'Propeller Chip MV4th16 Timer COG demo 
'
CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

PUB Start
  ''does nothing but must be there for spin IDE to compile
  
DAT
              org       0
Entry         mov       TimeAdr, PAR
              mov       DateAdr, PAR
              add       DateAdr, #4
              mov       LockAdr, PAR
              add       LockAdr, #8
              mov       TimeDelayAdr, PAR
              add       TimeDelayAdr, #12
              mov       LastTime, #0
              mov       LastDate, #0
              call      #ClrLock
              'Start time timer
              rdlong    TimeDelay, TimeDelayAdr
              mov       LoopWait, cnt
              add       LoopWait, TimeDelay
loop          waitcnt   LoopWait, TimeDelay
              'Ready to update the time
              'See if time was set
              call      #SetLock
              rdlong    T1, TimeAdr
              cmp       T1, LastTime            wz
              mov       LastTime, T1
        if_nz call      #GetTime
              rdlong    T1, DateAdr
              cmp       T1, LastDate            wz
              mov       LastDate, T1
        if_nz call      #GetDate
              call      #IncTime
              call      #SetTime
              call      #ClrLock
              'if we loaded a new time reset counter
        if_nz mov       LoopWait, cnt
        if_nz add       LoopWait, TimeDelay
              jmp       #loop
'
'This function sets the lock bit for the time data
SetLock       rdlong    T1, LockAdr
SetLock1      lockset   T1                      wc
        if_c  jmp       #SetLock1
SetLock_ret   ret
'                      
'This function clears the lock bit for the time data
ClrLock       rdlong    T1, LockAdr
              lockclr   T1
ClrLock_ret   ret
'                      
'This function reads the 32 bit BCD packet time value and unpacks
'the time value into the Hour, Minute, Second, SecHund variables
GetTime       rdlong    T1, TimeAdr
              'Get the hundreths of a second value.
              mov       SecHund, T1
              and       SecHund, #$FF
              'Get the seconds value.
              shr       T1, #8
              mov       Second, T1
              and       Second, #$FF
              'Get the minutes value.
              shr       T1, #8
              mov       Minute, T1
              and       Minute, #$FF
              'Get the hours value.
              shr       T1, #8
              mov       Hour, T1
GetTime_ret   ret
'
'                      
'This function reads the 32 bit BCD packet date value and unpacks
'the date value into the Year, Month, and Day variables
GetDate       rdlong    T1, DateAdr
              'Get the day value.
              mov       Day, T1
              and       Day, #$FF
              'Get the Month value.
              shr       T1, #8
              mov       Month, T1
              and       Month, #$FF
              'Get the year value.
              shr       T1, #8
              mov       Year, T1
              and       Year, #$FF
              'Get the Century value.
              shr       T1, #8
              mov       Century, T1
              and       Century, #$FF
GetDate_ret   ret
'
'
'This function builds a 32 bit BCD packed time. The 32 bit BCD
'is formed by packing the time values Hour, Minute, Second, SecHund. The
'result is saved to the 32 bit global memory address of the system time.
              'Set the hundreths of a second value.
SetTime       mov       T1, SecHund
              'Set the seconds value.
              mov       T2, Second
              shl       T2, #8
              or        T1, T2
              'Set the minutes value.
              mov       T2, Minute
              shl       T2, #16
              or        T1, T2
              'Set the hours value.
              mov       T2, Hour
              shl       T2, #24
              or        T1, T2
              wrlong    T1, TimeAdr
              call      #SetDate
SetTime_ret   ret
'
'
'This function builds a 32 bit BCD packed date. The 32 bit BCD
'is formed by packing the date values Year, Month, and Day. The
'result is saved to the 32 bit global memory address of the system date.
              'Set the day value.
SetDate       mov       T1, Day
              'Set the month value.
              mov       T2, Month
              shl       T2, #8
              or        T1, T2
              'Set the year value.
              mov       T2, Year
              shl       T2, #16
              or        T1, T2
              'Set the century value
              mov       T2, Century
              shl       T2, #24
              or        T1, T2
              'save the BCD date value to global memory
              wrlong    T1, DateAdr
SetDate_ret   ret
'
'
'This function increments the two digit BCD value in T1 (lower 8 bits)
BcdInc        add       T1, #1
              mov       T2, T1
              and       T2, #$0F
              cmp       T2, #$0A                 wc
        if_c  jmp       #BcdInc_ret                    
              'have a carry to tens digit
              and       T1, #$F0                'clear first digit
              add       T1, #$10                'increment tens digit
BcdInc_ret    ret
'
'
'This function adds 1 hundreth of a second to the time saved in the
'variables Hour, Minute, Second, and SecHund.
IncTime       mov       T1, SecHund
              call      #BcdInc
              mov       SecHund, T1
              cmp       SecHund, #$9A           wc
        if_c  jmp       #IncTime_ret
              'Hundreths of a seconds overflow 
              mov       SecHund, #0
              'increment seconds
              mov       T1, Second
              call      #BcdInc
              mov       Second, T1
              cmp       Second, #$5A             wc
        if_c  jmp       #IncTime_ret
              'Seconds overflow
              mov       Second, #0
              'increment minutes
              mov       T1, Minute
              call      #BcdInc
              mov       Minute, T1
              cmp       Minute, #$5A             wc
        if_c  jmp       #IncTime_ret
              'Minutes overflow
              mov       Minute, #0
              'increment hours
              mov       T1, Hour
              call      #BcdInc
              mov       Hour, T1
              cmp       Hour, #$24               wc
        if_c  jmp       #IncTime_ret
              'Minutes overflow
              mov       Hour, #0
              call      #IncDate
IncTime_ret   ret
'
'
'This function mutiplies the value in T1 by 10
MultTemp      long      0
Mult10        mov       MultTemp, T1
              shl       MultTemp, #3            'multiply by 8
              shl       T1, #1                  'multiply by 2
              add       T1, MultTemp
Mult10_ret    ret        
'
'
'Convert a packed BCD number to a byte The number is store in T1
BcdToByte               and     T1, #$FF        'mask off lower 8 bits
                        mov     T3, T1          'save packed BCD number
                        and     T3, #$0F        'T3 = least significant digit of BCD number
                        shr     T1, #4          'T1 = most significant digit of BCD number
                        call    #Mult10         'Multiply most significant digit by 10
                        add     T1, T3          'Add in least significant digit
BcdToByte_ret           ret
'
'
'This function returns the number of days in the current Month and Year.
'If the Month variable is out of range Month is set to 1 and days to 31.
'The number of days is returned in T1.
DaysInMonth             mov     T1, Month
                        call    #BcdToByte
                        'see if we are in range
                        cmp     Month, #1       wc
              if_c      jmp     #MonthOutRange
                        cmp     Month, #13      wc
              if_nc     jmp     #MonthOutRange
                        'Month in range (1..12) lookup days in table
                        sub     T1, #1
                        add     T1, #DaysPerMonth
                        movs    DaysInMonthRead, T1
                        nop
DaysInMonthRead         mov     T1, (T1)
                        'T1 has the days in current month
                        'see if month = 2 (feb) and variable Year is a leap year
                        cmp     Month, #2       wz
              if_nz     jmp     #DaysInMonth_ret
                        'Month = 2, see if we have a leap year in Year
                        mov     T2, T1                  'save number of days in T2
                        mov     T1, Year
                        call    #BcdToByte
                        test    T1, #3          wz
                        mov     T1, T2                  'restore days to T1                        
              if_nz     jmp     #DaysInMonth_ret
                        'Year divides by 4, if year divides by 100 evenly not a leap year
                        test    Year, #$00FF    wz
              if_z      jmp     #DaysInMonthCentury
                        'Year does not divide by 100 so we have a leap year
                        call    #BcdInc
                        jmp     #DaysInMonth_ret
                        'The year divides by 100 evenly see if it divides by 400
DaysInMonthCentury      mov     T2, T1                  'save number of days in T2
                        mov     T1, Century
                        call    #BcdToByte
                        test    T1, #3          wz
                        mov     T1, T2                  'restore days to T1                        
              if_nz     jmp     #DaysInMonth_ret
                        'Year divides by 400 so we have a leap year
                        call    #BcdInc
                        jmp     #DaysInMonth_ret
MonthOutRange           mov     Month, #$01
                        mov     T1, #$31             
DaysInMonth_ret         ret
'
'
'This function adds 1 day to the date saved in the
'variables Year, Month, and day.
IncDate                 mov     T1, Day
                        call    #BcdInc
                        mov     Day, T1
                        call    #DaysInMonth
                        cmp     Day, T1                 wc, wz
        if_c_or_z       jmp     #IncDate_ret
                        'Day of month overflow
                        mov     Day, #1
                        mov     T1, Month
                        call    #BcdInc
                        mov     Month, T1
                        cmp     Month, #$13             wc
        if_c            jmp     #IncDate_ret
                        'Month overflow
                        mov     Month, #1
                        'Increment the year
                        mov     T1, Year
                        call    #BcdInc
                        mov     Year, T1
                        cmp     Year, #$9A              wc
        if_c            jmp     #IncDate_ret
                        'Year overflow
                        mov     Year, #0
                        'Increment Century  
                        mov     T1, Century
                        call    #BcdInc
                        mov     Century, T1           
                        cmp     Century, #$9A           wc
        if_c            jmp     #IncDate_ret
                        'Century overflowed
                        mov     Century, #0
IncDate_ret             ret
'
'
'Constats start here
TimeAdr       long      0
DateAdr       long      0
LockAdr       long      0
TimeDelayAdr  long      0          
TimeDelay     long      800_000    '80mhz CPU clock ferquency divided by 100 for example
'Table of days in each month in packed BCD
DaysPerMonth  long      $31      'jan
              long      $28      'feb
              long      $31      'mar
              long      $30      'apr
              long      $31      'may
              long      $30      'jun
              long      $31      'jul
              long      $31      'aug
              long      $30      'sep
              long      $31      'oct
              long      $30      'nov
              long      $31      'dec
                        
'Variables start here
Century       res       1
Year          res       1
Month         res       1
Day           res       1
Hour          res       1
Minute        res       1
Second        res       1
SecHund       res       1
LoopWait      res       1
T1            res       1
T2            res       1
T3            res       1
LastTime      res       1
LastDate      res       1
