'MV4th16 Forth basic I/O EEPROM, I2C, COM1 Stuff
'
CON
  'I2C EEprom stuf
  'The prototype board has a 64K serial eeprom. The Prop chip only uses the first 32K leaving 32K for your program to store data.
  '
  'Changing the following constants should allow 32K to 128K serial EEProms to be supported. (only tested on 64K and 128K)
  c_EEpromSCLBit  = 28                                  'I2C clock signal pin number
  c_EEpromSDABit  = 29                                  'I2C data signal pin number
  c_EEpromAdrTx   = $A0                                 'First byte of eeprom write command
  c_EEpromAdrRx   = $A1                                 'First byte of eeprom read command
  
  { 32K EEPROM settings 16 blocks }
  {
  'EEprom address constants
  c_EEpromPageBytes        =  64                        'Bytes per eeprom page
  c_EEpromStartingAdr      =  $4000                     'Starting EEprom address of first page we can use.
                                                        'First 16K used by prop chip for forth program. 
  c_EEpromBlocks           =  16                        'Number of Forth blocks available (1024 bytes each)
  }

  { 64K EEPROM settings 32 blocks }
  {
  'EEprom address constants
  c_EEpromPageBytes        =  128                       'Bytes per eeprom page
  c_EEpromStartingAdr      =  $8000                     'Starting EEprom address of first page we can use.
                                                        'First 32K used by prop chip for forth program. 
  c_EEpromBlocks           =  32                        'Number of Forth blocks available (1024 bytes each)
  }

  { 64K EEPROM settings 48 blocks }
  'EEprom address constants
  c_EEpromPageBytes        =  128                       'Bytes per eeprom page
  c_EEpromStartingAdr      =  $4000                     'Starting EEprom address of first page we can use.
                                                        'First 16K used by prop chip for forth program. 
  c_EEpromBlocks           =  48                        'Number of Forth blocks available (1024 bytes each)

  { 128K EEPROM settings 96 blocks}
  {
  'EEprom address constants
  c_EEpromPageBytes        =  256                       'Bytes per eeprom page
  c_EEpromStartingAdr      =  $8000                     'Starting EEprom address of first page we can use.
                                                        'First 32K used by prop chip for forth program. 
  c_EEpromBlocks           =  96                        'Number of Forth blocks available (1024 bytes each)
  }

  { 128K EEPROM settings 112 blocks}
  {
  'EEprom address constants
  c_EEpromPageBytes        =  256                       'Bytes per eeprom page
  c_EEpromStartingAdr      =  $4000                     'Starting EEprom address of first page we can use.
                                                        'First 32K used by prop chip for forth program. 
  c_EEpromBlocks           =  112                       'Number of Forth blocks available (1024 bytes each)
  }

  'Forth Command constants
  c_ForthCmdCount    = 26
  'Forth com port fifo sizes
  c_ComTxFifoSize = Com#c_ComTxFifoSize
  c_ComRxFifoSize = Com#c_ComRxFifoSize
  c_bit_ticks_adr = Com#bit_ticks_adr
  'Fifo buffers
  c_ComRxFifoAdr  = Com#c_ComRxFifoAdr
  c_ComTxFifoAdr  = Com#c_ComTxFifoAdr
  'Start address of I/O data area, runs from c_ComDataStartAdr to $7FFF
  c_IODataStartAdr = Com#c_ComDataStartAdr
  
OBJ
  Com: "ForthBaseSerial"
  
VAR
  'com port changable settings
  word  settings_adr
  'rx fifo settings
  word  rx_head_adr             
  word  rx_tail_adr
  word  rx_buf_size
  word  rx_buf_adr
  'tx fifo settings
  word  tx_head_adr             
  word  tx_tail_adr
  word  tx_buf_size
  word  tx_buf_adr
  'com port hardware settings
  word  rx_pin
  word  tx_pin
  'End of com port settings
  
PUB Start(rxpin, txpin, mode, baudrate, SystemCommandAdr)
  long[SystemCommandAdr] := 0
  'Start serial port, 2 wire version
  Com.Start(rxpin, txpin, mode, baudrate, @settings_adr)
  ComRxHeadAdr := rx_head_adr
  ComRxTailAdr := rx_tail_adr
  ComRxBufferSize := rx_buf_size
  ComRxBufferAdr := rx_buf_adr
  ComTxHeadAdr := tx_head_adr
  ComTxTailAdr := tx_tail_adr
  ComTxBufferSize := tx_buf_size
  ComTxBufferAdr := tx_buf_adr
  'Start our cog if we can 
  cognew(@Entry, SystemCommandAdr)

DAT
'
'***********************************
'* Assembly language I2C Code      *
'***********************************
'
                        org   0
Entry                   rdlong  T1, #0                                          'read system frequency from long[0] in global memory
                        mov     T2, #500
                        call    #DoDivU32    
                        'Retry delay waiting for eeprom to ready ( @ .5ms )
                        'WriteWaitDelay := clkfreq / 2000 = 0.5ms
                        mov     WriteWaitDelay, T3                'WriteWaitDelay = system clocks per milli-second
                        'Initialize eeprom
                        call    #I2C_InitEEprom
                        call    #F_CMD_I2C_RESET
                        'Process commands
                        jmp     #F_Do_Cmds
'
'See if we have a valid I2C command. If we do execute the command.
'The I2CCmdAdr contains the address of the long holding
'the I2C command to perform. The next three longs contain;
'EEprom address, Global memory address, and Byte count.
'Commands:
'        0 - no command
'        1 - EMIT write to serial device
'        2 - KEY read from serial device
'        3 - KEY? see if KEY will not block
'        4 - EMIT? see if EMIT will not block
'        5 - EMPTY-KEYS flush the serial rx buffer
'        6 - Copy EEprom to global memory
'        7 - Copy global memory to EEProm page
'        8 - Retrive block device block count
'        9 - Start new COG (auto select COD index)
'       10 - Initialize the selected COG
'       11 - Stop running COG by index
'       12 - Allocate a lock
'       13 - Release a lock
'       14 - FILL ( c-addr u c -- ) fill memory address c-adr with u characters c
'       15 - CMOVE ( c-addr1 c-addr2 u -- )
'       16 - CMOVE> ( c-addr1 c-addr2 u -- )
'       17 - BAUD ( d -- ) set serial port baud rate
' These commands require the "I2C Interface.forth" file be loaded
'       18 - I2C-RESET ( -- )  
'       19 - I2C-READ ( c-adr u -- )  
'       20 - I2C-WRITE ( c-adr u -- f )  
'       21 - I2C-START ( -- )
'       22 - I2C-STOP ( -- )
'       23 - I2C-FAST ( -- )
'       24 - I2C-SLOW ( -- )
' End of commands that require that "I2C Interface.forth" be loaded
'       25 - REBOOT ( -- )
'
'
'Initialize our I2C serial 64KB eeprom.
'
I2C_InitEEprom          cmp     t1, t1          wz       'Make Z flag = 1
                        'make sure SCL = 0 at eeprom
                        muxnz   outa, I2C_SCL_Bit        'SCL bit = 0
                        muxz    dira, I2C_SCL_Bit        'SCL dir = output, eeprom SCL = 0
                        call    #I2C_BitDelay
                        'make sure SDA = 1 at eeprom
                        muxnz   outa, I2C_SDA_Bit        'SDA output register bit = 0
                        muxnz   dira, I2C_SDA_Bit        'Set SDA as input = 1 at eeprom
I2C_InitEEprom_ret      ret
'
'                        
'Perform I2C one bit delay (Delay > 0.5usec), max eeprom clock = 1mhz
loop    long  0    
I2C_BitDelay            mov     loop, I2C_BIT_DELAY
I2C_BitDelay_Next       djnz    loop, #I2C_BitDelay_Next 
I2C_BitDelay_ret        ret
'
'
'Set SCL to low
I2C_SCL_LOW             cmp     t1, t1          wz       'Make Z flag = 1
                        'make sure SCL = 0 at eeprom
                        muxnz   outa, I2C_SCL_Bit        'SCL bit = 0
                        call    #I2C_BitDelay
I2C_SCL_LOW_ret         ret
'                        
'
'Set SCL to low
I2C_SCL_HIGH            cmp     t1, t1          wz       'Make Z flag = 1
                        'make sure SCL = 1 at eeprom
                        muxz    outa, I2C_SCL_Bit        'SCL bit = 1
                        call    #I2C_BitDelay
I2C_SCL_HIGH_ret        ret
'                        
'
'Set SDA to low, also SDA is output
I2C_SDA_LOW             cmp     t1, t1          wz       'Make Z flag = 1
                        muxz    dira, I2C_SDA_Bit        'Set SDA as output, data set to 0 at init
                        call    #I2C_BitDelay
I2C_SDA_LOW_ret         ret
'                        
'
'Set SDA to high, also SDA is input
I2C_SDA_HIGH            cmp     t1, t1          wz       'Make Z flag = 1
                        muxnz   dira, I2C_SDA_Bit        'Set SDA as input, pull up resistor forces = 1 at eeprom
                        call    #I2C_BitDelay
I2C_SDA_HIGH_ret        ret
'                        
'
'Perform a I2C start bit, SDA goes low while SCL is high
F_CMD_I2C_START         call    #I2C_SCL_LOW
                        call    #I2C_SDA_HIGH
                        'SCL = 0, SDA = output = 1 at eeprom
                        call    #I2C_SCL_HIGH
                        call    #I2C_SDA_LOW
                        'Next set SCL = 0, then wait
                        call    #I2C_SCL_LOW
                        'Finally bring SDA = 1
                        call    #I2C_SDA_HIGH
                        'we always exit with SCL = 0, SDA = 1
F_CMD_I2C_START_ret     ret
'                        
'
'Perform a I2C stop bit, SDA goes low while SCL is high
F_CMD_I2C_STOP          call    #I2C_SCL_LOW
                        call    #I2C_SDA_LOW
                        'SCL = 0, SDA = output = 0 at eeprom
                        call    #I2C_SCL_HIGH
                        call    #I2C_SDA_HIGH
                        'Next set SCL = 0, then wait
                        call    #I2C_SCL_LOW
                        'we always exit with SCL = 0, SDA = 1
F_CMD_I2C_STOP_ret      ret
'                        
'
'18) I2C-RESET ( -- )
'reset the I2C interface
'Software reset I2C eeprom
F_CMD_I2C_RESET         call    #F_CMD_I2C_START
                        'Pulse SCL 9 times with SDA = 1
                        mov     t1, #63
I2C_ResetRpt            call    #I2C_SCL_HIGH
                        call    #I2C_SCL_LOW
                        djnz    t1, #I2C_ResetRpt
                        'Perform another start bit
                        call    #F_CMD_I2C_START
                        'Perform stop bit
                        call    #F_CMD_I2C_STOP
                        'we alway exit with SCL = 0, SDA = 1
F_CMD_I2C_RESET_ret     ret
'                        
'
'Read a byte from the I2C buss.
'On entry T2 = 0 to write ack = 0, T2 <> 0 ack = 1
'On exit t1 = data byte. 
I2C_ReadByte            mov     t1, #0
                        call    #I2C_SCL_LOW
                        call    #I2C_SDA_HIGH   'Set SDA = input                                                        
                        mov     t3, #8          'Need to receive 8 bit plus send ack/nak
I2C_ReadByteRpt         shl     t1, #1          'Shift bits ready for next msb of byte
                        'now set SCL high then low to clock data bit
                        call    #I2C_SCL_HIGH
                        'read data bit
                        mov     t4, ina
                        and     t4, I2C_SDA_Bit wz
              if_nz     or      t1, #1                        
                        'Set SCL = 0
                        call    #I2C_SCL_LOW
                        djnz    t3, #I2C_ReadByteRpt
                        'clocked in our eight bits, 
                        'now clock out our ack bit
                        mov     t2, t2          wz      'see if ack is 1 or 0
              if_z      call    #I2C_SDA_LOW                                                           
                        mov     t2, t2          wz      'see if ack is 1 or 0
              if_nz     call    #I2C_SDA_HIGH
                        'Clock SCL to send ack/nak         
                        call    #I2C_SCL_HIGH
                        call    #I2C_SCL_LOW
                        'we alway exit with SCL = 0, SDA = 1
                        call    #I2C_SDA_HIGH
I2C_ReadByte_ret        ret
'
'
'Write a byte to the I2C buss, MSb first.
'After byte is sent eeprom returns ack/nak on 9th clock.
'On entry T1 = data byte to write.
I2C_WriteByte           shl     t1, #24         'move data byte into MSB of T1
                        mov     t2, #8          'number of bits to write to eeprom
I2C_WriteByteRpt        shl     t1, #1    wc    'Move next bit into carry flag
              if_c      call    #I2C_SDA_HIGH                                                           
              if_nc     call    #I2C_SDA_LOW
                        'now set SCL high then low to clock data bit
                        call    #I2C_SCL_HIGH
                        call    #I2C_SCL_LOW
                        djnz    t2, #I2C_WriteByteRpt
                        'clocked out our eight bits, set SDA as input to read act
                        call    #I2C_SDA_HIGH
                        call    #I2C_SCL_HIGH
                        mov     EEpromResp, ina
                        and     EEpromResp, I2C_SDA_Bit
                        'we alway exit with SCL = 0, SDA = 1
                        call    #I2C_SCL_LOW
I2C_WriteByte_ret       ret
'
'
'Local variables for I2C read/write to global memory functions.
EEpromAdr     long      0
MemAdr        long      0
CopyCnt       long      0
EEpromResp    long      0
'
'
'Set the eeprom start address for a write, also use for read
'On Entry: T1 = eeprom start address, T2 = global memory address
'T3 = byte count to copy.
I2C_StartWrite          mov     EEpromResp, #0                                  'Indicate a valid read
                        mov     MemAdr, t2
                        mov     CopyCnt, t3     wz
              if_z      jmp     #I2C_StartWrite_ret                             'if nothing to copy just quit
                        'bit 16 of address goes in first command byte B[1]
                        mov     EEpromAdr, #c_EEpromAdrTx
                        test    t1, Bit16Mask   wz
              if_nz     or      EEpromAdr, #2
                        'add in remaing 16 bits of address
                        shl     EEpromAdr, #16
                        and     t1, DataWordMask
                        or      EEpromAdr, t1
                        'First send a start bit
                        call    #F_CMD_I2C_START
                        'next send a write command
                        mov     t1, EEpromAdr
                        shr     t1, #16
                        call    #I2C_WriteByte
                        tjnz    EEpromResp, #I2C_StartWrite_ret
                        'Send the MSB for 16 bit address
                        mov     t1, EEpromAdr
                        shr     t1, #8
                        call    #I2C_WriteByte
                        tjnz    EEpromResp, #I2C_StartWrite_ret
                        'Send the LSB for 16 bit address
                        mov     t1, EEpromAdr
                        call    #I2C_WriteByte
I2C_StartWrite_ret      ret
'
'
'Copy a region from the eeprom to global memoru
'On Entry: T1 = eeprom start address, T2 = global memory address
'T3 = byte count to copy.
'On Exit: EEpromResp = 0 if OK or <> 0 if failed
I2C_CopyToMem           call    #I2C_StartWrite                                         
                        tjnz    EEpromResp, #I2C_CopyToMem_ret                  'if EEpromResp <> 0, no responce from EEprom
                        'Send a start bit to exit write mode
                        call    #F_CMD_I2C_START
                        'finally send command to read
                        mov     t1, EEpromAdr
                        shr     t1, #16
                        or      t1, #1                  'make it a read command (Cmd[0] = 1)
                        call    #I2C_WriteByte
                        tjnz    EEpromResp, #I2C_CopyToMem_ret
                        'now read the bytes up to count -1
                        sub     CopyCnt, #1     wz
              if_z      jmp     #I2C_CopyToMemLast
I2C_CopyToMemNext       mov     t2, #0          'Send ACK = 0
                        call    #I2C_ReadByte
                        wrbyte  t1, MemAdr
                        add     MemAdr, #1
                        djnz    CopyCnt, #I2C_CopyToMemNext
                        'Read all the bytes except one, needs a nak
I2C_CopyToMemLast       mov     t2, #1           'Send NAK = 1
                        call    #I2C_ReadByte
                        wrbyte  t1, MemAdr
                        call    #F_CMD_I2C_STOP
I2C_CopyToMem_ret       ret
'
'
'Wait for eeprom to ready after a write page, normaly < 5 Msec
I2C_WriteReady          call    #F_CMD_I2C_START
                        mov     t1, #c_EEpromAdrTx                              'Actual address does not matter only device select bits
                        call    #I2C_WriteByte
                        call    #F_CMD_I2C_STOP
I2C_WriteReady_ret      ret
'
' 
'Write data to a page on the eeprom
'On Entry: T1 = eeprom start address, T2 = global memory address
'T3 = byte count to copy.
I2C_WritePage           call    #I2C_StartWrite                                         
                        tjnz    EEpromResp, #I2C_WritePage_ret                  'if EEpromResp <> 0, no responce from EEprom
                        'now write the page data
I2C_WritePageNext       rdbyte  t1, MemAdr
                        call    #I2C_WriteByte
                        tjnz    EEpromResp, #I2C_WritePage_ret
                        add     MemAdr, #1
                        djnz    CopyCnt, #I2C_WritePageNext
                        call    #F_CMD_I2C_STOP
I2C_WritePage_ret       ret
'
'Write data to a page on the eeprom, then wait for eeprom to ready.
'If the eeprom takes more then 50ms the wait is aborted and EEpromResp
'will be set to indicate the error.
'On Entry: T1 = eeprom start address, T2 = global memory address
'T3 = byte count to copy.
WaitReryCnt             long    0
'
I2C_WriteWait           call    #I2C_WritePage
                        tjnz    EEpromResp, #I2C_WriteWait_ret                  'if EEpromResp <> 0, no responce from EEprom
                        'now wait for the EEprom to complete the write max of 50ms
                        mov     WaitReryCnt, #100       'Delay is .5ms per loop @ 50ms total
                        'delay for 500us
I2C_WriteWaitDelay      mov     T4, cnt
                        add     T4, WriteWaitDelay
                        waitcnt T4, #0
                        'see if eeprom is ready
                        call    #I2C_WriteReady                                                
                        tjz     EEpromResp, #I2C_WriteWait_ret                  'if EEpromResp = 0, EEprom is ready
                        'eeprom is not ready see if we should try again
                        djnz    WaitReryCnt, #I2C_WriteWaitDelay
I2C_WriteWait_ret       ret
'
'
CopyFromMemAdr          long    0
CopyFromEEpromAdr       long    0
CopyFromByteCount       long    0
CopyFromBlockBytes      long    0
'
I2C_CopyParams          mov     CopyFromMemAdr, T2
                        mov     CopyFromEEpromAdr, T1
                        mov     CopyFromByteCount, T3
I2C_CopyParams_ret      ret
'
'Copy a region from global memoru to the eeprom
'On Entry: T1 = eeprom start address, T2 = global memory address
'T3 = byte count to copy.
'On Exit: EEpromResp = 0 if OK or <> 0 if failed
                        
I2C_CopyFromMem         call    #I2C_CopyParams
                        'see if we are starting the first write on a page boundry.
                        'Page bytes are 32, 64, 128 or 256 bytes, dec by one to get
                        'Page offset mask
I2C_CopyFromMemNext     mov     T2, #c_EEpromPageBytes
                        sub     T2, #1
                        and     T2, CopyFromEEpromAdr   wz                      'see if we are on a page boundry
                        mov     T3, #c_EEpromPageBytes                          'Assume a full page write
              if_nz     sub     T3, T2                                          'We have a partial page to write
                        'See if the number of bytes to write is less then byte we want to write
                        cmp     CopyFromByteCount, T3   wc
              if_c      mov     T3, CopyFromByteCount
                        'We have a copy count so prepare to write.
                        mov     CopyFromBlockBytes, T3
                        mov     T1, CopyFromEEpromAdr
                        mov     T2, CopyFromMemAdr
                        call    #I2C_WriteWait
                        tjnz    EEpromResp, #I2C_CopyFromMem_ret                'if EEpromResp = 0, EEprom is ready
                        'We wrote the block, Update pointers
                        add     CopyFromMemAdr, CopyFromBlockBytes
                        add     CopyFromEEpromAdr, CopyFromBlockBytes
                        sub     CopyFromByteCount, CopyFromBlockBytes           wc, wz
                        'see if we need to write another page
        if_nc_and_nz    jmp     #I2C_CopyFromMemNext
I2C_CopyFromMem_ret     ret
'
                        
'FORTH system command support routines
'
' Remove top 16 bit value from forth stack.T1 = 16 bit value on return
                        'make sure we are in a command and stack pointer is valid
F_PopStack              tjz     ForthStackPtr, #F_PopStack_Ret
                        'We are in a command OK to pop value from stack
                        rdword  T1, ForthStackPtr
                        add     ForthStackPtr, #2
F_PopStack_Ret          ret             
'
' Add 16 bit value to top of forth stack.T1 = 16 bit value to write.
                        'make sure we are in a command and stack pointer is valid
F_PushStack             tjz     ForthStackPtr, #F_PushStack_Ret
                        'We are in a command OK to push a value to the stack
                        sub     ForthStackPtr, #2
                        wrword  T1, ForthStackPtr
F_PushStack_Ret         ret             
'
'
'Unsigned 32 bit division with 32 bit result and 32 bit remainder.
'On entry T1 - dividend, T2 = divisor. On exit T3 = T1 / T2 
DoDivU32                tjnz    T2, #DoDivU32_1         'If T2 = 0 we have a divide by zero error
                        'Handle division by zero, return max value and remainder of 0
                        mov     T3, #0
                        sub     T3, #1
                        jmp     #DoDivU32_ret
DoDivU32_1              'Perform a normal division                                                              
                        mov     T3, #0
                        mov     T4, #0
                        mov     T5, #32                                   
                        'Do the shift and sub a max of 32 times.
DoDivU32_2              shl     T1, #1          wc
                        rcl     T4, #1
                        cmpsub  T4, T2         wc      'C flag is set (1) if a subtraction is possible.
                        rcl     T3, #1
                        djnz    T5, #DoDivU32_2
DoDivU32_ret            ret
'
'
'read or write a 1k block to or from the block device using supplied block number and memory address.
'A block is 1024 bytes and will span multiple eeprom pages.
'Read/Write block ( u adr -- f ), u = mass storage block number, adr = buffer memory address,
'f = 0 I/O failed, f = -1 I/O performed.
IO_OperationCode        long    0                'Set to 0 for read 1 for write
'
F_DO_BLK_IO             call    #F_PopStack     'T1 = memory address
                        mov     T2, T1          'T2 = global memory address
                        call    #F_PopStack     'T1 = block number
                        'see if block number in range
                        cmp     T1, #c_EEpromBlocks     wc
        if_nc           mov     T1, #0
        if_nc           jmp     #F_DO_BLK_IO_DONE
                        'Request is valid, try to read/write eeprom device
                        shl     T1, #10         'T1 = (BlockNumber * 1024) 
                        add     T1, EEpromStartingAdr   'T1 = eeprom start address of selected block
                        mov     T3, #1
                        shl     T3, #10         'T3 = 1024
                        'we have a valid request and all params are set, read/write eeprom
                        cmp     IO_OperationCode, #0    wz
        if_z            call    #I2C_CopyToMem
                        cmp     IO_OperationCode, #0    wz
        if_nz           call    #I2C_CopyFromMem
                        'I/O operation completed report results
                        cmp     EEpromResp, #0          wz
                        mov     T1, DataWordMask        'T1 = -1 16 bits
        if_nz           mov     T1, #0
                        'push result to stack and exit
F_DO_BLK_IO_DONE        call    #F_PushStack
F_DO_BLK_IO_ret         ret
'
'Run the cog T3 = #8 COG auto selected, 0..7 run on selected cog
COG_RUN                 call    #F_PopStack     'T1 = PAR (14 bit) value
                        mov     T2, T1          'T2 = PAR value
                        call    #F_PopStack     'T1 = COG (14 bit ) code address
                        shr     T2, #2          'Convert T2 to 14 bit number
                        shr     T1, #2          'Convert T1 to 14 bit number
                        'build destination word
                        shl     T2, #18
                        shl     T1, #4
                        or      T1, T2
                        or      T1, T3
                        'params are ready start new cog
                        coginit T1              wc, wr
              if_c      mov     T1, DataWordMask
              if_nc     and     T1, #7
COG_RUN_ret             ret
'
'
'FORTH system commands
'
'1) EMIT ( c -- ) Write char to console, c = 8 bit char to send to console.
F_CMD_EMIT              call    #F_PopStack             'T1 = character value to send
                        rdword  T2, ComTxHeadAdr        'T2 = tx fifo head
F_CMD_EMIT_WAIT         rdword  T3, ComTxTailAdr        'T3 = tx fifo Tail
                        'see if fifo is full (Head + 1 == Tail)
                        mov     T4, T2                  'T4 = tx fifo head
                        add     T4, #1
                        cmp     T4, ComTxBufferSize     wz
        if_z            mov     T4, #0                  'Head wrapped back to fifo start
                        cmp     T4, T3                  wz
        if_z            jmp     #F_CMD_EMIT_WAIT       'Head + 1 = Tail, no room in fifo, try again
                        'There is room write character
                        mov     T5,  ComTxBufferAdr
                        add     T5, T2                  'Calculate tx fifo write address
                        wrbyte  T1, T5
                        wrword  T4, ComTxHeadAdr        'update tx fifo head with head new value
F_CMD_EMIT_ret          ret
'
'
'2) KEY ( -- c ) Read char from console, c = 8 bit char read from console.
F_CMD_KEY               rdword  T3, ComRxTailAdr        'T3 = tx fifo Tail
F_CMD_KEY_WAIT          rdword  T2, ComRxHeadAdr        'T2 = tx fifo Head
                        cmp     T2, T3                  wz
        if_z            jmp     #F_CMD_KEY_WAIT
                        'read from fifo
                        mov     T4, ComRxBufferAdr
                        add     T4, T3
                        rdbyte  T1, T4
                        call    #F_PushStack
                        'Update tail to show we read fifo
                        add     T3, #1
                        cmp     T3, ComRxBufferSize      wz
        if_z            mov     T3, #0
                        wrword  T3, ComRxTailAdr        'update tx fifo Tail with T3
F_CMD_KEY_ret           ret
'
'
'3) KEY? ( -- f )
'Test if console has data to read, f = 0 no char ready
'from console, f= -1 char available from console.
F_CMD_CAN_KEY           rdword  T3, ComRxTailAdr        'T3 = tx fifo Tail
                        rdword  T2, ComRxHeadAdr        'T2 = tx fifo Head
                        cmp     T2, T3                  wz
        if_z            mov     T1, #0
        if_nz           mov     T1, DataWordMask
                        call    #F_PushStack
F_CMD_CAN_KEY_ret       ret
'
'
'4) EMIT? ( -- f ) f = -1 can send, f= 0 unable to send.
F_CMD_CAN_EMIT          rdword  T2, ComTxHeadAdr        'T2 = tx fifo head
                        rdword  T3, ComTxTailAdr        'T3 = tx fifo Tail
                        add     T2, #1
                        cmp     T2, ComTxBufferSize     wc
        if_nc           mov     T2, #0                  'Head wrapped back to fifo start
                        cmp     T2, T3                  wz
        if_z            mov     T1, #0
        if_nz           mov     T1, DataWordMask
                        call    #F_PushStack
F_CMD_CAN_EMIT_ret      ret
'
'
'5) EMPTYKEYS ( -- ) Request the console input buffer be cleared.
F_CMD_EMPTYKEYS         mov     T1, #0
                        wrword  T1, ComRxHeadAdr        'rx fifo head = 0
                        wrword  T1, ComRxTailAdr        'rx fifo Tail = 0
                        call    #F_CompleteCommand
F_CMD_EMPTYKEYS_ret     ret
'
'
'6) READ_BLK ( block mem-adr -- f )
'Read block from eeprom device to memory buffer.
'f = 0 operation failed, f = -1 (16 bit) operation completed.
F_CMD_READ_BLK          mov     IO_OperationCode, #0
                        call    #F_DO_BLK_IO
F_CMD_READ_BLK_ret      ret
'
'
'7) WRITE_BLK ( block mem-adr -- f )
'Write forth block in memory to the eeprom device.
'f = 0 operation failed, f = -1 (16 bit) operation completed.
F_CMD_WRITE_BLK         mov     IO_OperationCode, DataWordMask
                        call    #F_DO_BLK_IO
F_CMD_WRITE_BLK_ret     ret
'
'
'8) BLK_COUNT ( -- u )
'Returns number blocks avaulable, u = mass storage block count.
F_CMD_BLK_COUNT         mov     T1, #c_EEpromBlocks
                        call    #F_PushStack
F_CMD_BLK_COUNT_ret     ret
'
'
'9) COD-INIT ( Code-adr u1 u2 -- )
'Start the COG index number u2, if the COG is already running stop the COG and restart
'the COG with the new code. q-adr1 points to a long aligned address containg
'the cog code to execute. u1 (upper 14 bits) is passes via the COG PAR register.
F_CMD_COG_INIT          call    #F_PopStack     'T1 = PAR (14 bit) value
                        mov     T3, T1          'T2 = PAR value
                        and     T3, #7
                        call    #COG_RUN
F_CMD_COG_INIT_ret      ret
'
'                      
'10) COG-NEW ( Code-adr u1 -- u2 )
'Start a new COG Code-adr points to a long aligned address containg
'the cog code to execute. u1 (upper 14 bits) is passes via the COG PAR register.
'The function returns the COG number started in u2 or u2 = -1 if no cog available.
F_CMD_COG_NEW           mov     T3, #8          'T3 = auto COG select
                        call    #COG_RUN
                        'save result to forth data stack
                        call    #F_PushStack
F_CMD_COG_NEW_ret       ret
'
'
'11) COG-STOP ( u -- )
'Stop the COG index number u if it is running.
F_CMP_COG_STOP          call    #F_PopStack     'T1 = cog index to stop
                        cogstop T1
F_CMP_COG_STOP_ret      ret
'
'
'12) LOCK_NEW ( -- u )
'Allocate a lock u is the lock index or -1 if no locks avaiable
'Execute next system command.
F_CMD_LOCK_NEW          locknew T1              wc
        if_c            mov     T1, DataWordMask
                        call    #F_PushStack
F_CMD_LOCK_NEW_ret      ret
'
'
'13) LOCK_RET ( u -- )
'Release a lock, u is the lock index to release/return
F_CMD_LOCK_RET          call    #F_PopStack     'T1 = lock index to return
                        lockret T1              wc
F_CMD_LOCK_RET_ret      ret
'
'14) C FILL   ( c-addr u char -- ) fill memory with char
'u bytes are filled starting at address c-adr
F_CMD_FILL              call    #F_PopStack     'T1 = char to fill with
                        mov     T2, T1          'T2 = char to fill with
                        call    #F_PopStack     'T1 = fill count
                        mov     T3, T1          'T3 = fill count
                        call    #F_PopStack     'T1 = fill address
                        'Fill memory bytes (count, T3)
                        tjz     T3, #F_CMD_FILL_ret
                        'Count is positive fill memory                        
F_CMD_FILL_NXT          wrbyte  T2, T1
                        add     T1, #1
                        djnz    T3, #F_CMD_FILL_NXT
F_CMD_FILL_ret          ret     
'
'15) CMOVE ( c-addr1 c-addr2 u -- )
'If u is greater than zero, copy u consecutive characters from the data space starting at c-addr1
'to that starting at c-addr2, proceeding character-by-character from lower addresses to higher
'addresses.
F_CMD_CMOVE             call    #F_PopStack     'T1 = char to copy
                        mov     T3, T1          'T3 = chars to copy
                        call    #F_PopStack     'T1 = dest address
                        mov     T2, T1          'T2 = dest address
                        call    #F_PopStack     'T1 = source address
                        'Fill memory bytes (count, T3)
                        tjz     T3, #F_CMD_CMOVE_ret
F_CMD_CMOVE_NEXT        rdbyte  T4, T1
                        add     T1, #1
                        wrbyte  T4, T2     
                        add     T2, #1
                        djnz    T3, #F_CMD_CMOVE_NEXT
F_CMD_CMOVE_ret         ret
'
'16) CMOVE> ( c-addr1 c-addr2 u -- )
'If u is greater than zero, copy u consecutive characters from the data space starting at c-addr1
'to that starting at c-addr2, proceeding character-by-character from higher addresses to lower
'addresses.
F_CMD_CMOVE_REV         call    #F_PopStack     'T1 = char to copy
                        mov     T3, T1          'T3 = chars to copy
                        call    #F_PopStack     'T1 = dest address
                        mov     T2, T1          'T2 = dest address
                        call    #F_PopStack     'T1 = source address
                        'copy from end of region to start
                        add     T1, T3
                        sub     T1, #1
                        add     T2, T3
                        sub     T2, #1
                        'Do copy operation
F_CMD_CMOVE_REV1        rdbyte  T4, T1
                        sub     T1, #1
                        wrbyte  T4, T2     
                        sub     T2, #1
                        djnz    T3, #F_CMD_CMOVE_REV1
F_CMD_CMOVE_REV_ret     ret                        
'
'17) BAUD ( d -- ) set serial port baud rate
F_CMD_BAUD              call    #F_PopStack     'T1 = high word of double
                        mov     T2, T1          'T2 = high word of double 
                        shl     T2, #16         'move word into bits 16..31
                        call    #F_PopStack     'T1 = low word of double
                        or      T2, T1          'T2 = 32bit double value of baud rate
                        rdlong  T1, #0          'read the system frequency from global memory
                        call    #DoDivU32
                        wrlong  T3, ComBitTicksAdr
F_CMD_BAUD_ret          ret
'
'19) I2C-READ ( uDataAdr uByteCount -- )
'read specified number of bytes from I2C interface and save to global memory
F_CMD_I2C_READ          call    #F_PopStack     'CopyCnt = number bytes to copy
                        mov     CopyCnt, T1
                        call    #F_PopStack     'MemAdr = global memory address
                        mov     MemAdr, T1
                        'loop and copy from I2C to memory
F_CMD_I2C_READ_NEXT     mov     t2, #0          'Send ACK = 0
                        call    #I2C_ReadByte
                        wrbyte  T1, MemAdr
                        add     MemAdr, #1
                        djnz    CopyCnt, #F_CMD_I2C_READ_NEXT
F_CMD_I2C_READ_ret      ret
'
'20) I2C-WRITE ( uDataAdr uByteCount -- fAcked )
'read specified number og bytes from I2C interface
F_CMD_I2C_WRITE         mov     EEpromResp, #0  'Indicate no error
                        call    #F_PopStack     'CopyCnt = number bytes to copy
                        mov     CopyCnt, T1
                        call    #F_PopStack     'MemAdr = global memory address
                        mov     MemAdr, T1
                        'loop and copy from I2C to memory
F_CMD_I2C_WRITE_NEXT    mov     t2, #0          'Send ACK = 0
                        rdbyte  T1, MemAdr
                        add     MemAdr, #1
                        call    #I2C_WriteByte
                        tjnz    EEpromResp, #F_CMD_I2C_WRITE_ERROR
                        djnz    CopyCnt, #F_CMD_I2C_WRITE_NEXT
                        'All done now report positive result and return
                        mov     T1, DataWordMask
                        jmp     #F_CMD_I2C_WRITE_DONE
                        '( did not get responce report error )
F_CMD_I2C_WRITE_ERROR   mov     T1, #0
F_CMD_I2C_WRITE_DONE    call    #F_PushStack
F_CMD_I2C_WRITE_ret     ret
'

'23) I2C-FAST ( -- )
'Set the I2C interface to run at @ 1 Mbs.
F_CMD_I2C_FAST          mov    I2C_BIT_DELAY, #1
F_CMD_I2C_FAST_ret     ret

'24) I2C-SLOW ( -- )
'Set the I2C interface to run at @ 400 Kbs.
F_CMD_I2C_SLOW          mov    I2C_BIT_DELAY, #10
F_CMD_I2C_SLOW_ret      ret
'
'31) REBOOT ( -- )
'Hardware reset of CPU
F_CMD_REBOOT            mov     T1, #$80
                        clkset  T1
F_CMD_REBOOT_WAIT       jmp     #F_CMD_REBOOT_WAIT
'
'                        
'Wait for a system command to be requested by forth engine
F_WaitForCmd            rdlong  ForthCmd, par
                        test    ForthCmd, DataWordMask  wz
        if_z            jmp     #F_WaitForCmd
                        'Seperate stack address from command number
                        mov     ForthStackPtr, ForthCmd
                        shr     ForthStackPtr, #16
                        and     ForthCmd, DataWordMask
                        'see if we have a valid command
                        cmp     ForthCmd, #c_ForthCmdCount                      wc
        if_c            jmp     #F_WaitForCmd_ret
                        ' we have an invalid command
                        call    #F_CompleteCommand                        
                        jmp     #F_WaitForCmd
                        ' we have a valid command, get the stack pointer address
F_WaitForCmd_Ret        ret

'
'Update stack pointer then exit
F_SaveStackptr          mov     T1, par
                        add     T1, #2
                        wrword  ForthStackPtr, T1
'Indicate system command requested by forth engine is completed
F_CompleteCommand       mov     ForthCmd, #0                          
                        wrword  ForthCmd, par
                        mov     ForthStackPtr, #0
F_SaveStackptr_ret                              
F_CompleteCommand_ret   ret
'
'
F_Do_Cmds               call    #F_WaitForCmd
                        'we have a cmd index 1..25
                        mov     T1, #F_JUMP_TABLE
                        add     T1, ForthCmd
                        movs    :F_Set_Cmd_Jump, T1
                        nop
:F_Set_Cmd_Jump         mov     :F_Do_Cmd_Jump, (T1)
                        nop
:F_Do_Cmd_Jump          nop
                        'The called function will return here
F_Do_Cmds_Return        call    #F_SaveStackptr
                        jmp     #F_Do_Cmds
'
'   
'Table of Jump opcodes for each of the base I/O functions
F_JUMP_TABLE            nop                             '(0)
                        call    #F_CMD_EMIT             '(1)
                        call    #F_CMD_KEY              '(2)
                        call    #F_CMD_CAN_KEY          '(3)
                        call    #F_CMD_CAN_EMIT         '(4)
                        call    #F_CMD_EMPTYKEYS        '(5)
                        call    #F_CMD_READ_BLK         '(6)
                        call    #F_CMD_WRITE_BLK        '(7)
                        call    #F_CMD_BLK_COUNT        '(8)
                        call    #F_CMD_COG_NEW          '(9)
                        call    #F_CMD_COG_INIT         '(10)
                        call    #F_CMP_COG_STOP         '(11)
                        call    #F_CMD_LOCK_NEW         '(12)
                        call    #F_CMD_LOCK_RET         '(13)
                        call    #F_CMD_FILL             '(14)
                        call    #F_CMD_CMOVE            '(15)
                        call    #F_CMD_CMOVE_REV        '(16)
                        call    #F_CMD_BAUD             '(17)
                        call    #F_CMD_I2C_RESET        '(18)
                        call    #F_CMD_I2C_READ         '(19)
                        call    #F_CMD_I2C_WRITE        '(20)
                        call    #F_CMD_I2C_START        '(21)
                        call    #F_CMD_I2C_STOP         '(22)
                        call    #F_CMD_I2C_FAST         '(23)
                        call    #F_CMD_I2C_SLOW         '(24)
                        jmp     #F_CMD_REBOOT           '(25)
'
'
'                        
'Pre defined constants
I2C_SCL_Bit             long    1  << c_EEpromSCLBit
I2C_SDA_Bit             long    1  << c_EEpromSDABit
Bit16Mask               long    1 << 16
DataWordMask            long    $FFFF
'
EEpromStartingAdr       long    c_EEpromStartingAdr
I2C_BIT_DELAY           long    1
'
'Forth command variables ForthCmdAdr (16 bits) followed directly by StackPtrAdr (16 bits)
ForthCmd                long    0
ForthStackPtr           long    0
'
ComBitTicksAdr          long    c_bit_ticks_adr
'Com port rx fifo addresses in global memory
ComRxHeadAdr            long    0
ComRxTailAdr            long    0
ComRxBufferSize         long    0
ComRxBufferAdr          long    0
'Com port tx fifo addresses in global memory
ComTxHeadAdr            long    0
ComTxTailAdr            long    0
ComTxBufferSize         long    0
ComTxBufferAdr          long    0

'Variables initialized by start up code
'
'Uninitialized data
t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1
t5                      res     1
t6                      res     1
WriteWaitDelay          res     1
'
' ******** END OF COG CODE ********
'