''MV4th16 is a port of ANS 83 Forth to the Propeller Chip.
''MV4th16 uses a 16 bit cell size and 32 bit double percision numbers
''character data is one byte in size.
''       
''I developed this code from four main sources;
''CamelForth for Z80 by Bradford J. Rodriguez,
''Z80 Forth (I do not have the authors name),
''eForth by Ken Chen
''ANS Forth draft (X3J14 dpANS-6 June 30, 1993)

'' 2010 2 15:
''This version is the first to include the full ANS core word set.
''The optimizer is not implemented at this point. Dictionary address
''adjustment done in Spin startup code.
''
'' 2010 2 17:
''This version is the first to include the code optimizer.
''Optimizer replaces native forth word addreses with there virtual CPU opcode.
''Optimization is done before the virtual CPU is started. The forth compiler
''does not preform code optimization (next thing to fix). Forth word "EXECUTE"
''needs to not be tagged as a native opcode. This forth word must be called with its
''calling address placed on the stack, side effect of optimizer.

'' 2010 2 18:
''This version is the first to include the compile time optimization.
''The optimization is done in the forth word INTERPERT when in compile mode.
''I should move this to the COMPILE, foth word, This will allow all execution 
''tokens to be optimized as they are added to the dictionary. 
''Added the forth word MARKER to allow the unloading
''of dictionary words (replaces FORGET). Added the foth word COGD@ and COGD!
''to allow read/write access to any COG memory location or register (ie. INA, OUTA...).
''
'' 2010 2 20:
''Added a number of double precision forth words; DXOR, DOR, DAND, DLSHIFT, DRSHIFT, DARSHIFT
''as native virtual CPU opcodes to make working with 32 values simpiler and faster. Also
''added 2CONSTANT, 2VARIABLE, D2/ and D2* as native foth words.
''Started to add support for double literals, constants and variables. Still have not modified the
''forth parser to accept double literals from input (double numbers end in a period).
''When completed the entire DOUBLE word set should be implemented.
''General dictionary cleanup and debuging.
''The forth cog now uses the global memory it was loaded from for working memory.
''This means the forth cog can not be stopped and then restarted from spin. I do not
''see this as a limitation. MV4th16 was modified to no longer use the USER
''memory region leaving it completly to the programer to use. Still need to add a forth
''word to indicate the size of the USER memory area.  
''
''2010 2 21:
''Added code to locate the end of the spin code used in this object. Got the idea from user BradC on Parallax forum.
''MV4th16 code is made up of COG asm code followed by a "DAT" section that contains the forth dictionary. 
''Once the forth object is started the spin code is just wasted memory space (@ 700 bytes).
''The memory ocupied by the spin code for the forth object is located after the end of the hard coded forth dictionary portion of memory.
''if we know where the end of the spin code is in memory we can reuse the memory to extend the forth dictionary.
''Completed support for double literals by adding a virtual CPU opcode NFW_DLIT. This opcode allows a double to be loaded from
''program memory to the data stack. Created a new forth word "?DOUBLE" that checks if a literal 
''string ends in a "." (period) charateter. I also modified ">NUMBER" to return a value of one if the 
''converted string is a double cell literal and still return -1 for a valid single cell literal.  The toral
''effect of the changes allow a double literal to be entered by ending a standard numeric literal string with a period.
''The forth words "2CONSTANT" and "2LITERAL" were modified to use the new "NFW_DLIT" CPU opcode.
''The forth word ":NOMANE" was added to suport headless forth words. Curently the ";" forth word always calls "REVEAL"
''as it exits compile mode. Since a headless fort word has no header "REVEAL" will actualy be applied to the last 
''standard (has a header) forth word pointed to by "LATEST". This should not realy be a problem but it does not seam
''like it should be that way so a fix is required.
''
''2010 3 2:
''Need to add support for forth word "DOES>". COG memory was all in use so I removed the native opcode "NFW_EXIT_LIT".
''This native code was used by the forth words "USER" and "CONSTANT". These words where modified to use the
''native opcode "NFW_LIT" and append an "NFW_EXIT".
''The System command variable is now passed using the PAR register. This means the two word variables must be long
''aligned. This change allowed for one extra byte of COG memory.
''To support the forth "DOES>" word I added a native opcode called "NFW_DOES_BRANCH". I also made the "NOP" forth word
''a native opcode. This allows the "CREATE" word to reserve the first two cells after the header allowing for the use
''of the "DOES>" word in compiler mode.
''
''2010 3 5:
''Starting to add block support. Initialy will support extra 32k on prototype board
''eeprom. Would like to use 2MB SPI flash later.
''Started to add missing words to get to X3J14 dpANS-6 (June 30, 1993) ANS forth standard.
''Added native opcodes for forth words; 2>R, 2DROP, 0<>, and 2RDROP.
''Block support is implemented using a single fixed block buffer for now. Since there
''is 32K of free space in the eeprom we have valid block numbers of 0..31.
''There is no support for LOAD from a block still must modify the parser to read from
''current block buffer. Will also need to add forth words; \, REFILL, LOAD, LIST.
''  
''2010 3 8:
''Added the forth word EXIT and renamed U* and U/MOD to UM* and UM/MOD per ANS spec.
''Added suport for .( forth word to display a comment while the compiling. Also
''added forth words TRUE, and FALSE.
''  
''2010 3 9:
''Added forth defining word VALUE and complementry forth word TO.
''Got rid of the seperate execution stack used by the virtual CPU.
''The virtual CPU now uses the return stack. Moved all the return
''stack based forth words to native opcodes these include;
''R@ R> >R 2R@ 2>R 2R>. With all return stack forth words as
''native opcodes we can get rid of execution stack.
''Moved forth words S0 and R0 from native opcodes to forth words
''to free up space in cog memory for the return stack functions.
''Added support for ?DO forth word this needs more testing but does
''seam to work. Added forth word ERASE as I work to meet the full 93
''ANS core and extened core word sets (if I don't run out of memory first).
''
''2010 3 11:
'' Added the forth words; PARSE and SOURCE-ID.
''
''2010 4 8:
''Added the forth words SAVE-INPUT, RESTORE-INPUT, THRU, LOAD, LIST.
''BLK was being set at wrong time. the \ comment does not work as it will
''skip all of remaining block. Fixed DLSHIFT, DRSHIFT and DARSHIFT to
''return a double prec. value, only low word was returned.
''
''2010 4 10:
''LOAD was change to save the input state to the return stack
''instead of using SAVE-INPUT / RESTORE_INPUT.
''Rename forth work ?KEY to KEY?.
''Added the forth word pair DEFER and IS. DEFER creates defered words
''that can be assigned to execute an execution token using the forth word IS.  
''
''2010 4 19: Added forth words to support the case statement; CASE,
''OF, OF-RANGE, ENDOF, and ENDCASE. Added the forth word OTHERWISE as
''it is used in many forth compilers but it is a do nothing word in MV4th16. 
''
''2010 4 22: Changed forth words 1+ 2+ to save 4 cog opcodes. Bug found in
''forth word CMOVE> reverse copy addresses where off by one, added subtract one
''from calculated source address and destination address to correct the problem.
''Added the forth word [COMPILE] to round out the ANS XCORE word set.
''Added the ability to prefix numbers with a "$" to indicate a hex number.
''
''2010 4 27: Corrected a problemm with the forth word LEAVE. Changed the name
''of the forth word (unloop) to UNLOOP.
''
''2010 5 3:
''Added the non ANS forth word USERSIZE to return the number of bytes in the USER memory area.
'' 
''2010 5 6:
''Added better error reporting to forth word ' (tick). ' now correctly reports an error
''when the name is not found because the forth interprutter reached the end of the input buffer.
''Also if name is not in the dictionary name is diplayed followed by a question mark instead
''of just displaying a ' character as it did before.
''The forth word ALLOT now reports an "out of memory?" error if the DP address + 128 > SP.
''The data stack and dictionary share a common memory region with the dictionary pointer, DP
''increasing as memory is allocated for the dictionary and the stack pointer SP decreasing it address as values
''are save to the data stack. If DP >= SP we have run out of memory. The 128 is a pad to attempt
''to detect the out of memory before we crash.
''Added the non ANS forth word RDROP as a native virtual cpu opcode to make it
''simpler to use return stack.   
'' 
''2010 5 6:
''Added the non ANS forth word :ABORT this is a factor of ABORT. The intent of this word is
''to cleanup after ABORT is called while compiling a definition into the forth dictionary.
'':ABORT will remove the last definition in the dictionary if the word is still set
''to be hidden (high bit of name length byte = 1). This will not work for headless
''definitions created with the forth word :NONAME. I will need a better way to track
''headless definitions but for now works most of the time.
''Added the non ANS forth word EMPTY-KEYS to clear the input buffer.
''Added the non ANS forth word EMIT? test if console output will block or not.
''Added the non ANS forth word BACKSPACES to emit X number of backspace characters. 
''
''2010 5 14:
''Changed the name of foth words COGD@ abd COGD! to COG@ and COG!.
''Added the forth words COG-NEW, COG_INIT and COG-STOP these words are implemented as system callbacks.
'' 
''2010 5 19:
''Fixed the forth word .S to report when data stack has underflowed instead of dumbing entire memory image.
''Added the non ANS forth word THRU. this works exactly like THRU but displays the block numbers as they loaded.
''
''FIRST RELEASE ver 0.90
''
''2010 6 5:
''Added the non ANS forth words Q@ and Q! to read and write a double precision number to or from memory.
''The supplied memory address must be alligned on a quad byte address (0, 4, 8, 12....). These function allow MV4th16 to
''read or write a double value as an atomic operation.
'' 
''2010 6 15:
''Added the non ANS forth words LOCK-NEW and LOCK-RET to allocate and release propeller locks.
''These words are implemented as system callbacks to save space in the COG. Since the locks are not
''normally allocated and released frequently no harm in using a system callback. Also added the non ANS
''forth words LOCK-SET and LOCK-CLR as virtual CPU opcodes so that locks can be tested and set with the
''least posible delay.
''
''SECOND RELEASE ver 0.91
''
''2011 7 10
''Fixed forth word ?NUMBER that I broke when adding support for the forth word ?DNUMBER. The forth word ?NUMBER is not used by
''the current forth implementation and because it is a support function not part of ANS standard forth we could remove ?NUMBER and just use
''the ?DNUMBER forth word in its place (also just a support funtion).
''
''
''2012 12 23
''Made EMIT and EMIT? defered forth words to allow for I/O redirection. The old EMIT abd EMIT? words where renamed
''to (com1out) and (com1out?) and are assigned to EMIT and EMIT? after a reset.
''Added new forth word "NFA" to return the name field address of a forth word matching the supplied name string.
''This word is needed by "TO" and "IS" as the forth word "FIND" returns a matching forth word's code address not
''the forth word's name address. "FIND" was updated to use the forth word "NFA" as a factor.
''The forth word "VALUE" was updated to set a flag so that the forth word "TO" can tell if the specified word was
''created with the forth word "VALUE". 
''
''2012 12 23
''Modified forth word "DEFER" to set opcode flags to indicate a defered forth word. Modified forth word "IS" to
''only modify forth words that have the defered opcode flag set. Added the forth word "DEFER?" to test a
''forth word for the defer opcode flag. Commented out the forth word "DUMP" to save space, "DUMP" can be found
''in the "TOOLS.BLK" file included with MV4th16.
''Commented out the forth word "blk-wrt-cnt" as it has no real use for EEPROM devices.
''
''2012 12 24
''Forth word "?NUMBER" was redone to call "?DNUMBER" to save space.
''Forth word "EMPTY-KEYS" had its name field length set wrong (4 instead of 10) and was corrected.
''
''2012 01 26
''Forth word "?HEX" was removed and replaced with "?CHAR" that tests if the first character of a string
''is a given character. "?CHAR" is used by the new forth word "?BASE" to set the system BASE variable to one
''of three possible values based on the first character of a numeric string ($ = 16, # = 10, % = 2).
''
''2012 01 31
''Added missing Forth word "RP@" to read the current return stack address. Native forth word "I" was mapped
''to native forth word "R@". To improve forth tool word "SEE" forth word "I" has its own native word.
''
''2012 02 03
''Rewrote the native opcode "OP_XLOOP" to use 6 less longs then before. Need space to support multi-tasking
''in future release. Started to remove DAT section Param_xxxx system variables. Removed Param_UserBufferAdr
''replaced with call to opcode OP_XUSER.
''Removed the forth word USERSIZE as it realy serves no purpose, also removed local variable "Param_UserBufferSize".
''
''2012 02 05
''Added the forth words "0" and "1" to allow the character 0 and 1 to have there own native forth words.
''This saves 30 longs of memory from base forth dictionary and makes the symbols 0 and 1 compile and execute much faster.
''
''2012 02 08
''Created the forth word "LIT," to replace the common forth thread "NFW_LIT NFW_LIT XT, ," to save memory.
''The cpu opcode "OP_STOREPLUS" was modified to share code with "OP_STORE", this saved 2 COG longs.
''
''2012 02 13
''Added the custom forth word "OPFLAG?" as a factor to check for any given forth word flag.
''Updated the forth word "(init)" to save memory.
''
''2012 02 15
''Changed code for opcode "NFW_XLOOP" to speed up DO...LOOP code. This change required
''7 additional longs of COG memory. 
''Updated the forth word "TYPE" to save memory space (uses 6 longs less then before). 
''Updated the forth word "S=" to save memory space and increase speed. Improved compile times by @ 10% 
''
''2012 02 15
''Updated the forth word "SKIP", now uses 14 words less memory and is faster.
''Updated the forth word "SCAN", now uses 10 words less memory and is faster.
''
''2012 03 08
''Added forth terminal control words; AT-XY to set terminal cursor, PAGE to clear the terminal screen,
''XY to fetch the terminal cursor position, MAX-XY to read the terminal display bounds, CLR-EOL to
''clear the current row on the terminal display from the cursor, and CURSOR to show/hide
''the terminal cursor. These words are all defered and do not have any default values. 
''
''2012 03 13
''Moved the memory commands FILL CMOVE and CMOVE> to the system I/O functions to free up COG memory.
''Renamed the default serial port funtions to COM1-IN COM1-IN? COM1-OUT COM1-OUT? COM1-CLEAR.
''EMIT EMIT? KEY KEY? and EMPTY-KEYS are now defered words, at startup they are assigned to COM1. 
''
''2012 04 08
''The forth word "ACCEPT" was modified to not emit a backspace character when the input buffer
''is empty, and not emit characters after the buffer is full.
''Added the function READ_RTC and WRITE_RTC to access the RTC clock chip on the spinnet board.
''These forth words should be made more generic by changing them to a set of generic
''I2C functions such as; "I2C-START", "I2C-STOP", "I2C-READ", and "I2C-WRITE".
''
''2012 04 13
''Removed the forth word "THRU." as it is a debuging tool I placed it into the tools file.
''RE above: now have I2C functions; "I2C-START", "I2C-STOP", "I2C-READ", "I2C-WRITE",
''"I2C-FAST", and "I2C-SLOW". Last two words control the I2C speed ( 1 Mbs or 400 Kbs).
''
''2012 05 21
''Removed the forth word "?NUMBER" as it not required in the core dictionary.
''Use the word ?DNUMBER instead.
''
''2012 05 29
''Forth words PICK and ROLL where modified to use a zero based index instead of a one based
''index. The forth word 2ROT uses ROLL and was modified to use a zero based index.
''
CON
  c_MaxLocalCode    = $1F0           'largest value of a cog local opcode.
  '
  c_ReservedWords   = 9000           'number of words of memory to reserve for the forth dictionary and data stack.
  '
  c_FirstFortAddress = 8             'Start of Leave and Return stack, reserve first 2 longs for normal spin clkfreq                   
  '                     
 'Following Opcode constants are used to directly refrence virtual CPU opcodes.
 'If you modify the cog code these may need to be adjusted.
 'These functions all modify the IP directly and can not be called
 'from a different forth code thread. ALSO the forth word SEE in the seperate Tools.forth file
 'will need to be modified if these constants are changed.
 '
  NFW_NOP       = $07            'Must be set to cog start offset of OP_NOP
  NFW_EXIT_ADR  = $08            'Must be set to cog start offset of OP_EXIT_ADR
  NFW_EXIT      = $0B            'Must be set to cog start offset of OP_EXIT
  NFW_DLIT      = $0E            'Must be set to cog start offset of OP_DLITERAL
  NFW_ALIT      = $12            'Must be set to cog start offset of OP_ALITERAL
  NFW_LIT       = $13            'Must be set to cog start offset of OP_LITERAL
  NFW_QBRANCH   = $16            'Must be set to cog start offset of OP_QBRANCH
  NFW_BRANCH    = $1C            'Must be set to cog start offset of OP_BRANCH
  NFW_XLOOP     = $1E            'Must be set to cog start offset of OP_XLOOP
  NFW_XPLUSLOOP = $28            'Must be set to cog start offset of OP_XPLUSLOOP
  NFW_XSQUOTE   = $38            'Must be set to cog start offset of OP_XSQUOTE
  NFW_DOES_BRANCH = $40          'Must be set to cog start offset of OP_DOES_BRANCH
  '
  'System command index for basic forth functions implemented by another cog
  cmd_Done        =  0    '                           Indicates that the last command is done and a new command can be requested.
  
  cmd_COM_OUT     =  1    ' ( c -- )                  Write char to console, c = 8 bit char to send to console.
  
  cmd_COM_IN      =  2    ' (  -- c )                 Read char from console, c = 8 bit char read from console.
  
  cmd_COM_CAN_IN  =  3    ' (  -- f )                 Test if console has data to read, f = 0 no char ready
'                                                     from console, f= -1 char abailable from console.

  cmd_COM_CAN_OUT =  4    ' (  -- f )                 Test if console can write data, f = 0 no char can be set to
'                                                     the console, f= -1 char can be written to console.

  cmd_COM_CLEAR =  5    ' ( -- )                    Request the console input buffer be cleared
                                              
  cmd_READBLK   =  6    ' ( u adr -- f )            Read block, u = mass storage block number, adr = buffer
'                                                   memory address, f = 0 read failed, f = -1 block read.
 
  cmd_WRITEBLK  =  7    ' ( u adr -- f )            Write block, u = mass storage block number, adr = buffer
'                                                   memory address, f = 0 write failed, f = -1 block written.

  cmd_BLK_COUNT =  8    ' ( -- u )                  Returns number blocks avaulable, u = mass storage block count.
  
  cmd_COG_NEW   =  9    ' ( q-adr1 u1 -- u2 )       Start a new COG q-adr1 points to a long aligned address containg
'                                                   the cog code to execute. u1 (upper 14 bits) is passes via the COG PAR register.
'                                                   The function returns the COG number started in u2 or u2 = -1 if no cog available.                            

  cmd_COG_INIT  = 10    ' ( q-adr1 u1 u2 --  )      Start the COG index number u2, if the COG is already running stop the COG and restart
'                                                   the COG with the new code. q-adr1 points to a long aligned address containg
'                                                   the cog code to execute. u1 (upper 14 bits) is passes via the COG PAR register.
     
  cmd_COG_STOP  = 11    ' ( u --  )                 Stop the COG index number u if it is running.

  cmd_LOCK_NEW  = 12    ' ( -- u )                  Allocate a lock u is the lock index or -1 if no locks avaiable            
  
  cmd_LOCK_RET  = 13    ' ( u -- )                  Release a lock, u is the lock index to release/return

  cmd_FILL      = 14    ' ( c-adr u c -- )         'Fill global memory from address c-adr for u consecutive bytes with character c.
  
  cmd_CMOVE     = 15    ' ( c-adr1 c-adr2 u -- )   'Copy u characters from c-adr1 to c-adr2 
  
  cmd_CMOVE_REV = 16    ' ( c-adr1 c-adr2 u -- )   'Copy u characters from c-adr1 to c-adr2 in reverse order (bottom up) 
  
  cmd_COM_BAUD  = 17    ' ( d -- )                  Set the baud rate for the default console device

  cmd_I2C_RESET = 18    ' ( -- )                    Reset the I2C interface

  cmd_I2C_READ  = 19    ' ( c-Adr uCount -- )       Read a specified number of bytes from the I2C interface to global memory.

  cmd_I2C_WRITE = 20    ' ( c-adr uCount -- fAck )  Write specified number of bytes from global memory to the I2C interface.
                        '                           if fAck is true the device on the I2C interface responded.
                        '
  cmd_I2C_START = 21    ' ( -- )                    Send a start sequence on the I2C interface.

  cmd_I2C_STOP  = 22    ' ( -- )                    Send a stop sequence on the I2C interface.

  cmd_I2C_FAST  = 23    ' ( -- )                    Set the I2C interface to run at @ 1 Mbs.

  cmd_I2C_SLOW  = 24    ' ( -- )                    Set the I2C interface to run at @ 400 Kbs.

  cmd_REBOOT    = 25    ' ( -- )                    Reset forth system.

  '
  'Forth definition flags
  of_NONE       =  0
  of_IMMED      =  1
  of_NATIVE     =  2
  of_VALUE      =  4
  of_DEFER      =  8

  'Size of various forth data areas in bytes
  size_ReturnStack              = 1024      'Return stack and Leave stack share same memory region
  size_OutBuffer                =   80      'Temporary Output Buffer is used to convert numbers to string
  size_PadBuffer                =  128      'Scrach pad work area not used by Forth itesef
  size_TibBuffer                =  160      'Terminal input buffer, size determins max text length for forth input sting
  size_DiskBuffer               = 1024      'Block device I/O buffer size if bytes
  '
  'Memory offsets of various forth data areas in bytes
  ofs_ReturnStack              =  0
  '      
  'USER area variable offsets for base forth system,
  'These need to be in. user area to support multi-tasking.
  uof_ToIn                =    0
  uof_Base                =    2
  uof_State               =    4
  uof_DictionaryPointer   =    6
  uof_SourceID            =    8
  uof_Source              =    10             'two words of storage
  uof_Latest              =    14
  uof_HoldPointer         =    16
  uof_LeavePointer        =    18
  uof_BlockNumber         =    20
  uof_ScreenBlockNumber   =    22
  uof_TIB                 =    24                        'TIB buffer is 160 bytes
  uof_OUT                 =    uof_TIB + size_TibBuffer  'OUT buffer is 80 bytes long.
  uof_PAD                 =    uof_OUT + size_OutBuffer  'PAD buffer is 128 bytes, PAD must be directly after OUT buffer
  uof_SIZE_USER_VARS      =    uof_PAD + size_PadBuffer               

  size_FirstDataSpace = c_FirstFortAddress + size_DiskBuffer + size_ReturnStack + uof_SIZE_USER_VARS            
  
PUB Start(SystemCommandAdr, DataStackBaseAdr) | ObjectOffset, BaseMemoryAdr
  Prompt(SystemCommandAdr)
  'For now all Forth working storage will use COG code memory after COG loads
  'Data stack starts at end of data memory and grows to lower memory addresses.
  'Save first 2 longs so COG assemblies will find system clock settings,
  'long[0] = system clock frequency and byte[5] = clock mode settings. 
  BaseMemoryAdr:= c_FirstFortAddress
  'Initialize the param/data stack to supplied address.
  'The supplied address is the last address of the stack region.
  'The data stack address is reduced each time a value is added to the stack.
  DatStkPtr := DataStackBaseAdr
  Param_DataStackBase:= DatStkPtr
  'Setup virtual CPU return stack address. Stack grows towards lower memory (PUSH is; adr = adr - 2, write adr value)
  'The supplied address is the last address of the return/leave stack region.
  'The return stack address is reduced each time a value is added to the stack.
  RetStkPtr:= BaseMemoryAdr + ofs_ReturnStack + size_ReturnStack
  Param_ReturnStackBase:= RetStkPtr
  'Initialize block buffer address
  Param_BlockBufferBase:= RetStkPtr
  'Initialize the leave stack address.
  'The supplied address is the first address of the return/leave stack region.
  'The leave stack address is increased each time a value is added to the stack.
  Param_LeaveStackBase:= BaseMemoryAdr + ofs_ReturnStack
  'User variable allocation address
  UserVarsAdr:= Param_BlockBufferBase + 1024
  'Calculate the global memory address offset required to convert
  'object based addresses to global memory based addresses.
  'On first cog start the IP variable is set to object address of "COLD" forth word.
  ObjectOffset:= @COD_COLD - IP                         'Calculate global memory object offset, IP = object offset of forth word "COLD"                
  IP := @COD_COLD                                       'Set IP variable to global address, so on next start we do not correct addresses.
  'If ObjectOffset is not zero correct the memory address refrences in the base forth words.
  if ObjectOffset   
    CorrectAddresses(ObjectOffset)
    OptimizeDictionary
  'Init the system command interface variables
  WORD[SystemCommandAdr]:= 0                             'Indicate there is no command to process
  'Save current system clock frequency
  'Start our cog if we can
  coginit(0, @Entry, SystemCommandAdr)

PRI OutChar(SystemCommandAdr, Char) | Resp, Stack
  Stack := Char
  word[SystemCommandAdr + 2] := @Stack
  word[SystemCommandAdr] := 1
  repeat
     Resp := word[SystemCommandAdr]
  until (Resp == 0)
        
PRI Prompt(SystemCommandAdr) | Idx, Str1
      Str1 := string("Starting MV4th16...",13,10)
      repeat Idx FROM 1 TO 21
        OutChar(SystemCommandAdr, byte[Str1])
        Str1 += 1
        
PRI CorrectAddresses(ObjectOffset) | CurOP, LinkedOP, aCodeWord, IsNative, PriorOpStart, PrioLinkAdr, x
'Correct the code addresses to compensate for the objects placement in memory.
'When the Forth words are complied into the object all addresses are relative
'to the start of the object. Forth requires the actual memory addresses so
'we must add an offset to all address values in the forth code threads.
    LinkedOP:= @WRD_COLD
    PriorOpStart:= @UserCodeEnd
    repeat
      CurOp:= LinkedOP
      'First read link field address to prior forth word
      PrioLinkAdr:= LinkedOP
      LinkedOP:= WORD[CurOp - 4]
      IsNative:= WORD[CurOp - 2] & of_NATIVE
      if LinkedOp
        'Link field not 0 so correct it
        LinkedOp += ObjectOffset
        WORD[CurOp - 4]:= LinkedOp
      'Skip over the Name field to get to start of code.
      'First byte of name field is its length  
      x:= BYTE[CurOp]
      'add one to string length for length field byte itself.
      x += 1
      'if string length is odd, add one to keep word alignment
      if (x & 1)
        x += 1
      CurOP += x
      'We are pointing to first word in code thread
      if (IsNative == 0)
        repeat
          aCodeWord:= WORD[CurOp]
          if (aCodeWord <> NFW_LIT)        
            if (aCodeWord <> NFW_DLIT)        
              if (aCodeWord <> NFW_XSQUOTE)        
                if (aCodeWord > c_MaxLocalCode)
                  aCodeWord += ObjectOffset
                  WORD[CurOp]:= aCodeWord
              else
                'We have an embeded string, skip over it.
                'Point to start of string
                'Read length byte
                x:= BYTE[CurOp + 2]
                x += 1
                'if string length is odd, add one to keep word alignment
                if (x & 1)
                  x += 1
                CurOP += x
            else    
             'we have a double literal value, skip over it    
              CurOp += 4
          else
           'we have a literal value, skip over it    
            CurOp += 2
          'Point to start of next opcode  
          CurOp += 2
        until (aCodeWord == NFW_EXIT) or (CurOp => PriorOpStart)
        PriorOpStart:= PrioLinkAdr
    until (LinkedOP == 0)

PRI BuildNativeTable : Count | TableAdr, CurOP, LinkedOP, aCodeWord, IsNative, x
    ''Create a lookup table of native word entry address and there actual native opcode.
      Count:= 0
      TableAdr:= @UserCodeEnd   'Use memory region after last forth word for table
      LinkedOP:= @WRD_COLD
      repeat
        CurOp:= LinkedOP
        'First read link field address to prior forth word
        LinkedOP:= WORD[CurOp - 4]
        IsNative:= WORD[CurOp - 2] & of_NATIVE
        if IsNative
          'Calculate code start address, CurOP = address of name field.
          x:= BYTE[CurOp]
          x += 1
          'if string length is odd, add one to keep word alignment
          if (x & 1)
            x += 1
          CurOP += x
          'CurOP = forth word code start address
          'First code word of a native forth word contains the opcode
          aCodeWord:= WORD[CurOP]
          'Add the code address and native opcode to the table
          WORD[TableAdr]:= CurOp
          TableAdr += 2
          WORD[TableAdr]:= aCodeWord
          TableAdr += 2
          Count += 1
      until (LinkedOP == 0)

PRI ConvertAddress(Address, Count): NativeCode | First, Last, Mid, MidVal, MidAdr, x
      'Convert a forth code address to a native opcode if posible.
      'Return either the native opcode for the address or if no
      'native code at address returns the same address.
      'Address table is sorted by code address so do a binary search
      NativeCode:= Address
      First:= 0
      Last:=  Count - 1
      repeat
        Mid:= First + ((Last - First) >> 1)
        MidAdr:= @UserCodeEnd + (Mid << 2)              'Use memory region after last forth word for translation table
        MidVal:= WORD[MidAdr]
        if (MidVal == Address)
          'Found the address in code table
          NativeCode:= WORD[MidAdr + 2]
          Last:= Mid
          First:= Mid
        else
          'Did not match try again
          if (MidVal > Address)
            First:= Mid
          else
            Last:= Mid
      until (Last - First) =< 1
      if (NativeCode == Address)
        'Did not find see if First or Last matches
        MidAdr:= @UserCodeEnd + (Last << 2)
        MidVal:= WORD[MidAdr]
        if (MidVal == Address)
          NativeCode:= WORD[MidAdr + 2]
        else
          MidAdr:= @UserCodeEnd + (First << 2)
          MidVal:= WORD[MidAdr]
          if (MidVal == Address)
            NativeCode:= WORD[MidAdr + 2]
      
PRI OptimizeDictionary : ChangeCount | NativeCount, CurOP, LinkedOP, aCodeWord, IsNative, PriorOpStart, PrioLinkAdr, x
     'Scan entire dictionary and replace all refrences to forth code threads
     'that have native virtual CPU opcodes with the actual opcode.
      ChangeCount:= 0
      PriorOpStart:= @UserCodeEnd
      NativeCount:= BuildNativeTable
      LinkedOP:= @WRD_COLD
      repeat
        CurOp:= LinkedOP
        'First read link field address to prior forth word
        PrioLinkAdr:= LinkedOP
        LinkedOP:= WORD[CurOp - 4]
        IsNative:= WORD[CurOp - 2] & of_NATIVE
        if (IsNative == 0)
          'Skip over the Name field to get to start of code.
          'First byte of name field is its length  
          x:= BYTE[CurOp]
          'add one to string length for length field byte itself.
          x += 1
          'if string length is odd, add one to keep word alignment
          if (x & 1)
            x += 1
          CurOP += x
          'We are pointing to first word in code thread
          repeat
            aCodeWord:= WORD[CurOp]
            case aCodeWord
              NFW_LIT, NFW_QBRANCH, NFW_BRANCH, NFW_XLOOP, NFW_XPLUSLOOP:
               'we have a literal value or branch address, skip over it
                CurOp += 2
              NFW_DLIT:
                CurOp += 4
              NFW_XSQUOTE:
                'We have an embeded string, skip over it.
                'Point to start of string
                'Read length byte
                x:= BYTE[CurOp + 2]
                x += 1
                'if string length is odd, add one to keep word alignment
                if (x & 1)
                  x += 1
                CurOP += x
              OTHER:
                if (aCodeWord > c_MaxLocalCode)
                  x:= ConvertAddress(aCodeWord, NativeCount)
                  if (aCodeWord <> x)
                    ChangeCount += 1
                    WORD[CurOp]:= x
           'Point to start of next opcode   
            CurOp += 2
          until (aCodeWord == NFW_EXIT) or (CurOp => PriorOpStart)
          PriorOpStart:= PrioLinkAdr
      until (LinkedOP == 0)
      
{
PRI  EndSpinAddress
''This is a dummy funtion that is used to find the global memory address
''of the end of the spin code used to initialize the MV4th16 object.
    Return string(1)
}          
DAT
'
                        org     0
'Execute the next opcode pointed to by IP
Entry                   rdword  T1, IP                  'Read next forth word and save in T1
                        add     IP, #2                  'Point IP to next code address cell
                        cmp     T1, #c_MaxLocalCode       wc
              if_c      jmp     T1                      'If code is native opcode execute it
                        'We want to execute a code thread
                        'Push current IP to return stack.
                        sub     RetStkPtr, #2           'Set return stack address for push
                        wrword  IP, RetStkPtr           'Save current IP adddress to return stack
                        mov     IP, T1                  'Set IP to new execution address
OP_NOP                  jmp     #Entry
'
'Exit current code thread leaving the address of the next instruction
'opcode (1 cell ahead) and return to prior code thread..
'This is used for CREATE and VRIABLE.
OP_EXIT_ADR             mov     TOS, IP         'IP always points to next execution address
DO_EXIT_PUSH            sub     DatStkPtr, #2   'Point to next top of data stack address
                        wrword  TOS, DatStkPtr  'Save TOS to data stack
                        'Fall into normal EXIT
'Exit current code thread and return to prior code thread.
'IP is poped from the return stack. The cog address of OP_EXIT must be
'set in the constant NFW_EXIT. If you modifiy the code make sure you set
'the constant. NFW_EXIT to the new cog address of OP_EXIT opcode.  
OP_EXIT                 rdword  IP, RetStkPtr
OP_RDROP                add     RetStkPtr, #2                                         
                        jmp     #Entry
'
'Copy the next two execution words to the top of the data stack
'then advance instruction pointer to next word.
OP_DLITERAL             rdword  TOS, IP
                        add     IP, #2
                        sub     DatStkPtr, #2           'Point to next top of data stack
                        wrword  TOS, DatStkPtr          'Write correct value to top of data stack.
'                        
'Copy the next execution word to the top of the data stack
'then advance instruction pointer to next word.
'This is not a forth word. Instead it is used by spin code
'that adjusts the base words addresses to convert the addresses
'from object relative to global addresses.
'This opcode should only be used by the base word set.
'Normaly a literal value is not adjusted during address conversion.
'If the literal is an object based address the "ALITERAL" opcode will
'allow the value of the literal to be adjusted.
OP_ALITERAL             nop
'Copy the next execution word to the top of the data stack
'then advance instruction pointer to next word.
OP_LITERAL              rdword  TOS, IP
                        add     IP, #2
                        jmp     #DO_PUSH_TOS            'Push value in TOS variable (16 bits) and then execute next opcode.
'
'
'Jump to the address at the next code word if TOS = 0
OP_QBRANCH              mov     T1, TOS                 'Save current TOS value.
                        add     DatStkPtr, #2           'Point to second data stack cell.
                        rdword  TOS, DatStkPtr          'Read the value of the second data stack cell.
                        tjz     T1, #OP_BRANCH
                        'Skip the brach address
                        add     IP, #2
                        jmp     #Entry
'Jump to the address at the next code word
OP_BRANCH               rdword  IP, IP
                        jmp     #Entry
'
'Loop back to last DO loop if limit is not crossed when 1 is added to loop count
OP_XLOOP                rdword  T1, RetStkPtr          'Read loop index value.
                        'Set T2 = address of loop limit value 
                        mov     T2, RetStkPtr
                        add     T2, #2                 
                        rdword  T2, T2                 'Read loop limit value
                        'increment the loop index, then test if we reached limit.
                        add     T1, #1
                        and     T1, CELLMASK
                        wrword  T1, RetStkPtr          'Write new loop index value.
                        cmp     T2, T1          wz
            if_nz       jmp     #OP_BRANCH             'If not done take branch
                        'All done so clean-up return stack and skip branch
                        jmp     #OP_XLOOP_1
'
'Loop back to last DO if boundry of Limit is not crossed when
'Loop Step is added to the current loop count. The step size is
'the top value on the data stack. The current Limit and Current loop
'values are on top of the return stack.
OP_XPLUSLOOP            add     DatStkPtr, #2           'Pop top value from data stack, loop step (Step saved in TOS var).
                        mov     T2, TOS                 'Save Loop step in T2.
                        'Get loop count and limit from return stack.
                        call    #DO_GET_2R              'Get the current loop index on the return stack, TOS = loop index  T1 = loop limit. 
                        'We have our three values to calc loop state, Update loop values.
                        mov     T3, T1
                        sub     T3, TOS                 'Distance from loop limit to loop index in T3
                        add     TOS, T2                 'Calc next loop index
                        sub     T1, TOS           wz    'Distance from loop limit to next loop index in T1
                        'See if the sign bit changed
                        xor     T1, T3
                        test    T1, CellSignMask  wc    
                        'if sign bit changed on distances or index = limit, DO has completed
          if_z_or_c     jmp     #OP_XLOOP_1                                
                        'Update current count and take branch                        
                        wrword  TOS, RetStkPtr
                        rdword  TOS, DatStkPtr          'Write correct value to top of data stack variable. 
                        jmp     #OP_BRANCH
                        'Loop completed pop top 2 return stack values and skip branch
OP_XLOOP_1              add     RetStkPtr, #4           'Pop top two values from the return stack
                        add     IP, #2                  'Skip over loop brach back address
                        jmp     #DO_LOAD_TOS            'Update the TOS variable to reflect the change in the stack pointer.
'
'Push the address and count of the counted string following this opcode to the data stack.
'Adjust the virtual CPU instruction pointer to point to the opcode following the string.
'P (S")     -- c-addr u   run-time code for S"
OP_XSQUOTE              mov     T1, IP                  'T1 = start address of the string
                        rdbyte  TOS, T1                 'read string length byte
                        add     IP, TOS                 'IP address of last string character
                        add     IP, #1                  'Add one for length byte
                        test    IP, #1          wz      'see if IP is on even byte boundry
              if_nz     add     IP, #1
                        'we need to save address in T1 and count in TOS to stack
                        add     T1, #1
                        jmp     #DO_PUSH_T1_TOS                        
'
'Push the IP address plus 2 to the data stack, then branch to the address
'at the next cell.
OP_DOES_BRANCH          mov     TOS, IP
                        add     TOS, #2
                        rdword  IP, IP
                        jmp     #DO_PUSH_TOS 
                        
'Push 0 at the top of the data stack.
OP_ZERO                 mov     TOS, #0
                        jmp     #DO_PUSH_TOS            'Push value 0 in TOS variable (16 bits) and then execute next opcode.
'                        
'Push 1 at the top of the data stack.
OP_ONE                  mov     TOS, #1
                        jmp     #DO_PUSH_TOS            'Push value 1 in TOS variable (16 bits) and then execute next opcode.
'                        
'Execute the code thread whose address is at the top of the data stack
'and return to calling code thread when done.
OP_EXECUTE              mov     IP, TOS
                        jmp     #OP_DROP
'
'                        
'HELPER ROUTINES
'
'Save the TOS and T1 variables to top of data stack and do next OpCode
DO_PUSH_T1_TOS          sub     DatStkPtr, #2           'Point to next top of data stack
'Save the TOS and T1 variables to top of data stack and do next OpCode
DO_SAVE_T1_TOS          wrword  T1, DatStkPtr
'Push the value in the variable TOS and perform next opcode.
DO_PUSH_TOS             sub     DatStkPtr, #2           'Point to next top of data stack
'
'Save the value in the variable TOS to the current data stack position and perform next opcode.
DO_SAVE_TOS             wrword  TOS, DatStkPtr          'Write correct value to top of data stack.
                        and     TOS, CellMask
                        jmp     #Entry                 'Perform next opcode.

'Load the value in the variable TOS with the current data stack position value, then perform next opcode.
DO_LOAD_TOS             rdword  TOS, DatStkPtr          'Write correct value to top of data stack.
                        jmp     #Entry                 'Perform next opcode.


'Read and pop top 3 values into TOS, T1, T2,
'stack points to forth stack cell.
DO_POP3                 call    #DO_POP_LOAD_T1         'TOS variable contains top value, remove top value.
                                                        'T1 = next value from data stack. 
                        add     DatStkPtr, #2           'Remove value from data stack.
                        rdword  T2, DatStkPtr           'T2 = third value on data stack
                        add     DatStkPtr, #2           'Remove value from data stack.
DO_POP3_ret             ret
'
'
'Advance data stack to pop TOS and read next stack value into T1.
DO_POP_LOAD_T1          add     DatStkPtr, #2
                        rdword  T1, DatStkPtr
DO_POP_LOAD_T1_ret      ret
'
'
'Load the two double values at the top of the data stack.
'On exit TOS variable containsis the top double, T1 = second double
DO_LOAD_2D              call    #DO_POP3
                        'T1 = low word of long, TOS = high word of long
                        shl     TOS, #16
                        or      TOS, T1                                         'TOS = first signed double, 32 bits.
                        'now read next double, high word is already in T2
                        rdword  T1, DatStkPtr
                        'T1 = low word of long, T2 = high word of long
                        shl     T2, #16
                        or      T1, T2                                         'T1 = second signed double, 32 bits.
DO_LOAD_2D_ret          ret
'
'SYSTEM RETQUEST COMMAND
'
'Request a system command. This OpCode uses a least one parameter from the data stack.
'This is used request a support function from a helper cog running spin or asm.
'The variable "SystemCmdAdr" holds the address to a two word structure.
'The first word holds the command number to be executed.
'The first word is set to zero to indicate that the last command is completed.
'A non zero value indicates a command index to be executed.
'The second value is set to the curent data stack pointer
'value after the command index is removed from the data stack.
'After a command executes the command can set a new value for the data stack pointer
'if any parameters where added or removed from the data stack by the command.
'P SYSTEM ( n -- )
'
OP_SYSTEM               add     DatStkPtr, #2           'POP TOS from data stack
                        shl     DatStkPtr, #16
                        or      TOS, DatStkPtr
                        wrlong  TOS, PAR                'Write the command index to be executed
                        'Command is set, now wait for it to be completed
OP_SYSTEM_WAIT          rdlong  DatStkPtr, PAR
                        test    DatStkPtr, CellMask     wz
        if_nz           jmp     #OP_SYSTEM_WAIT
                        'Command completed, update the data stack pointer                        
                        shr     DatStkPtr, #16
                        jmp     #DO_LOAD_TOS            'Update the TOS variable
'                        
'
'FORTH FUNCTIONS THAT MODIFY RETURN STACK
'
'All functions that modifiy the return stack must be native codes.
                        
'Begin execution of a "DO" loop construct, or perform 2>R, both do same thing
OP_2TO_R                nop
OP_XDO                  call    #DO_POP_LOAD_T1         'Pop top value from data stack (START saved in TOS var)
                        'Push loop Limit to return stack
                        sub     RetStkPtr, #2
                        wrword  T1, RetStkPtr
                        'Push loop Start to return stack
                        'fall into >R to complete
                        'Move the TOS cell to the return stack.
OP_TO_R                 sub     RetStkPtr, #2
                        wrword  TOS, RetStkPtr
                        add     DatStkPtr, #2           'Pop value from data stack (T1 var has a copy)
                        jmp     #DO_LOAD_TOS            'Update the TOS variable to reflect the change in the stack pointer.
'
'Move the top value on the return stack to the data stack
OP_R_TO                 rdword  TOS, RetStkPtr
                        add     RetStkPtr, #2
                        jmp     #DO_PUSH_TOS            'Push value in TOS variable (16 bits) and then execute next opcode.
'                         
'Copy the top value on the return stack to the data stack,
'"I" and "R@" need different addresses for "SEE" to work.
OP_I                    nop
OP_R_FETCH              rdword  TOS, RetStkPtr
                        jmp     #DO_PUSH_TOS            'Push value in TOS variable (16 bits) and then execute next opcode.
'
'Read top two values from return stack and Return in TOS and T1.
'TOS has top of resturn stack value. 
DO_GET_2R               rdword  TOS, RetStkPtr           'Get the current loop index on the return stack
                        add     RetStkPtr, #2           'Point to the loop limit on the return stack
                        rdword  T1, RetStkPtr           'Get the loop limit value
                        sub     RetStkPtr, #2           'Restore the return stack point to orignal value
DO_GET_2R_ret           ret
'
'Fetch top 2 return stack values and place on data stack ( 2R@ )
OP_2R_FETCH             call    #DO_GET_2R
                        jmp     #DO_PUSH_T1_TOS

'MOVE top 2 return stack values and place on data stack ( 2R> )
OP_2R_TO                call    #DO_GET_2R
                        add     RetStkPtr, #4           'Drop top 2 entries on return stack
                        jmp     #DO_PUSH_T1_TOS
'
'Remove the two loop parameters from the return stack
OP_XUNLOOP              add     RetStkPtr, #4           'Pop top two values from the return stack
                        jmp     #Entry
'
'Forth word "I" returns the current DO loop count.
'The current loop count is the top value on the return stack.
'This is the same as the forth word "R>"
'
'Forth word "J" returns the next outer DO loop count.
'The next outer loop count is the 3rd value on the return stack.
'Push the outer do loop count onto the data stack
OP_J                    mov     T1, RetStkPtr
                        add     T1, #4
                        rdword  TOS, T1
                        jmp     #DO_PUSH_TOS            'Push value in TOS variable (16 bits) and then execute next opcode.
'
'                        
'PARAMATER STACK OPCODES
'
'Duplicate cell at top of data stack if value <> 0
OP_IFDUP                tjz     TOS, #Entry
'     
'Duplicate cell at top of data stack.
'( x -- x x )
OP_DUP                  jmp     #DO_PUSH_TOS
'
'
'Duplicate the top two cells of data stack.
'( x1 x2 -- x1 x2 x1 x2 )
OP_2DUP                 call    #DO_POP_LOAD_T1
                        sub     DatStkPtr, #4
                        jmp     #DO_SAVE_T1_TOS
'
'X 2DROP ( x1 x2 -- )
'Drop top two data stack cells.
OP_2DROP                add     DatStkPtr, #2
'
'Drop top data stack cell.
'C DROP ( x1 -- )
OP_DROP                 add     DatStkPtr, #2
                        jmp     #DO_LOAD_TOS
'
'Drop second data stack cell.
'X NIP ( x1 x2 -- x2 ) 
OP_NIP                  add     DatStkPtr, #2
                        jmp     #DO_SAVE_TOS
'                        
'Insert the a copy of the TOS after the second stack cell
'X TUCK ( x1 x2 -- x2 x1 x2 ) 
OP_TUCK                 call    #DO_POP_LOAD_T1           'Point to second param
                        wrword  TOS, DatStkPtr
                        jmp     #DO_PUSH_T1_TOS
'                        
'Swap the top two cells on the data stack.
'C SWAP ( x1 x2 -- x2 x1 ) 
OP_SWAP                 mov     T1, TOS                 'Save current top of data stack
                        add     DatStkPtr, #2           'Point to second cell on data stack
                        rdword  TOS, DatStkPtr          'Read the second parameter from memory.
                        jmp     #DO_SAVE_T1_TOS         'Save T1 to data stack then push value in TOS variable (16 bits), and then execute next opcode.
'
'Copy the second cell to the top of the data stack.
'C OVER ( x1 x2 -- x1 x2 x1 ) 
OP_OVER                 add     DatStkPtr, #2           'Point to second data stack cell.
                        rdword  TOS, DatStkPtr          'Read the value of the second data stack cell.
                        sub     DatStkPtr, #2           'Restore stack pointer to top of stack.
                        jmp     #DO_PUSH_TOS            'Push value in TOS variable (16 bits) and then execute next opcode.
'
'Remove u. Copy the xu to the top of the stack.
'E PICK ( xu ... x1 x0 u -- xu ... x1 x0 xu )                        
' ie: 0 PICK == DUP, 1 PICK = OVER, index is zero based.
OP_PICK                 add     TOS, #1                 'Skip over current TOS value
                        shl     TOS, #1                 'Mult by 2, bytes per cell.
                        add     TOS, DatStkPtr          'Add address of current top of data stack.
                        rdword  TOS, TOS                'Read nth cell value.
                        jmp     #DO_SAVE_TOS            'Save value in TOS variable (16 bits) over old TOS index value
                                                        'and then execute next opcode.
'
'Roll the nth cell to the top of the stack
' ie. 1 ROLL == SWAP, 2 ROLL == ROT index is one based.
'E ROLL ( xu xu-1 ... x0 u -- xu-1 ... x0 xu )
OP_ROLL                 add     DatStkPtr, #2           'Remove TOS point to second value on stack.
                        'If TOS = 0 update TOS and exit
                        tjz     TOS, #DO_LOAD_TOS
                        'Read value at nth position
OP_ROLL_INT             mov     Loop, TOS                        
                        shl     TOS, #1                 'Mult by 2, bytes per cell.
                        add     TOS, DatStkPtr          'Add address of current top of data stack.
                        mov     T3, TOS
                        rdword  TOS, T3                 'Read nth cell value and save in T2
                        'Need to move enries on stack down for at least on entry
                        'T3 = address of last cell to fill
                        sub     T3, #2                  'Point to prior cell on stack
                        'move data stack cells down
OP_ROLL1                rdword  T1, T3
                        add     T3, #2
                        wrword  T1, T3
                        sub     T3, #4                  'Point to prior cell on stack
                        djnz    Loop, #OP_ROLL1
                        jmp     #DO_SAVE_TOS            'Save value in TOS variable (16 bits) over old TOS index value
                        
'
'Rotate the third data stack value to the top of the stack
OP_ROT                  mov     TOS, #2
                        jmp     #OP_ROLL_INT
'
'Copy the data stack pointer value to the data stack
OP_SP_FETCH             mov     TOS, DatStkPtr
                        jmp     #DO_PUSH_TOS            'Push value in TOS variable (16 bits) and then execute next opcode.
'                         
'Copy the top value of the data stack to the data stack pointer
OP_SP_STORE             mov     DatStkPtr, TOS          'Set data stack pointer to the TOS value
                        jmp     #DO_LOAD_TOS            'Update the TOS variable to reflect the change in the stack pointer.
'
'Copy the value of the return stack pointer to the data stack
OP_RP_FETCH             mov     TOS, RetStkPtr
                        jmp     #DO_PUSH_TOS            'Push value in TOS variable (16 bits) and then execute next opcode.
'
'
'Copy the top value of the data stack to the return stack pointer
OP_RP_STORE             mov     RetStkPtr, TOS          'Set data stack pointer to the TOS value
                        add     DatStkPtr, #2           'Remove top entry from data stack.
                        jmp     #DO_LOAD_TOS            'Update the TOS variable to reflect the change in the stack pointer.
'
'
'LOGICAL COMPARISION OPCODES
'
'
'
' SIGNED COMPARE WORDS
'
'        
'Compare top two signed cells on data stack to see if second cell is equal to first cell.
OP_EQ                   mov     CmpOpCode, CmpEQ
                        jmp     #CmpSignedW
'
'Compare top two signed cells on data stack to see if second cell is equal to first cell.
OP_NEQ                  mov     CmpOpCode, CmpNEQ
                        jmp     #CmpSignedW
'
'Compare top two signed cells on data stack to see if second cell is less then first cell.
OP_LT                   mov     CmpOpCode, CmpLT
                        jmp     #CmpSignedW
'
'Compare top two signed cells on data stack to see if second cell is less then or equal to the first cell.
OP_LTE                  mov     CmpOpCode, CmpLTE
                        jmp     #CmpSignedW
'
'Compare top two signed cells on data stack to see if second cell is greater then first cell.
OP_GT                   mov     CmpOpCode, CmpGT
                        jmp     #CmpSignedW
'
'Compare top two signed cells on data stack to see if second cell is greater then or equal to the first cell.
OP_GTE                  mov     CmpOpCode, CmpGTE
                        'Fall into #CmpSignedW, saves one opcode
'This function does a signed compare of the top two cells
'on the data stack. The TOS variable is set to zero if
'the comapre fails or -1 (FFFF) if true.
CmpSignedW              call    #DO_POP_LOAD_T1
                        test    T1, CellSignMask        wz
              if_nz     or      T1, SignExtMask
              'Compare 0 Entry point (T1 = 0)
CmpSignedZ              test    TOS, CellSignMask       wz
              if_nz     or      TOS, SignExtMask
                        cmps    T1, TOS                 wc, wz
                        jmp     #DO_CMP_RESULT
'
'
' SIGNED COMPARE WITH ZERO WORDS
'
'Compare top signed cell on data stack with 0 and return true (FFFF) if TOS is equal to 0.
OP_ZEQ                  mov     CmpOpCode, CmpEQ
DO_Z_CMP                mov     T1, #0
                        jmp     #CmpSignedZ
'
'Compare top signed cell on data stack with 0 and return true (FFFF) if the cell is less then 0 or return 0.
OP_ZLT                  mov     CmpOpCode, CmpGT
                        jmp     #DO_Z_CMP
'
'Compare top signed cell on data stack with 0 and return true (FFFF) if the cell is greater then 0 or return 0.
OP_ZGT                  mov     CmpOpCode, CmpLT
                        jmp     #DO_Z_CMP
'
'Compare top signed cell on data stack with 0 and return true (FFFF) if the cell is <> 0 or return 0.
OP_ZNE                  mov     CmpOpCode, CmpNEQ
                        jmp     #DO_Z_CMP
'
'
' UNSIGNED COMPARE WORDS
'        
'Compare top two unsigned cells on data stack to see if second cell is less then first cell.
OP_ULT                  mov     CmpOpCode, CmpLT
                        jmp     #CmpUSignedW
'
'Compare top two signed cells on data stack to see if second cell is greater then first cell.
OP_UGT                  mov     CmpOpCode, CmpGT
                        'Fall into CmpUSignedW, saves one opcode
'This function does an unsigned compare of the top two cells
'on the data stack. The TOS variable is set to zero if
'the comapre fails or -1 (FFFF) if true.
CmpUSignedW             call    #DO_POP_LOAD_T1
CmpUSignedD             cmp     T1, TOS                 wc, wz
                        jmp     #DO_CMP_RESULT
'
'
' UNSIGNED DOUBLE COMPARE WORDS
'
'Compare top two signed doubles on data stack to see if second double is greater then first double.
OP_UDGT                 mov     CmpOpCode, CmpGT
                        call    #DO_LOAD_2D
                        jmp     #CmpUSignedD
'                        
' SIGNED DOUBLE COMPARE WORDS
'        
'Compare top two signed doubles on data stack to see if second double is less then first double.
OP_DLT                  mov     CmpOpCode, CmpLT
                        jmp     #CmpSignedD
'
'Compare top two signed doubles on data stack to see if second double is greater then first double.
OP_DGT                  mov     CmpOpCode, CmpGT
                        jmp     #CmpSignedD
'
'Compare top two signed doubles on data stack to see if second double is equal to first double.
OP_DEQ                  mov     CmpOpCode, CmpEQ
                        'Fall into CmpSignedD, saves one opcode
'This function does a signed compare of the top two double values
'on the data stack. The TOS variable is set to zero if
'the comapre fails or -1 (FFFF) if true.
CmpSignedD              call    #DO_LOAD_2D
                        'Do Compare
                        cmps    T1, TOS                 wc, wz
                        'Result of compare
DO_CMP_RESULT           mov     TOS, #0
CmpOpCode  if_c_and_nz  mov     TOS, CmpTrue            'This OPCODE is changed by each compare function
                        jmp     #DO_SAVE_TOS
'
'
'ARITHMETIC OPCODES
'
'
'If TOS is negative replace value with positive value
OP_ABS                  test    TOS, CellSignMask      wz                       
        if_nz           or      TOS, SignExtMask
                        abs     TOS, TOS
                        jmp     #DO_SAVE_TOS
'
'
'Leave min of top two signed cells on data stack.
OP_MIN                  mov     AluSignedOP, AluMinS
                        jmp     #AluSignedW
'
'
'Leave min of top two signed cells on data stack.
OP_MAX                  mov     AluSignedOP, AluMaxS
                        jmp     #AluSignedW
'
'
'Add top two cell values on data stack, pop top two cells and push result.
OP_ADD                  mov     AluSignedOP, AluAddS
                        jmp     #AluSignedW
'
'Arithmetic right shift of 2nd value on data stack by TOS (Max 16)
OP_ARSHIFT              mov     AluSignedOP, AluLogSAR
                        jmp     #AluSignedW
'
'Subtract top two cell values on data stack, pop top two cells and push result.
OP_SUB                  mov     AluSignedOP, AluSubS
                        'Fall into AluSignedW, saves one opcode
'Get the two top stack values sign extnded them then perform operation.
'The finall result is saved to the stack and the TOS variable.
AluSignedW              call    #DO_POP_LOAD_T1
                        test    TOS, CellSignMask       wz
              if_nz     or      TOS, SignExtMask
                        test    T1, CellSignMask        wz
              if_nz     or      T1, SignExtMask
'Third entry point: ALU operation involves a constant and TOS value, T1 = TOS value and TOS holds constant
AluSignedOP             subs    T1, TOS
                        mov     TOS, T1
                        jmp     #DO_SAVE_TOS            'Mask off upper 16 bits and save to data stack top value
'
'Add 2 to top cell value on data stack, pop top cell and push result.
OP_ADD_2                add     TOS, #1                        
'Add 1 to top cell value on data stack, pop top cell and push result.
OP_ADD_1                add     TOS, #1                        
                        jmp     #DO_SAVE_TOS            'Mask off upper 16 bits and save to data stack top value
'
'Subtract 2 from top cell value on data stack, pop top cell and push result.
OP_SUB_2                add     TOS, CellMask           'Add -1 same as subtract 1
'Subtract 1 from top cell value on data stack, pop top cell and push result.
OP_SUB_1                add     TOS, CellMask           'Add -1 same as subtract 1                        
                        jmp     #DO_SAVE_TOS            'Mask off upper 16 bits and save to data stack top value
'
'
'LOGICAL CPU OPCODES
'
'Logical AND of top to data stack cells
OP_AND                  mov      DO_Logical_OP, AluLogAND 
                        jmp      #DO_Logical
'                       
'Logical OR of top to data stack cells
OP_OR                   mov      DO_Logical_OP, AluLogOR 
                        jmp      #DO_Logical
'
'Logical OR of top to data stack cells
OP_XOR                  mov      DO_Logical_OP, AluLogXOR 
                        jmp      #DO_Logical
'
'Logical right shift of 2nd value on data stack by TOS (Max 16)
OP_RSHIFT               mov      DO_Logical_OP, AluLogSHR
                        jmp      #DO_Logical
'
'Logical right shift of 2nd value on data stack by TOS (Max 16)
OP_LSHIFT               mov      DO_Logical_OP, AluLogSHL
                        'Fall into DO_Logical, saves one opcode
'Logical operation of top two data stack cells.
DO_Logical              call    #DO_POP_LOAD_T1
DO_Logical_OP           and     T1, TOS
                        mov     TOS, T1
                        jmp     #DO_SAVE_TOS
'
'Perform a bit wise invert on the top data stack value
OP_INVERT               xor     TOS, CellMask
                        jmp     #DO_SAVE_TOS
'
'Add top two double signed values on data stack,
'pop top two doubles add them and push double signed result.
OP_DADD                 mov     AluSignedDOP, AluAddS
                        jmp     #AluSignedD
'
'Binaru XOR top two double values on data stack,
'pop top two doubles XOR them and push double signed result.
OP_DXOR                 mov     AluSignedDOP, AluLogXOR
                        jmp     #AluSignedD
'
'Binaru OR top two double values on data stack,
'pop top two doubles OR them and push double signed result.
OP_DOR                  mov     AluSignedDOP, AluLogOR
                        jmp     #AluSignedD
'
'Binaru AND top two double values on data stack,
'pop top two doubles AND them and push double signed result.
OP_DAND                 mov     AluSignedDOP, AluLogAND
                        jmp     #AluSignedD
'
'Subtract top two double signed values on data stack,
'pop top two doubles subtract them and push double signed result.
OP_DSUB                 mov     AluSignedDOP, AluSubS
                        'Fall into AluSignedD, saves one opcode
'Get the two top stack double values then perform operation.
'The final result is saved to the stack as a double value and the
'TOS variable will contain the high word.
AluSignedD              call    #DO_LOAD_2D
                        'We have operand 1 (T1) and operand 2 (TOS), perform double ALU operation
AluSignedDOP            subs    T1, TOS
                        'save low byte of result to data stack
DO_PUSH_T1_32           wrword  T1, DatStkPtr
                        'save high byte of result to data stack
                        shr     T1, #16
                        mov     TOS, T1
                        'TOS = top data stack value, hogh word of double result
                        jmp     #DO_PUSH_TOS
'
'Logical right shift of double value on data stack by single TOS (Max 31) value.
'DRSHIFT      d n -- d'
OP_DRSHIFT              mov      DO_SHIFT_D_OP, AluLogSHR
                        jmp      #DO_SHIFT_D
'
'Logical left shift of double value on data stack by TOS (Max 31) value
'DLSHIFT      d n -- d'
OP_DLSHIFT              mov      DO_SHIFT_D_OP, AluLogSHL
                        jmp      #DO_SHIFT_D
'
'Arithmetic right shift of double value on data stack by single TOS (Max 31) value.
'DARSHIFT      d n -- d'
OP_DARSHIFT             mov      DO_SHIFT_D_OP, AluLogSAR
                        'Fall into DO_SHIFT_D, saves one opcode
'Generic fuction to perform double shift operation
DO_SHIFT_D              call    #DO_POP3          'TOS = bits to shift T1 = High Byte, T2 = Low Byte
                        shl     T1, #16
                        or      T1, T2
DO_SHIFT_D_OP           shr     T1, TOS
                        sub     DatStkPtr, #2      'Point to next top of data stack
                        jmp     #DO_PUSH_T1_32
'
'Read a COG memory location "n" and return as a double.
'(n -- ud)
OP_COGDFETCH            movs    OP_COGDFETCH_ADR, TOS
                        nop
OP_COGDFETCH_ADR        mov     T1, (TOS)
                        jmp     #DO_PUSH_T1_32
'                            
'Write to COG memory location "n" the double "ud".
'(ud n -- )
OP_COGDSTORE            movd    OP_COGDSTORE_ADR, TOS
                        call    #DO_POP3                'TOS = n,  T1 = high word of ud, T2 = low word of ud
                        shl     T1, #16
                        or      T1, T2
OP_COGDSTORE_ADR        mov     (TOS), T1
                        jmp     #DO_LOAD_TOS
'                            
'Unsigned 32 bit multiply with 32 bit result. TOS = T1 * T2,
DoMultU32               'If T2 > T1 then flip the operands
                        cmp     T2, T1          wc
        if_nc           mov     TOS, T1
        if_nc           mov     T1, T2
        if_nc           mov     T2, TOS
                        mov     TOS, #0                                   
                        'Do the add and shift a max of 32 times, quit when (T1 << n) = 0
DoMultU32_1             shr     T2, #1          wc, wz
        if_c            add     TOS, T1
                        shl     T1, #1
        if_nz           jmp     #DoMultU32_1
DoMultU32_ret           ret
'
'
'Unsigned 32 bit division with 32 bit result and 32 bit remainder.
'On entry T1 - dividend, T2 = divisor. On exit TOS = T1 / T2, T1 = remainder. 
DoDivU32                tjnz    T2, #DoDivU32_1         'If T2 = 0 we have a divide by zero error
                        'Handle division by zero, return max value and remainder of 0
                        mov     TOS, #0
                        sub     TOS, #1
                        jmp     #DoDivU32_ret
DoDivU32_1              'Perform a normal division                                                              
                        mov     TOS, #0
                        mov     Rem, #0
                        mov     Loop, #32                                   
                        'Do the shift and sub a max of 32 times.
DoDivU32_2              shl     T1, #1          wc
                        rcl     Rem, #1
                        cmpsub  Rem, T2         wc      'C flag is set (1) if a subtraction is possible.
                        rcl     TOS, #1
                        djnz    Loop, #DoDivU32_2
                        mov     T1, Rem
DoDivU32_ret            ret
'
'
'Sign extend one or two 16 bit values variables
                        'If bit 15 = 1 we have a negative operand 1 value, sign extend it, then convert to ABS value.
DO_SIGN_16x16           test    T1, CellSignMask       wz                       
        if_nz           or      T1, SignExtMask
                        'If bit 15 = 1 we have a negative operand 2 value, sign extend it, then convert to ABS value.
DO_SIGN_32x16           test    T2, CellSignMask       wz
        if_nz           or      T2, SignExtMask
DO_SIGN_16x16_ret
DO_SIGN_32x16_ret       ret
'
'Convert signed 32bit values into positive number for division and mult.
'T1 and T2 contain the (Multiplicand, Multiplier or Dividend, Divisor).
'On exit T3 bit 31 is result sign bit,
'T3 bit 0 contains sign of multiplier or divisor.
DO_SIGN_TO_ABS          mov     T3, T1              
                        xor     T3, T2              'T3 bit 31 = result sign bit, if = 1 sign bits not same so invert result.
                        'If T1 bit 31 = 1 we have a negative operand 1 value, convert to ABS value.
                        shl     T1, #1          nr, wc
                        negc    T1, T1
                        muxc    T3, #1              'T3 bit 1 = sign of dividend
                        'If T1 bit 31 = 1 we have a negative operand 1 value, convert to ABS value.
                        shl     T2, #1          nr, wc
                        negc    T2, T2
DO_SIGN_TO_ABS_ret      ret
'
'
'Signed 16x16 bit multiply with 32 bit signed result. TOS = T1 * T2
DoMultS16               call    #DO_SIGN_16x16
'                        
'Signed 32 bit multiply with 32 bit signed result. TOS = T1 * T2
DoMultS32               call    #DO_SIGN_TO_ABS
                        'perform unsigned mult
                        call    #DoMultU32
                        'Done mutiply set sign bit, test bit 31 of T3                                                            
                        shl     T3, #1          wc
                        negc    TOS, TOS
DoMultS16_ret
DoMultS32_ret           ret
'
'
'Signed multiply (*) of top two data stack values, 16 bit signed result placed on stack.
OP_MULTS                call    #DO_POP_LOAD_T1
                        mov     T2, TOS
                        call    #DoMultS16
                        jmp     #DO_SAVE_TOS
'    
'Signed multiply (M*) of top two data stack values, 32 bit signed result placed on stack.
OP_MULTS_16_32          call    #DO_POP_LOAD_T1
                        mov     T2, TOS
                        call    #DoMultS16
DO_MULT_DONE            wrword  TOS, DatStkPtr
                        shr     TOS, #16
                        jmp     #DO_PUSH_TOS
     
'
'Signed multiply (D*) of top two double data stack values, 32 bit signed result placed on stack.
OP_MULTS_32_32          call    #DO_LOAD_2D
                        'TOS variable containsis the top double, T1 = second double
                        mov     T2, TOS
                        call    #DoMultS32
                        'TOS = 32 bit result
                        mov     T1, TOS
                        shr     TOS, #16 
                        jmp     #DO_SAVE_T1_TOS


'Unsigned multiply (U*) of top two data stack values, 32 bit unsigned result placed on stack.
OP_MULTU_16_32          call    #DO_POP_LOAD_T1
                        mov     T2, TOS
                        call    #DoMultU32
                        jmp     #DO_MULT_DONE
'                        
'Signed 16 bit division with 32 bit signed result and 32 bit signed remainder. 
DoDivS16                call    #DO_POP_LOAD_T1
                        mov     T2, TOS
                        'If bit 15 = 1 we have a negative dividend value, sign extend it.
                        test    T1, CellSignMask       wz                       
        if_nz           or      T1, SignExtMask
                        'If bit 15 = 1 we have a negative divisor value, sign extend it.
DoDivS32byS16           test    T2, CellSignMask       wz
        if_nz           or      T2, SignExtMask
                        'Perform a 32 bit signed divide
'                        
'Signed 32 bit division with 32 bit signed result and 32 bit signed remainder.
'On entry T1 - dividend, T2 = divisor.
'On exit TOS = floored quotient (T1 / T2), and T1 = floored remainder
'T2 = symetrical quotient (T1 /T2), and Rem = symetrical remainder
DoDivS32                mov     Divisor, T2             'Save divisor  
                        call    #DO_SIGN_TO_ABS
                        'perform unsigned mult
                        call    #DoDivU32
                        'Done division set sign bit of result, test bit 31 of T3                                                  
                        shl     T3, #1          nr, wc
                        negc    TOS, TOS
                        'set sign bit of remainder to sign bit of dividend
                        shr     T3, #1          nr, wc
                        negc    T1, T1
                        'Save symetrical division results to T2 = quotient, Rem = remainder
                        mov     Rem, T1
                        mov     T2, TOS
                        'if Quotient (TOS) >= 0 then MOD = REM.
                        'if Quotient (TOS) < 0 then MOD = REM + Divisor, Quotient -= 1
                        cmps    TOS, #0         wr, wc
        if_ae           jmp     #DoDivS32_ret
                        subs    TOS, #1
                        adds    T1, Divisor
DoDivS32byS16_ret                        
DoDivS16_ret                        
DoDivS32_ret            ret
'
'
'Signed floored division with modulas (/MOD) of top two data stack values,
'two 16 bit signed results placed on stack REM then Quot.
OP_DIVMOD               call    #DoDivS16
                        'save the remainder first
                        'then save the result of the div
                        jmp     #DO_SAVE_T1_TOS
'
'Signed symetrical division with remainder (/REM) of top two data stack values,
'two 16 bit signed results placed on stack REM then Quot.
OP_DIVREM               call    #DoDivS16
                        'save the remainder first
                        'then save the result of the div
                        mov     T1, Rem
                        mov     TOS, T2
                        jmp     #DO_SAVE_T1_TOS
'
'Prepare for a divistion of a double by a single.
PREP_DIV                call    #DO_POP3
                        shl     T1, #16
                        or      T1, T2
                        mov     T2, TOS
PREP_DIV_ret            ret
'
'Unigned division of a double by a single data stack values,
'16 bit unsigned results placed on stack, Quot.
'On entry TOS = unsigned single cell value
OP_UDIVMOD              call    #PREP_DIV
                        call    #DoDivU32
                        'save the remainder first
                        'then save the result of the div
                        jmp     #DO_PUSH_T1_TOS
'
'Signed floored division of a double by a signed single data stack value.
'16 bit signed results placed on stack, Quot, Mod.
'On entry TOS = signed single cell value
OP_S32DIVMOD            call    #PREP_DIV
                        call    #DoDivS32byS16
                        'save the remainder first
                        'then save the result of the div
                        jmp     #DO_PUSH_T1_TOS
'
'Signed symetrical division of a signed double by a signed single data stack value.
'16 bit signed results placed on stack, Quot, Rem.
'On entry TOS = signed single cell value
OP_S32DIVREM            call    #PREP_DIV
                        call    #DoDivS32byS16
                        'save the remainder first
                        'then save the result of the div
                        mov     T1, Rem
                        mov     TOS, T2
                        jmp     #DO_PUSH_T1_TOS
'
'Signed floored division of a signed double by a signed double data stack value.
'32 bit signed results placed on stack, Quot, Mod.
'On entry TOS = high word of 32 bit signed double cell value
OP_DIV_MOD_S32          call    #DO_LOAD_2D
                        'TOS variable containsis the top double, T1 = second double
                        mov     T2, TOS
                        call    #DoDivS32
                        'TOS = 32 bit result, T1 = remainder
                        'first save remainder
                        wrword  T1, DatStkPtr
                        sub     DatStkPtr, #2           'Point to next top of data stack
                        shr     T1, #16
                        wrword  T1, DatStkPtr
                        sub     DatStkPtr, #2           'Point to next top of data stack
                        'next save the quot 32 bits
                        mov     T1, TOS
                        jmp     #DO_PUSH_T1_32
'
'Forth Memory Manipulation Commands
'
'Read the cell value from the address on the top of the data stack.
'The address is replaced by the value read.
OP_FETCH                rdword  TOS, TOS
                        jmp     #DO_SAVE_TOS
'
'
'Read the byte value from the address on the top of the data stack.
'The address is replaced by the byte value read.
OP_CFETCH               rdbyte  TOS, TOS                      
                        jmp     #DO_SAVE_TOS
'
'Write the address on the top of the data stack with the
'cell value at the second data stack cell.
'The address and value are is popped from the data stack.
OP_STORE                call    #DO_POP_LOAD_T1         'Get second value from data stack. TOS variable contains top value.
                                                        'Read the word value, Address word in TOS variable prior to read.
OP_DO_STORE             add     DatStkPtr, #2           'Remove value from data stack
                        wrword  T1, TOS                 'Write the data value to the address suplied. This opcode is modified by OP_STORE and OP_CSTORE
                        jmp     #DO_LOAD_TOS            'Update the TOS variable to reflect the change in the stack pointer.
'
'
'Write the address on the top of the data stack with the
'value at the address added to the cell value at the second
'data stack cell.
'The address and value are is poped from the data stack.
OP_STORE_PLUS           call    #DO_POP_LOAD_T1         'Get second value from data stack. TOS variable contains top value.
                                                        'Read the word value, Address word in TOS variable prior to read.
                        rdword  T2, TOS
                        add     T1, T2
                        jmp     #OP_DO_STORE
'                        
'Write the address on the top of the data stack with the
'cell low byte value from the second data stack cell.
'The address and value are is popped from the data stack.
OP_CSTORE               call    #DO_POP_LOAD_T1         'Get second value from data stack. TOS variable contains top value.
                                                        'Read the word value, Address word in TOS variable prior to read.
                        add     DatStkPtr, #2           'Remove value from data stack
                        wrbyte  T1, TOS                 'Write the data value to the address suplied. This opcode is modified by OP_STORE and OP_CSTORE
                        jmp     #DO_LOAD_TOS            'Update the TOS variable to reflect the change in the stack pointer.
'
'Return the address for a user variable.
'The TOS contains the index of the variable.
OP_XUSER                add     TOS, UserVarsAdr
                        jmp     #DO_SAVE_TOS
'
'Read a quad byte aligned 32 bit value from memory, MSW is on top of stack
'  ( q-adr -- d )
OP_QFETCH               rdlong  T1, TOS
                        jmp     #DO_PUSH_T1_32
'
'Write a double precision value to a quad byte aligned memory location, MSW is on top of stack
'  ( d q-adr -- )
OP_QSTORE               call    #DO_POP3                'TOS = q-adr, T1 = d high word, T2 = d low word.
                        shl     T1, #16
                        or      T1, T2
                        wrlong  T1, TOS
                        jmp     #DO_LOAD_TOS
'
'Clear the value of the COG LOCK whose ID is in TOS and return prior lock state
OP_LOCKCLR              lockclr TOS             wc
                        jmp     #DO_LOCK
'                        
'Set the value of the COG LOCK whose ID is in TOS and return prior lock state
OP_LOCKSET              lockset TOS             wc
DO_LOCK                 muxc    TOS, CellMask
                        jmp     #DO_SAVE_TOS
'
'P  C@+ ( c-addr -- c-addr+1 c )
'Duplicate c-addr then read character c at address c-addr and increment c-addr
OP_CFETCH_PLUS          mov     T1, TOS         'T1 = character address
                        rdbyte  TOS, T1         'TOS = character to read c
                        add     T1, #1          'Increment address
                        jmp     #DO_SAVE_T1_TOS
'
'P REV ( u -- u' ) reverse the bits in a word
OP_REV                  rev     TOS, #16
                        jmp     #DO_SAVE_TOS
'                        
'P COG-CNT ( -- d ) return the value of the COG CNT register
OP_COG_CNT              mov     T1, cnt
                        sub     DatStkPtr, #2      'Point to next top of data stack
                        jmp     #DO_PUSH_T1_32
'                        
'Comapre Signed Operators
'                        
CmpEQ      if_z         mov     TOS, CmpTrue
CmpNEQ     if_nz        mov     TOS, CmpTrue
CmpLT      if_b         mov     TOS, CmpTrue
CmpLTE     if_be        mov     TOS, CmpTrue
CmpGT      if_a         mov     TOS, CmpTrue
CmpGTE     if_ae        mov     TOS, CmpTrue
'
'Signed arithmetic operators
'
AluAddS                 adds    T1, TOS
AluSubS                 subs    T1, TOS
AluMaxS                 mins    T1, TOS
AluMinS                 maxs    T1, TOS
'
'Logical operator code, operator order not important,
AluLogXOR               xor     T1, TOS
AluLogOR                or      T1, TOS
AluLogAND               and     T1, TOS
AluLogSHR               shr     T1, TOS
AluLogSHL               shl     T1, TOS
AluLogSAR               sar     T1, TOS
'
'
'Pre initialized variables                        
CmpTrue
CellMask      long      $FFFF
CellSignMask  long      $8000
SignExtMask   long      $FFFF0000
NegativeOne   long      $FFFFFFFF
'
'User variables pointers
UserVarsAdr   long      0
'
'Current top of stack pointers
DatStkPtr     long      0
RetStkPtr     long      0
IP            long      @COD_COLD               'Used to calculate startup global memory offset
TOS           long      0
'
'Mark end of COG image so we know how many bytes we can reclaim.
'
END_OF_COG
'
' Uninitialized variables
'
'Virtual CPU variables
''
'Temp variables
T1            res       1
T2            res       1
T3            res       1
'
'Working variables
Loop          res       1
Rem           res       1
Divisor       res       1

  
'END OF COG CODE

DAT

'Leave space for remainder of return stack, disk buffer, and User area
ExtraArea               word    0 [(size_FirstDataSpace - @ExtraArea) >>1]               
'Paramaters for forth
Param_DataStackBase     word    0
Param_ReturnStackBase   word    0
Param_LeaveStackBase    word    0
Param_BlockBufferBase   word    0
'
'Internal variables for block extension words
'Mass store internal variables
'
'Block Device IO buffer, needs to start on a long aligned address to load COG code.
var_BufferBlock         word    0               'block index 1..xx of block in buffer, 0 = no block in buffer 
var_BufferDirty         word    0               '0 = block unchanged, -1 = buffer is updated
'
'
WordTableStart
'
              word  0
              word  of_NONE                       'opcode flag set to native
WRD_COUNT     byte  5, "COUNT"
COD_COUNT     word  (@OP_CFETCH_PLUS - @Entry) >> 2
              word  NFW_EXIT
'              
'B  COUNT or C@+ ( c-addr -- c-addr+1 c )
'Duplicate c-addr then read character c at address c-addr and increment c-addr
              word  @WRD_COUNT
              word  of_NATIVE                     'opcode flag set to native
WRD_ZERO      byte  1, "0"
COD_ZERO      word  (@OP_ZERO - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_ZERO
              word  of_NATIVE                     'opcode flag set to native
WRD_ONE       byte  1, "1"
COD_ONE       word  (@OP_ONE - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_ONE
              word  of_NATIVE                     'opcode flag set to native
WRD_IFDUP     byte  4, "?DUP"
COD_IFDUP     word  (@OP_IFDUP - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_IFDUP
              word  of_NATIVE                     'opcode flag set to native
WRD_DUP       byte  3, "DUP"
COD_DUP       word  (@OP_DUP - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_DUP
              word  of_NATIVE                     'opcode flag set to native
WRD_DROP      byte  4, "DROP"
COD_DROP      word  (@OP_DROP - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_DROP
              word  of_NATIVE                     'opcode flag set to native
WRD_SWAP      byte  4, "SWAP"
COD_SWAP      word  (@OP_SWAP - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_SWAP
              word  of_NATIVE                     'opcode flag set to native
WRD_OVER      byte  4, "OVER"
COD_OVER      word  (@OP_OVER - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_OVER
              word  of_NATIVE                     'opcode flag set to native
WRD_TUCK      byte  4, "TUCK"
COD_TUCK      word  (@OP_TUCK - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_TUCK
              word  of_NATIVE                     'opcode flag set to native
WRD_NIP       byte  3, "NIP"
COD_NIP       word  (@OP_NIP - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_NIP
              word  of_NATIVE                     'opcode flag set to native
WRD_PICK      byte  4, "PICK"
COD_PICK      word  (@OP_PICK - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_PICK
              word  of_NATIVE                     'opcode flag set to native
WRD_ROLL      byte  4, "ROLL"
COD_ROLL      word (@OP_ROLL - @Entry) >> 2
              word NFW_EXIT
'
              word  @WRD_ROLL
              word  of_NATIVE                     'opcode flag set to native
WRD_ROT       byte  3, "ROT"
COD_ROT       word (@OP_ROT - @Entry) >> 2
              word NFW_EXIT
'
'C >R   (x1 -- ) (R: -- x1 ) move top cell to return stack
              word  @WRD_ROT
              word  of_NATIVE                     'opcode flag set to native
WRD_TO_R      byte  2, ">R"
COD_TO_R      word  (@OP_TO_R - @Entry) >> 2
              word  NFW_EXIT
'
'C R>   ( -- x1 ) (R: x1 -- ) move top cell from return stack to data stack
              word  @WRD_TO_R
              word  of_NATIVE                     'opcode flag set to native
WRD_R_TO      byte  2, "R>"
COD_R_TO      word  (@OP_R_TO - @Entry) >> 2
              word  NFW_EXIT
'
'C R@   ( -- x1 ) (R: x1 -- x1 ) copy top cell from return stack to data stack
              word  @WRD_R_TO
              word  of_NATIVE                     'opcode flag set to native
WRD_R_FETCH   byte  2, "R@"
COD_R_FETCH   word  (@OP_R_FETCH - @Entry) >> 2
              word  NFW_EXIT
'
'P RDROP   ( --  ) (R: x1 --  ) drop top cell from return stack
              word  @WRD_R_FETCH
              word  of_NATIVE                     'opcode flag set to native
WRD_R_DROP    byte  5, "RDROP"
COD_R_DROP    word  (@OP_RDROP - @Entry) >> 2
              word  NFW_EXIT
'
'X 2R@   ( -- x1 x2 ) (R: x1 x2 -- x1 x2 ) copy top 2 cells from return stack to data stack
              word  @WRD_R_DROP
              word  of_NATIVE                     'opcode flag set to native
WRD_2R_FETCH  byte  3, "2R@"
COD_2R_FETCH  word  (@OP_2R_FETCH - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_2R_FETCH
              word  of_NATIVE                     'opcode flag set to native
WRD_2R_TO     byte  3, "2R>"
COD_2R_TO     word  (@OP_2R_TO - @Entry) >> 2
              word  NFW_EXIT
'
'X 2>R   (x1 x2 -- ) (R: -- x1 x2 ) move top 2 cells to return stack
              word  @WRD_2R_TO
              word  of_NATIVE                     'opcode flag set to native
WRD_2TO_R     byte  3, "2>R"
COD_2TO_R     word  (@OP_2TO_R - @Entry) >> 2
              word  NFW_EXIT
'              
'C 2RDROP  (R: x1 x2 --  )        drop 2 cells from return stack
              word  @WRD_2TO_R
              word  of_NATIVE                     'opcode flag set to native
WRD_2RDROP    byte  6,"2RDROP"              
COD_2RDROP    word  (@OP_XUNLOOP - @Entry) >> 2     
              word  NFW_EXIT
'
              word  @WRD_2RDROP
              word  of_NATIVE                     'opcode flag set to native
WRD_EQ        byte  1, "="
COD_EQ        word  (@OP_EQ - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_EQ
              word  of_NATIVE                     'opcode flag set to native
WRD_NEQ       byte  2, "<>"
COD_NEQ       word  (@OP_NEQ - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_NEQ
              word  of_NATIVE                     'opcode flag set to native
WRD_LT        byte  1, "<"
COD_LT        word  (@OP_LT - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_LT
              word  of_NATIVE                     'opcode flag set to native
WRD_LTE       byte  2, "<="
COD_LTE       word  (@OP_LTE - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_LTE
              word  of_NATIVE                     'opcode flag set to native
WRD_GT        byte  1, ">"
COD_GT        word  (@OP_GT - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_GT
              word  of_NATIVE                     'opcode flag set to native
WRD_GTE       byte  2, ">="
COD_GTE       word  (@OP_GTE - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_GTE
              word  of_NATIVE                     'opcode flag set to native
WRD_ULT       byte  2, "U<"
COD_ULT       word  (@OP_ULT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_ULT
              word  of_NATIVE                     'opcode flag set to native
WRD_UGT       byte  2, "U>"
COD_UGT       word  (@OP_UGT - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_UGT
              word  of_NATIVE                     'opcode flag set to native
WRD_ZEQ       byte  2, "0="
COD_ZEQ       word  (@OP_ZEQ - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_ZEQ
              word  of_NATIVE                     'opcode flag set to native
WRD_ZNE       byte  3, "0<>"
COD_ZNE       word  (@OP_ZNE - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_ZNE
              word  of_NATIVE                     'opcode flag set to native
WRD_ZLT       byte  2, "0<"
COD_ZLT       word  (@OP_ZLT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_ZLT
              word  of_NATIVE                     'opcode flag set to native
WRD_ZGT       byte  2, "0>"
COD_ZGT       word  (@OP_ZGT - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_ZGT
              word  of_NATIVE                     'opcode flag set to native
WRD_DEQ       byte  2, "D="
COD_DEQ       word  (@OP_DEQ - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_DEQ
              word  of_NATIVE                     'opcode flag set to native
WRD_DLT       byte  2, "D<"
COD_DLT       word  (@OP_DLT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_DLT
              word  of_NATIVE                     'opcode flag set to native
WRD_DGT       byte  2, "D>"
COD_DGT       word  (@OP_DGT - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_DGT
              word  of_NATIVE                     'opcode flag set to native
WRD_UDGT      byte  3, "UD>"
COD_UDGT      word  (@OP_UDGT - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_UDGT
              word  of_NATIVE                     'opcode flag set to native
WRD_NOT       byte  3, "NOT"
COD_NOT       word  (@OP_ZEQ - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_NOT
              word  of_NATIVE                     'opcode flag set to native
WRD_ADD       byte  1,"+"              
COD_ADD       word  (@OP_ADD - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_ADD
              word  of_NATIVE                     'opcode flag set to native
WRD_SUB       byte  1,"-"              
COD_SUB       word  (@OP_SUB - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_SUB
              word  of_NATIVE                     'opcode flag set to native
WRD_DADD      byte  2,"D+"              
COD_DADD      word  (@OP_DADD - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_DADD
              word  of_NATIVE                     'opcode flag set to native
WRD_DSUB      byte  2,"D-"              
COD_DSUB      word  (@OP_DSUB - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_DSUB
              word  of_NATIVE                     'opcode flag set to native
WRD_DXOR      byte  4,"DXOR"              
COD_DXOR      word  (@OP_DXOR - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_DXOR
              word  of_NATIVE                     'opcode flag set to native
WRD_DOR       byte  3,"DOR"              
COD_DOR       word  (@OP_DOR - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_DOR
              word  of_NATIVE                     'opcode flag set to native
WRD_DAND      byte  4,"DAND"              
COD_DAND      word  (@OP_DAND - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_DAND
              word  of_NATIVE                     'opcode flag set to native
WRD_ADD1      byte  2,"1+"              
COD_ADD1      word  (@OP_ADD_1 - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_ADD1
              word  of_NATIVE                     'opcode flag set to native
WRD_SUB1      byte  2,"1-"              
COD_SUB1      word  (@OP_SUB_1 - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_SUB1
              word  of_NATIVE                     'opcode flag set to native
WRD_ADD2      byte  2,"2+"              
COD_ADD2      word  (@OP_ADD_2 - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_ADD2
              word  of_NATIVE                     'opcode flag set to native
WRD_SUB2      byte  2,"2-"              
COD_SUB2      word  (@OP_SUB_2 - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_SUB2
              word  of_NATIVE                     'opcode flag set to native
WRD_MULT      byte  1,"*"              
COD_MULT      word  (@OP_MULTS - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_MULT
              word  of_NATIVE                     'opcode flag set to native
WRD_DIVMOD    byte  4,"/MOD"              
COD_DIVMOD    word  (@OP_DIVMOD - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_DIVMOD
              word  of_NATIVE                     'opcode flag set to native
WRD_DIVREM    byte  4,"/REM"              
COD_DIVREM    word  (@OP_DIVREM - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_DIVREM
              word  of_NATIVE                        'opcode flag set to native
WRD_FMDIVMOD  byte  6,"FM/MOD"                
COD_FMDIVMOD  word  (@OP_S32DIVMOD - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_FMDIVMOD
              word  of_NATIVE                     'opcode flag set to native
WRD_SMDIVREM  byte  6,"SM/REM"          
COD_SMDIVREM  word  (@OP_S32DIVREM - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_SMDIVREM
              word  of_NATIVE                     'opcode flag set to native
WRD_UMUL      byte  3,"UM*"              
COD_UMUL      word  (@OP_MULTU_16_32 - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_UMUL
              word  of_NATIVE                     'opcode flag set to native
WRD_MMUL      byte  2,"M*"              
COD_MMUL      word  (@OP_MULTS_16_32 - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_MMUL
              word  of_NATIVE                     'opcode flag set to native
WRD_DMUL      byte  2,"D*"              
COD_DMUL      word  (@OP_MULTS_32_32 - @Entry) >> 2
              word  NFW_EXIT

              word  @WRD_DMUL
              word  of_NATIVE                     'opcode flag set to native
WRD_DDIVMOD   byte  5,"D/MOD"              
COD_DDIVMOD   word  (@OP_DIV_MOD_S32 - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_DDIVMOD
              word  of_NATIVE                     'opcode flag set to native
WRD_UDIVMOD   byte  6,"UM/MOD"              
COD_UDIVMOD   word  (@OP_UDIVMOD - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_UDIVMOD
              word  of_NATIVE                     'opcode flag set to native
WRD_MAX       byte  3,"MAX"              
COD_MAX       word  (@OP_MAX - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_MAX
              word  of_NATIVE                     'opcode flag set to native
WRD_MIN       byte  3,"MIN"              
COD_MIN       word  (@OP_MIN - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_MIN
              word  of_NATIVE                     'opcode flag set to native
WRD_ABS       byte  3,"ABS"              
COD_ABS       word  (@OP_ABS - @Entry) >> 2
              word  NFW_EXIT
'              
              word  @WRD_ABS
              word  of_NATIVE                     'opcode flag set to native
WRD_AND       byte  3,"AND"              
COD_AND       word  (@OP_AND - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_AND
              word  of_NATIVE                     'opcode flag set to native
WRD_OR        byte  2,"OR"              
COD_OR        word  (@OP_OR - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_OR
              word  of_NATIVE                     'opcode flag set to native
WRD_XOR       byte  3,"XOR"              
COD_XOR       word  (@OP_XOR - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_XOR
              word  of_NATIVE                     'opcode flag set to native
WRD_ARSHIFT   byte  7,"ARSHIFT"              
COD_ARSHIFT   word  (@OP_ARSHIFT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_ARSHIFT
              word  of_NATIVE                     'opcode flag set to native
WRD_RSHIFT    byte  6,"RSHIFT"              
COD_RSHIFT    word  (@OP_RSHIFT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_RSHIFT
              word  of_NATIVE                     'opcode flag set to native
WRD_LSHIFT    byte  6,"LSHIFT"              
COD_LSHIFT    word  (@OP_LSHIFT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_LSHIFT
              word  of_NATIVE                     'opcode flag set to native
WRD_DARSHIFT  byte  8,"DARSHIFT"              
COD_DARSHIFT  word  (@OP_DARSHIFT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_DARSHIFT
              word  of_NATIVE                     'opcode flag set to native
WRD_DRSHIFT   byte  7,"DRSHIFT"              
COD_DRSHIFT   word  (@OP_DRSHIFT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_DRSHIFT
              word  of_NATIVE                     'opcode flag set to native
WRD_DLSHIFT   byte  7,"DLSHIFT"              
COD_DLSHIFT   word  (@OP_DLSHIFT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_DLSHIFT
              word  of_NATIVE                     'opcode flag set to native
WRD_INVERT    byte  6,"INVERT"              
COD_INVERT    word  (@OP_INVERT - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_INVERT
              word  of_NATIVE                     'opcode flag set to native
WRD_FETCH     byte  1,"@"              
COD_FETCH     word  (@OP_FETCH - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_FETCH
              word  of_NATIVE                     'opcode flag set to native
WRD_STORE     byte  1,"!"              
COD_STORE     word  (@OP_STORE - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_STORE
              word  of_NATIVE                     'opcode flag set to native
WRD_PLUSSTORE byte  2,"+!"              
COD_PLUSSTORE word  (@OP_STORE_PLUS - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_PLUSSTORE
              word  of_NATIVE                     'opcode flag set to native
WRD_CFETCH    byte  2,"C@"              
COD_CFETCH    word  (@OP_CFETCH - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_CFETCH
              word  of_NATIVE                     'opcode flag set to native
WRD_CSTORE    byte  2,"C!"              
COD_CSTORE    word  (@OP_CSTORE - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_CSTORE
              word  of_NATIVE                     'opcode flag set to native
WRD_QFETCH    byte  2,"Q@"              
COD_QFETCH    word  (@OP_QFETCH - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_QFETCH
              word  of_NATIVE                     'opcode flag set to native
WRD_QSTORE    byte  2,"Q!"              
COD_QSTORE    word  (@OP_QSTORE - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_QSTORE
              word  of_NATIVE                     'opcode flag set to native
WRD_LOCKCLR   byte  8,"LOCK-CLR"              
COD_LOCKCLR   word  (@OP_LOCKCLR - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_LOCKCLR
              word  of_NATIVE                     'opcode flag set to native
WRD_LOCKSET   byte  8,"LOCK-SET"              
COD_LOCKSET   word  (@OP_LOCKSET - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_LOCKSET
              word  of_NATIVE                     'opcode flag set to native
WRD_RPFETCH   byte  3,"RP@"              
COD_RPFETCH   word  (@OP_RP_FETCH - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_RPFETCH
              word  of_NATIVE                     'opcode flag set to native
WRD_RPSTORE   byte  3,"RP!"              
COD_RPSTORE   word  (@OP_RP_STORE - @Entry) >> 2
              word  NFW_EXIT
'C SP@ ( -- n )          read current data stack pointer address
              word  @WRD_RPSTORE
              word  of_NATIVE                     'opcode flag set to native
WRD_SPFETCH   byte  3,"SP@"              
COD_SPFETCH   word  (@OP_SP_FETCH - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_SPFETCH
              word  of_NATIVE                     'opcode flag set to native
WRD_SPSTORE   byte  3,"SP!"              
COD_SPSTORE   word  (@OP_SP_STORE - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_SPSTORE
              word  of_NATIVE                     'opcode flag set to native
WRD_SYSTEM    byte  6,"SYSTEM"              
COD_SYSTEM    word  (@OP_SYSTEM - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_SYSTEM
              word  of_NATIVE                     'opcode flag set to native
WRD_XUSER     byte  6,"(user)"              
COD_XUSER     word  (@OP_XUSER - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_XUSER
              word  of_NATIVE                     'opcode flag set to native
WRD_XPLUSLOOP byte  7,"(+loop)"              
COD_XPLUSLOOP word  (@OP_XPLUSLOOP - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_XPLUSLOOP
              word  of_NATIVE                     'opcode flag set to native
WRD_XLOOP     byte  6,"(loop)"              
COD_XLOOP     word  (@OP_XLOOP - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_XLOOP
              word  of_NATIVE                     'opcode flag set to native
WRD_XDO       byte  4,"(do)"              
COD_XDO       word  (@OP_XDO - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_XDO
              word  of_NATIVE                     'opcode flag set to native
WRD_UNLOOP    byte  6,"UNLOOP"              
COD_UNLOOP    word  (@OP_XUNLOOP - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_UNLOOP
              word  of_NATIVE                     'opcode flag set to native
WRD_I         byte  1,"I"              
COD_I         word  (@OP_I - @Entry) >> 2
              word  NFW_EXIT
'
              word  @WRD_I
              word  of_NATIVE                     'opcode flag set to native
WRD_J         byte  1,"J"              
COD_J         word  (@OP_J - @Entry) >> 2
              word  NFW_EXIT
'
'C 2DUP   x1 x2 -- x1 x2 x1 x2   dup top 2 cells
'   OVER OVER ;
              word  @WRD_J
              word  of_NATIVE                     'opcode flag set to native
WRD_2DUP      byte  4, "2DUP"
COD_2DUP      word  (@OP_2DUP - @Entry) >> 2
              word  NFW_EXIT
'
'C 2DROP  x1 x2 --          drop 2 cells
'   DROP DROP ;
              word  @WRD_2DUP
              word  of_NATIVE                     'opcode flag set to native
WRD_2DROP     byte  5,"2DROP"              
COD_2DROP     word  (@OP_2DROP - @Entry) >> 2     
              word  NFW_EXIT
'
'P COGD@      u -- ud
'Read the 32 bit memory location "u" from cog memoru
              word  @WRD_2DROP
              word  of_NATIVE                     'opcode flag set to native
WRD_COGDFETCH byte  4, "COG@"
COD_COGDFETCH word  (@OP_COGDFETCH - @Entry) >> 2
              word  NFW_EXIT
'
'P COGD!     ud n -- 
'Write the memory location "n" in cog memoru with the double value ud
              word  @WRD_COGDFETCH
              word  of_NATIVE                     'opcode flag set to native
WRD_COGDSTORE byte  4, "COG!"
COD_COGDSTORE word  (@OP_COGDSTORE - @Entry) >> 2
              word  NFW_EXIT
'
'P (literal)  runtime code for literal
              word  @WRD_COGDSTORE
              word  of_NATIVE                    'opcode flag set to native
WRD_DOLIT     byte  9, "(literal)"
COD_DOLIT     word  (@OP_LITERAL - @Entry) >> 2
              word  NFW_EXIT
          
'P (2literal)  runtime code for literal
              word  @WRD_DOLIT
              word  of_NATIVE                    'opcode flag set to native
WRD_DO2LIT    byte  10, "(2literal)"
COD_DO2LIT    word  (@OP_DLITERAL - @Entry) >> 2
              word  NFW_EXIT
'
'P (branch)  native virtual cpu branch instruction
              word  @WRD_DO2LIT
              word  of_NATIVE                    'opcode flag set to native
WRD_BRANCH    byte  8, "(branch)"
COD_BRANCH    word  (@OP_BRANCH - @Entry) >> 2
              word  NFW_EXIT
'
'P (?branch)   ( -- )   exit current code thread
              word  @WRD_BRANCH
              word  of_NATIVE                       'opcode flag set to native
WRD_QBRANCH   byte  9,"(0branch)"
COD_QBRANCH   word  (@OP_QBRANCH - @Entry) >> 2
              word  NFW_EXIT
'             
'P (does)   ( -- )   exit current code thread
               word  @WRD_QBRANCH
               word  of_NATIVE                       'opcode flag set to native
WRD_DOESBRANCH byte  6,"(does)"
COD_DOESBRANCH word  (@OP_DOES_BRANCH - @Entry) >> 2
               word  NFW_EXIT
'             
'C EXIT   ( -- )   exit current code thread
              word  @WRD_DOESBRANCH
              word  of_NATIVE                       'opcode flag set to native
WRD_EXIT      byte  4,"EXIT"
COD_EXIT      word  NFW_EXIT
'             
'P EXIT-ADR  (  -- a-adr )
'Exit current code thread and leave next code address on data stack
              word  @WRD_EXIT
              word  of_NATIVE                       'opcode flag set to native
WRD_EXIT_ADR  byte  8,"EXIT-ADR"
COD_EXIT_ADR  word  NFW_EXIT_ADR
              word  NFW_EXIT
              
'P REV  ( u -- u' )
'Reverse the data bits in the word
              word  @WRD_EXIT_ADR
              word  of_NATIVE                       'opcode flag set to native
WRD_REV       byte  3,"REV"
              word  (@OP_REV - @Entry) >> 2
              word  NFW_EXIT
'
'P COG-CNT ( -- d )
'Return the value of the COG CNT register
              word  @WRD_REV
              word  of_NATIVE                       'opcode flag set to native
WRD_COG_CNT   byte  7,"COG-CNT"
COD_COG_CNT   word  (@OP_COG_CNT - @Entry) >> 2
              word  NFW_EXIT
'
'
'FORTH WORDS WRITTEN IN FORTH
'
'P NOP  --    no operation Forth word just eats up time
'NOP must be the first dictionary word that is not a native
'virtual CPU opcode. The address of NOP is used by the
'compiler's optimizer to determin if an address points to a
'NATIVE code thread.
              word  @WRD_COG_CNT
              word  of_NONE                       'opcode flag set to normal
WRD_NOP       byte  3,"NOP"
COD_NOP       word  (@OP_NOP - @Entry) >> 2
              word  NFW_EXIT
'
'C EXECUTE   xt --
'Execute the execution token "xt" and return to calling thread.
'This forth word must not be used as a native opcode.
              word  @WRD_NOP
              word  of_NONE                       'opcode flag set to normal
WRD_EXECUTE   byte  7,"EXECUTE"
COD_EXECUTE   word  (@OP_EXECUTE - @Entry) >> 2
              word  NFW_EXIT
'
'X TRUE    -- f             leave the value for logical true
              word  @WRD_EXECUTE
              word  of_NONE                       'opcode flag set to normal
WRD_TRUE      byte  4,"TRUE"
COD_TRUE      word  NFW_LIT, -1
              word  NFW_EXIT
'
'X FALSE    -- f             leave the value for logical true
              word  @WRD_TRUE
              word  of_NONE                       'opcode flag set to normal
WRD_FALSE     byte  5,"FALSE"
COD_FALSE     word  @COD_ZERO
              word  NFW_EXIT
'
'C FILL   ( c-addr u char -- ) fill memory with char
              word  @WRD_FALSE
              word  of_NONE                       'opcode flag set to normal
WRD_FILL      byte  4,"FILL"              
COD_FILL      word  NFW_LIT, cmd_FILL, @COD_SYSTEM
              word  NFW_EXIT
'
'
'P REBOOT  ( -- ) hardware restart of CPU
              word  @WRD_FILL
              word  of_NONE                       'opcode flag set to normal
WRD_REBOOT    byte  6,"REBOOT"              
              word  NFW_LIT, cmd_REBOOT, @COD_SYSTEM
              word  NFW_EXIT
'
'S CMOVE    ( c-addr1 c-addr2 u -- )
'If u is greater than zero, copy u consecutive characters from the data space starting at c-addr1
'to that starting at c-addr2, proceeding character-by-character from lower addresses to higher
'addresses.
              word  @WRD_REBOOT
              word  of_NONE                       'opcode flag set to normal
WRD_CMOVE     byte  5, "CMOVE"
COD_CMOVE     word  NFW_LIT, cmd_CMOVE, @COD_SYSTEM
              word  NFW_EXIT
'
'S CMOVE>    ( c-addr1 c-addr2 u -- )
'If u is greater than zero, copy u consecutive characters from the data space starting at c-addr1
'to that starting at c-addr2, proceeding character-by-character from higher addresses to lower
'addresses.
              word  @WRD_CMOVE
              word  of_NONE                       'opcode flag set to normal
WRD_CMOVEREV  byte  6, "CMOVE>"
COD_CMOVEREV  word  NFW_LIT, cmd_CMOVE_REV, @COD_SYSTEM
              word  NFW_EXIT
'
'FORTH ALIGNMENT AND PORTABILITY OPERATORS
'
'C ALIGNED  addr -- a-addr       align given addr
              word  @WRD_CMOVEREV
              word  of_NONE                       'opcode flag set to normal
WRD_ALIGNED   byte  7,"ALIGNED"
COD_ALIGNED   word  @COD_DUP, @COD_ONE, @COD_AND, @COD_ADD
              word  NFW_EXIT
'              
'C ALIGN    --                         align HERE
              word  @WRD_ALIGNED
              word  of_NONE                       'opcode flag set to normal
WRD_ALIGN     byte  5,"ALIGN"
COD_ALIGN     word  @COD_HERE, @COD_ALIGNED
              word  @COD_DP, @COD_STORE
              word  NFW_EXIT
'
'P QALIGNED ( uAdr -- uAdr' )
'Align TOS to a quad based boundry ( div 4 evenly )
              word  @WRD_ALIGN
              word  of_NONE                       'opcode flag set to normal
WRD_QALIGNED  byte  8,"QALIGNED"
COD_QALIGNED  word  @COD_DUP, NFW_LIT, 3, @COD_AND, NFW_QBRANCH, @COD_QALIGNED1
              word  NFW_LIT, $FFFC, @COD_AND, NFW_LIT, 4, @COD_ADD    
COD_QALIGNED1 word  NFW_EXIT
'
'Z CELL     -- n                 size of one cell
              word  @WRD_QALIGNED
              word  of_NONE                       'opcode flag set to normal
WRD_CELL      byte  4,"CELL"
COD_CELL      word  NFW_LIT, 2
              word  NFW_EXIT
'
'C CELL+    a-addr1 -- a-addr2      add cell size
              word  @WRD_CELL
              word  of_NONE                       'opcode flag set to normal
WRD_CELLPLUS  byte  5,"CELL+"
COD_CELLPLUS  word  @COD_ADD2
              word  NFW_EXIT
'
'C CELLS    n1 -- n2            cells->adrs units
              word  @WRD_CELLPLUS
              word  of_NONE                       'opcode flag set to normal
WRD_CELLS     byte  5,"CELLS"
COD_CELLS     word  @COD_ONE, @COD_LSHIFT
              word  NFW_EXIT
'
'C CHAR+     c-addr1 -- c-addr2   add char size
              word  @WRD_CELLS
              word  of_NONE                       'opcode flag set to normal
WRD_CHARPLUS  byte  5,"CHAR+"
COD_CHARPLUS  word  @COD_ADD1
              word  NFW_EXIT
'             
'C CHARS     n1 -- n2            chars->adrs units
              word  @WRD_CHARPLUS
              word  of_NONE                       'opcode flag set to normal
WRD_CHARS     byte  5,"CHARS"
COD_CHARS     word  NFW_EXIT
'
'X ERASE   c-addr u --  fill memory with zero
'Set all bits for u bytes to zero. 
              word  @WRD_CHARS
              word  of_NONE                       'opcode flag set to normal
WRD_ERASE     byte  5,"ERASE"              
COD_ERASE     word  @COD_ZERO, @COD_FILL
              word  NFW_EXIT
'
'S BLANK   c-addr u --  fill string with space characters
'Set all u characters to a space character. 
              word  @WRD_ERASE
              word  of_NONE                       'opcode flag set to normal
WRD_BLANK     byte  5,"BLANK"              
COD_BLANK     word  @COD_BL, @COD_FILL
              word  NFW_EXIT
'
'Forth system variables and constants
'
'
'P BL ( -- c ) Returns ASCI space character
              word  @WRD_BLANK
              word  of_VALUE                     'opcode flag set to value
WRD_BL        byte  2,"BL"              
COD_BL        word  NFW_LIT, 32
              word  NFW_EXIT
'
'P BS ( -- c ) Returns ASCI back space character
              word  @WRD_BL
              word  of_VALUE                     'opcode flag set to value
WRD_BS        byte  2,"BS"              
COD_BS        word  NFW_LIT, 8
              word  NFW_EXIT
'
'Returns the address of offset into TIB area.
              word  @WRD_BS
              word  of_NONE                     'opcode flag set to normal
WRD_TOIN      byte  3,">IN"              
COD_TOIN      word  NFW_LIT, uof_ToIn, @COD_XUSER
              word  NFW_EXIT
'
'Returns the address of numeric conversion base.
              word  @WRD_TOIN
              word  of_NONE                     'opcode flag set to normal
WRD_BASE      byte  4,"BASE"              
COD_BASE      word  NFW_LIT, uof_Base, @COD_XUSER
              word  NFW_EXIT
'
'Returns the address of the compiler STATE variable.
              word  @WRD_BASE
              word  of_NONE                     'opcode flag set to normal
WRD_STATE     byte  5,"STATE"              
COD_STATE     word  NFW_LIT, uof_State, @COD_XUSER
              word  NFW_EXIT
'
'Returns the dictionary pointer address.
              word  @WRD_STATE
              word  of_NONE                     'opcode flag set to normal
WRD_DP        byte  2,"DP"              
COD_DP        word  NFW_LIT, uof_DictionaryPointer, @COD_XUSER
              word  NFW_EXIT
'
'Stores address of last source ID
              word  @WRD_DP
              word  of_NONE                     'opcode flag set to normal
WRD_SOURCE_ID byte  9,"SOURCE-ID"              
COD_SOURCE_ID word  NFW_LIT, uof_SourceID, @COD_XUSER, @COD_FETCH
              word  NFW_EXIT
'
'Stores address of last source string two cells: len, adrs
              word  @WRD_SOURCE_ID
              word  of_NONE                     'opcode flag set to normal
WRD_TSOURCE   byte  7,$27,"SOURCE"              
COD_TSOURCE   word  NFW_LIT, uof_Source, @COD_XUSER
              word  NFW_EXIT
'
'Returns address of LATEST variable.
              word  @WRD_TSOURCE
              word  of_NONE                     'opcode flag set to normal
WRD_LATEST    byte  6,"LATEST"              
COD_LATEST    word  NFW_LIT, uof_Latest, @COD_XUSER
              word  NFW_EXIT
'
'Returns the hold pointer address.
              word  @WRD_LATEST
              word  of_NONE                     'opcode flag set to normal
WRD_HP        byte  2,"HP"              
COD_HP        word  NFW_LIT, uof_HoldPointer, @COD_XUSER
              word  NFW_EXIT
'
'Returns the Leave-stack pointer address.
              word  @WRD_HP
              word  of_NONE                     'opcode flag set to normal
WRD_LP        byte  2,"LP"              
COD_LP        word  NFW_LIT, uof_LeavePointer, @COD_XUSER
              word  NFW_EXIT
'
'Returns the user PAD buffer address.
              word  @WRD_LP
              word  of_NONE                     'opcode flag set to normal
WRD_PAD       byte  3,"PAD"              
COD_PAD       word  NFW_LIT, uof_PAD, @COD_XUSER
              word  NFW_EXIT
'
'Returns address to bottom of Leave stack.
              word  @WRD_PAD
              word  of_NONE                     'opcode flag set to normal
WRD_L0        byte  2,"L0"              
COD_L0        word  NFW_ALIT, @Param_LeaveStackBase, @COD_FETCH
              word  NFW_EXIT
'
'Returns address to current I/O block number set by BUFFER and BLOCK
              word  @WRD_L0
              word  of_NONE                     'opcode flag set to normal
WRD_BLK       byte  3,"BLK"              
COD_BLK       word  NFW_LIT, uof_BlockNumber, @COD_XUSER
              word  NFW_EXIT
'
'Returns address to last listed block number set by LIST
              word  @WRD_BLK
              word  of_NONE                     'opcode flag set to normal
WRD_SCR       byte  3,"SCR"              
COD_SCR       word  NFW_LIT, uof_ScreenBlockNumber, @COD_XUSER
              word  NFW_EXIT
'
'C S0 ( -- n )          read data start start address
              word  @WRD_SCR
              word  of_NONE                      'opcode flag set to normal
WRD_S0        byte  2, "S0"
COD_S0        word  NFW_ALIT, @Param_DataStackBase, @COD_FETCH
              word  NFW_EXIT
'              
'C R0 ( -- n )          read return start start address
              word  @WRD_S0
              word  of_NONE                     'opcode flag set to normal
WRD_R0        byte  2, "R0"
COD_R0        word  NFW_ALIT, @Param_ReturnStackBase, @COD_FETCH
              word  NFW_EXIT
'
'Returns size of the temporary input buffer
              word  @WRD_R0
              word  of_NONE                     'opcode flag set to normal
WRD_TIBSIZE   byte  7,"TIBSIZE"              
COD_TIBSIZE   word  NFW_LIT, size_TibBuffer
              word  NFW_EXIT
'              
'Returns address of the temporary input buffer
              word  @WRD_TIBSIZE
              word  of_NONE                     'opcode flag set to normal
WRD_TIB       byte  3,"TIB"              
COD_TIB       word  NFW_LIT, uof_TIB, @COD_XUSER
              word  NFW_EXIT
'              
'Returns start address of USER area.
              word  @WRD_TIB
              word  of_NONE                     'opcode flag set to normal
WRD_U0        byte  2,"U0"              
COD_U0        word  @COD_ZERO, @COD_XUSER
              word  NFW_EXIT
'
'P (set-source-id) ( f -- )  set the source id TRUE/FALSE
              word  @WRD_U0
              word  of_NONE                     'opcode flag set to normal
WRD_TO_SOURCE byte  11,"(source-id)"              
COD_TO_SOURCE_ID word  NFW_LIT, uof_SourceID, @COD_XUSER, @COD_STORE
                 word  NFW_EXIT
'

'Initialize forth variables
'
              word  @WRD_TO_SOURCE
              word  of_NONE                     'opcode flag set to normal
WRD_INIT      byte  6,"(init)"
              'Clear all user variables (set = 0).
              'set >IN = 0, STATE = 0, SOURCE-ID = 0, input from user device,
              'SOURCE address = 0, SOURCE length = 0, BLK = 0, SCR = 0
COD_INIT      word  @COD_U0, NFW_LIT, uof_SIZE_USER_VARS, @COD_ERASE
              'Initialize remaining variables
              word  NFW_LIT, 10, @COD_BASE, @COD_STORE                          'set BASE = 10
              word  NFW_ALIT, @UserCodeEnd, @COD_DP, @COD_STORE                 'Init DP to address of user code start
              word  NFW_ALIT, @WRD_COLD, @COD_LATEST, @COD_STORE                'Init LATEST to address of COLD word (always last word)
              'Reset base leave stack and hold pointers
              word  @COD_PAD, @COD_HP, @COD_STORE       'set HP = Start of PAD buffer, decrements as characters are added
              word  @COD_S0, @COD_SPSTORE               'Clear the data stack
              word  @COD_L0, @COD_LP, @COD_STORE        'Clear the leave (control) stack
              word  NFW_EXIT
'
'C DEPTH    -- +n        number of items on stack
'   SP@ S0 SWAP - 2/ ;   16-BIT VERSION!
'Write the number of cell values on the data stack.
              word  @WRD_INIT
              word  of_NONE                     'opcode flag set to normal
WRD_DEPTH     byte  5,"DEPTH"              
COD_DEPTH     word  @COD_SPFETCH, @COD_S0, @COD_SWAP, @COD_SUB, @COD_2DIV                 
              word  NFW_EXIT
'
'C UNUSED    -- +n        number of unused bytes
'   SP@ HERE - ;   16-BIT VERSION!
'Write the number of cell values on the data stack.
              word  @WRD_DEPTH
              word  of_NONE                     'opcode flag set to normal
WRD_UNUSED    byte  6,"UNUSED"              
COD_UNUSED    word  @COD_SPFETCH, @COD_HERE, @COD_SUB     
              word  NFW_EXIT
'
'
'FORTH ARITHMETIC OPERATORS
'
              word  @WRD_UNUSED
              word  of_NONE                     'opcode flag set to normal
WRD_DIV       byte  1,"/"              
COD_DIV       word  @COD_DIVMOD, @COD_NIP
              word  NFW_EXIT
'              
              word  @WRD_DIV
              word  of_NONE                     'opcode flag set to normal
WRD_SYM_DIV   byte  2,"/-"             
COD_SYM_DIV   word  @COD_DIVREM, @COD_NIP
              word  NFW_EXIT
'              
              word  @WRD_SYM_DIV
              word  of_NONE                     'opcode flag set to normal
WRD_MOD       byte  3,"MOD"              
COD_MOD       word  @COD_DIVMOD, @COD_DROP
              word  NFW_EXIT
'              
              word  @WRD_MOD
              word  of_NONE                     'opcode flag set to normal
WRD_REM       byte  3,"REM"              
COD_REM       word  @COD_DIVREM, @COD_DROP
              word  NFW_EXIT

              word  @WRD_REM
              word  of_NONE                     'opcode flag set to normal
WRD_DDIV      byte  2,"D/"              
COD_DDIV      word  @COD_DDIVMOD, @COD_2SWAP, @COD_2DROP
              word  NFW_EXIT
'              
              word  @WRD_DDIV
              word  of_NONE                     'opcode flag set to normal
WRD_DMOD       byte  3,"DMOD"              
COD_DMOD       word  @COD_DDIVMOD, @COD_2DROP
              word  NFW_EXIT
'              
'Convert the top data stack value into two cell signed double
              word  @WRD_DMOD
              word  of_NONE                     'opcode flag set to normal
WRD_STOD      byte  3,"S>D"              
COD_STOD      word  @COD_DUP, @COD_ZLT    
              word  NFW_EXIT
'
'Convert the top two cell signed double value into single signed value
              word  @WRD_STOD
              word  of_NONE                     'opcode flag set to normal
WRD_DTOS      byte  3,"D>S"              
COD_DTOS      word  @COD_DROP    
              word  NFW_EXIT
'
              word  @WRD_DTOS
              word  of_NONE                     'opcode flag set to normal
WRD_FMULDIV   byte  2,"*/"               
COD_FMULDIV   word @COD_TO_R, @COD_MMUL, @COD_R_TO, @COD_FMDIVMOD, @COD_NIP
              word  NFW_EXIT
'
              word  @WRD_FMULDIV
              word  of_NONE                     'opcode flag set to normal
WRD_SMULDIV   byte  3,"*/-"              
COD_SMULDIV   word @COD_TO_R, @COD_MMUL, @COD_R_TO, @COD_SMDIVREM, @COD_NIP
              word  NFW_EXIT
'
              word  @WRD_SMULDIV
              word  of_NONE                     'opcode flag set to normal
WRD_MULDIVMOD byte  5,"*/MOD"            
COD_MULDIVMOD word @COD_TO_R, @COD_MMUL, @COD_R_TO, @COD_FMDIVMOD
              word  NFW_EXIT
'
              word  @WRD_MULDIVMOD
              word  of_NONE                     'opcode flag set to normal
WRD_MULDIVREM byte  5,"*/REM"            
COD_MULDIVREM word @COD_TO_R, @COD_MMUL, @COD_R_TO, @COD_SMDIVREM
              word  NFW_EXIT
'
              word  @WRD_MULDIVREM
              word  of_NONE                     'opcode flag set to normal
WRD_NEGATE    byte  6,"NEGATE"              
COD_NEGATE    word  @COD_INVERT, @COD_ADD1
              word  NFW_EXIT
'              
'If the TOS value is negative, negate the second data stack value
              word  @WRD_NEGATE
              word  of_NONE                     'opcode flag set to normal
WRD_QNEGATE   byte  7,"?NEGATE"              
COD_QNEGATE   word  @COD_ZLT, NFW_QBRANCH, @COD_QNEGATE1, @COD_NEGATE    
COD_QNEGATE1  word  NFW_EXIT
'
'Add the top double value on the data stack to the next value on the stack, leaving a double result.
              word  @WRD_QNEGATE
              word  of_NONE                     'opcode flag set to normal
WRD_MADD      byte  2,"M+"              
COD_MADD      word  @COD_DUP, @COD_ZLT, @COD_DADD    
              word  NFW_EXIT
'
'Negate the top double value on the data stack.
              word  @WRD_MADD
              word  of_NONE                     'opcode flag set to normal
WRD_DNEGATE   byte  7,"DNEGATE"              
COD_DNEGATE   word  @COD_SWAP, @COD_INVERT, @COD_SWAP, @COD_INVERT, @COD_ONE, @COD_ZERO, @COD_DADD    
              word  NFW_EXIT
'
'If the TOS value is negative, Negate the following double value on the data stack.
              word  @WRD_DNEGATE
              word  of_NONE                     'opcode flag set to normal
WRD_QDNEGATE  byte  8,"?DNEGATE"              
COD_QDNEGATE  word  @COD_ZLT, NFW_QBRANCH, @COD_QDNEGATE1, @COD_DNEGATE    
COD_QDNEGATE1 word  NFW_EXIT
'
'If the double at the top of the data stack is negative, convert to positive double
              word  @WRD_QDNEGATE
              word  of_NONE                     'opcode flag set to normal
WRD_DABS      byte  4,"DABS"              
COD_DABS      word  @COD_DUP, @COD_QDNEGATE    
              word  NFW_EXIT
'
'D D0<  d -- f
'If the double at the top of the data stack is negative, return true -1, else return 0
              word  @WRD_DABS
              word  of_NONE                     'opcode flag set to normal
WRD_DZLT      byte  3,"D0<"               
COD_DZLT      word  NFW_DLIT, 0, 0, @COD_DLT
              word  NFW_EXIT
'
'D D0<  d -- f
'If the double at the top of the data stack equals zero return true -1, else return 0
              word  @WRD_DZLT
              word  of_NONE                     'opcode flag set to normal
WRD_DZEQ      byte  3,"D0="               
COD_DZEQ      word  NFW_DLIT, 0, 0, @COD_DEQ
              word  NFW_EXIT
'
'P D0>  d -- f
'If the double at the top of the data stack is greater then zero return true -1, else return 0
              word  @WRD_DZEQ
              word  of_NONE                     'opcode flag set to normal
WRD_DZGT      byte  3,"D0>"               
COD_DZGT      word  NFW_DLIT, 0, 0, @COD_DGT
              word  NFW_EXIT
'
'D DMIN d1 d2 -- d3
'leave the minimum of the two top double numbers
              word  @WRD_DZGT
              word  of_NONE                     'opcode flag set to normal
WRD_DMIN      byte  4,"DMIN"              
COD_DMIN      word  @COD_2DUP, @COD_2ROT, @COD_2DUP, @COD_2ROT, @COD_DLT, NFW_QBRANCH, @COD_DMIN1
              word  @COD_2SWAP
COD_DMIN1     word  @COD_2DROP
              word  NFW_EXIT
'
'D DMAX d1 d2 -- d3
'leave the maximun of the two top double numbers on the stack
              word  @WRD_DMIN
              word  of_NONE                     'opcode flag set to normal
WRD_DMAX      byte  4,"DMAX"              
COD_DMAX      word  @COD_2DUP, @COD_2ROT, @COD_2DUP, @COD_2ROT, @COD_DGT, NFW_QBRANCH, @COD_DMAX1
              word  @COD_2SWAP
COD_DMAX1     word  @COD_2DROP
              word  NFW_EXIT
'
'
'FORTH DOUBLE OPERATORS
'
'C 2@    a-addr -- x1 x2    fetch 2 cells
'   DUP CELL+ @ SWAP @ ;
'   the lower address will appear on top of stack
              word  @WRD_DMAX
              word  of_NONE                     'opcode flag set to normal
WRD_2FETCH    byte  2,"2@"              
COD_2FETCH    word  @COD_DUP, @COD_CELLPLUS, @COD_FETCH, @COD_SWAP, @COD_FETCH     
              word  NFW_EXIT
'
'C 2!    x1 x2 a-addr --    store 2 cells
'   SWAP OVER ! CELL+ ! ;
'   the top of stack is stored at the lower adrs
              word  @WRD_2FETCH
              word  of_NONE                     'opcode flag set to normal
WRD_2STORE    byte  2,"2!"              
COD_2STORE    word  @COD_SWAP, @COD_OVER, @COD_STORE, @COD_CELLPLUS, @COD_STORE     
              word  NFW_EXIT
'
'C 2ROT   x1 x2 x3 x4 x5 x6 -- x3 x4 x5 x6 x1 x2  roate 3rd double to TOS
'   DROP DROP ;
              word  @WRD_2STORE
              word  of_NONE                     'opcode flag set to normal
WRD_2ROT      byte  4,"2ROT"               
COD_2ROT      word  NFW_LIT, 5, @COD_ROLL, NFW_LIT, 5, @COD_ROLL
              word  NFW_EXIT
'
'C 2SWAP  x1 x2 x3 x4 -- x3 x4 x1 x2  per diagram
'   ROT >R ROT R> ;
              word  @WRD_2ROT
              word  of_NONE                     'opcode flag set to normal
WRD_2SWAP     byte  5,"2SWAP"              
COD_2SWAP     word  @COD_ROT, @COD_TO_R, @COD_ROT, @COD_R_TO      
              word  NFW_EXIT
'              
'C 2OVER  x1 x2 x3 x4 -- x1 x2 x3 x4 x1 x2
'   >R >R 2DUP R> R> 2SWAP ;
              word  @WRD_2SWAP
              word  of_NONE                     'opcode flag set to normal
WRD_2OVER     byte  5,"2OVER"              
COD_2OVER     word  @COD_TO_R, @COD_TO_R, @COD_2DUP, @COD_R_TO, @COD_R_TO, @COD_2SWAP      
              word  NFW_EXIT
'
'
'FORTH NUMERIC OUTPUT
'
' Numeric conversion is done l.s.digit first, so
' the output buffer is built backwards in memory.

' Some double-precision arithmetic operators are
' needed to implement ANSI numeric conversion.
'
'C 2/       n -- n'
'Signed floored division of n by 2, n' = n/2
              word  @WRD_2OVER
              word  of_NONE                     'opcode flag set to normal
WRD_2DIV      byte  2,"2/"            
COD_2DIV      word  @COD_ONE, @COD_ARSHIFT
              word  NFW_EXIT
'
'C 2*       n -- n'
'Signed multiplication of n by 2, n' = n * 2
              word  @WRD_2DIV
              word  of_NONE                     'opcode flag set to normal
WRD_2MUL      byte  2,"2*"            
COD_2MUL      word  @COD_ONE, @COD_LSHIFT
              word  NFW_EXIT
'
'D D2/       d -- d'
'Signed floored double precision division of d by 2, d' = d/2
              word  @WRD_2MUL
              word  of_NONE                     'opcode flag set to normal
WRD_D2DIV     byte  3,"D2/"            
COD_D2DIV     word  @COD_ONE, @COD_DARSHIFT
              word  NFW_EXIT
'
'D D2*       d -- d'
'Signed multiplication of double d by 2, d' = d * 2
              word  @WRD_D2DIV
              word  of_NONE                     'opcode flag set to normal
WRD_D2MUL     byte  3,"D2*"            
COD_D2MUL     word  @COD_ONE, @COD_DLSHIFT
              word  NFW_EXIT
'
'P UD/MOD   ud1 u2 -- u3 ud4   32/16->32 divide
'   >R 0 R@ UM/MOD  ROT ROT R> UM/MOD ROT ;
              word  @WRD_D2MUL
              word  of_NONE                     'opcode flag set to normal
WRD_UDDIVMOD  byte  6,"UD/MOD"            
COD_UDDIVMOD  word  @COD_TO_R, @COD_ZERO, @COD_R_FETCH, @COD_UDIVMOD, @COD_ROT, @COD_ROT
              word  @COD_R_TO, @COD_UDIVMOD, @COD_ROT
              word  NFW_EXIT
'
'P UD*       ud1 u -- ud3      32*16->32 multiply
'   DUP >R UM* DROP  SWAP R> UM* ROT + ;
              word  @WRD_UDDIVMOD
              word  of_NONE                     'opcode flag set to normal
WRD_UDMUL     byte  3,"UD*"               
COD_UDMUL     word  @COD_DUP, @COD_TO_R, @COD_UMUL, @COD_DROP, @COD_SWAP, @COD_R_TO
              word  @COD_UMUL, @COD_ROT, @COD_ADD
              word  NFW_EXIT
'
'C HOLD   char --        add char to output string
'   -1 HP +!  HP @ C! ;
              word  @WRD_UDMUL
              word  of_NONE                     'opcode flag set to normal
WRD_HOLD      byte  4,"HOLD"              
COD_HOLD      word  NFW_LIT, -1, @COD_HP, @COD_PLUSSTORE, @COD_HP, @COD_FETCH, @COD_CSTORE
              word  NFW_EXIT
'
'C <#     --             begin numeric conversion
'   PAD HP ! ;          (initialize Hold Pointer)
              word  @WRD_HOLD
              word  of_NONE                     'opcode flag set to normal
WRD_LESSNUM   byte  2,"<#"                
COD_LESSNUM   word  @COD_PAD, @COD_HP, @COD_STORE
              word  NFW_EXIT
'
'P >digit    n -- c      convert to 0..9 A..Z
'   [ HEX ] DUP 9 > 7 AND + 30 + ;
              word  @WRD_LESSNUM
              word  of_NONE                     'opcode flag set to normal
WRD_TODIGIT   byte  6,">DIGIT"            
COD_TODIGIT   word  @COD_DUP, NFW_LIT, 9, @COD_GT, NFW_LIT, 7, @COD_AND
              word  @COD_ADD, NFW_LIT, 48, @COD_ADD
              word  NFW_EXIT
'
'C #      ud1 -- ud2     convert 1 digit of output
'   BASE @ UD/MOD ROT >digit HOLD ;
              word  @WRD_TODIGIT
              word  of_NONE                     'opcode flag set to normal
WRD_NUM       byte  1,"#"              
COD_NUM       word  @COD_BASE, @COD_FETCH, @COD_UDDIVMOD, @COD_ROT, @COD_TODIGIT
              word  @COD_HOLD
              word  NFW_EXIT
'
'C #S    ud1 -- ud2     convert remaining digits
'   BEGIN # 2DUP OR 0= UNTIL ;
              word  @WRD_NUM
              word  of_NONE                     'opcode flag set to normal
WRD_NUMS      byte  2,"#S"              
COD_NUMS      word  @COD_NUM, @COD_2DUP, @COD_OR, @COD_ZEQ, NFW_QBRANCH, @COD_NUMS
              word  NFW_EXIT
'             
'C #>     ud1 -- c-addr u    end conv., get string
'   2DROP HP @ PAD OVER - ;
                word @WRD_NUMS
                word of_NONE                    'opcode flag set to normal
WRD_NUMGREATER  byte  2,"#>"              
COD_NUMGREATER  word @COD_2DROP, @COD_HP, @COD_FETCH, @COD_PAD, @COD_OVER, @COD_SUB
                word NFW_EXIT
'
'C SIGN   n --           add minus sign if n<0
'   0< IF 2D HOLD THEN ;
              word  @WRD_NUMGREATER
              word  of_NONE                     'opcode flag set to normal
WRD_SIGN      byte  4,"SIGN"            
COD_SIGN      word  @COD_ZLT, NFW_QBRANCH, @COD_SIGN1, NFW_LIT, $2D, @COD_HOLD
COD_SIGN1     word  NFW_EXIT
'
'C U.     u --           display u unsigned
'   <# 0 #S #> TYPE SPACE ;
              word  @WRD_SIGN
              word  of_NONE                     'opcode flag set to normal
WRD_UDOT      byte  2,"U."              
COD_UDOT      word  @COD_LESSNUM, @COD_ZERO, @COD_NUMS, @COD_NUMGREATER
              word  @COD_TYPE, @COD_SPACE
              word  NFW_EXIT
'
'P UD.     ud --           display double ud unsigned
'   <# #S #> TYPE SPACE ;
              word  @WRD_UDOT
              word  of_NONE                     'opcode flag set to normal
WRD_UDDOT     byte  3,"UD."              
COD_UDDOT     word  @COD_LESSNUM, @COD_NUMS, @COD_NUMGREATER
              word  @COD_TYPE, @COD_SPACE
              word  NFW_EXIT
'
'C .      n --           display n signed
'   <# DUP ABS 0 #S ROT SIGN #> TYPE SPACE
              word  @WRD_UDDOT
              word  of_NONE                     'opcode flag set to normal
WRD_DOT       byte  1,"."              
COD_DOT       word  @COD_LESSNUM, @COD_DUP, @COD_ABS, @COD_ZERO, @COD_NUMS
              word  @COD_ROT, @COD_SIGN, @COD_NUMGREATER
              word  @COD_TYPE, @COD_SPACE
              word  NFW_EXIT
'
'C DECIMAL ( -- )     set number base to decimal
'    10 BASE !
              word  @WRD_DOT
              word  of_NONE                     'opcode flag set to normal
WRD_DECIMAL   byte  7,"DECIMAL"              
COD_DECIMAL   word  NFW_LIT, 10, @COD_BASE, @COD_STORE
              word  NFW_EXIT
'
'X HEX     --       set number base to hex
'   16 BASE !   
              word  @WRD_DECIMAL
              word  of_NONE                     'opcode flag set to normal
WRD_HEX       byte  3,"HEX"              
COD_HEX       word  NFW_LIT, 16, @COD_BASE, @COD_STORE
              word  NFW_EXIT            
'
'X BINARY     --       set number base to hex
'   2 BASE !   
              word  @WRD_HEX
              word  of_NONE                     'opcode flag set to normal
WRD_BINARY    byte  6,"BINARY"              
COD_BINARY    word  NFW_LIT, 2, @COD_BASE, @COD_STORE
              word  NFW_EXIT            
'
'X D.R    d n --
'Display double d right aligned in a field n characters wide. If the number of characters required to
'display d is greater than n, all digits are displayed with no leading spaces in a field as wide as
'necessary
'   >R SWAP OVER DABS <# #S ROT SIGN #>
'   <R OVER SUB SPACES TYPE            
              word  @WRD_BINARY
              word  of_NONE                     'opcode flag set to normal
WRD_DDOTR     byte  3,"D.R"              
COD_DDOTR     word  @COD_TO_R, @COD_SWAP, @COD_OVER, @COD_DABS, @COD_LESSNUM, @COD_NUMS
              word  @COD_ROT, @COD_SIGN, @COD_NUMGREATER
              word  @COD_R_TO, @COD_OVER, @COD_SUB, @COD_SPACES, @COD_TYPE
              word  NFW_EXIT
'
'X .R    n1 n2 --
'Display single n1 right aligned in a field n2 characters wide. If the number of characters required to
'display n1 is greater than n2, all digits are displayed with no leading spaces in a field as wide as
'necessary
'   >R S>D <R D.R           
              word  @WRD_DDOTR
              word  of_NONE                     'opcode flag set to normal
WRD_DOTR      byte  2,".R"              
COD_DOTR      word  @COD_TO_R, @COD_STOD, @COD_R_TO, @COD_DDOTR
              word  NFW_EXIT
'
'X U.R    n1 n2 --
'Display unsigned single n1 right aligned in a field n2 characters wide. If the number of characters required to
'display n1 is greater than n2, all digits are displayed with no leading spaces in a field as wide as
'necessary.
'   0 SWAP D.R            
              word  @WRD_DOTR
              word  of_NONE                     'opcode flag set to normal
WRD_UDOTR     byte  3,"U.R"              
COD_UDOTR     word  @COD_ZERO, @COD_SWAP, @COD_DDOTR
              word  NFW_EXIT
'
'Display a double number free form
'X D.  d --
'   0 D.R SPACE
              word  @WRD_UDOTR
              word  of_NONE                     'opcode flag set to normal
WRD_DDOT      byte  2,"D."              
COD_DDOT      word  @COD_ZERO, @COD_DDOTR, @COD_SPACE
              word  NFW_EXIT            
'
'            
'FORTH TERMINAL IO
'
'P COM1-IN ( -- c )
'If a character is available from com port, return true. Otherwise, return false.
              word  @WRD_DDOT
              word  of_NONE                     'opcode flag set to normal
WRD_COM1IN    byte  7,"COM1-IN"              
COD_COM1IN    word  NFW_LIT, cmd_COM_IN, @COD_SYSTEM    
              word  NFW_EXIT
'
'Return TRUE if a character is available from the com port 1 input device or FALSE if no data ready
              word  @WRD_COM1IN
              word  of_NONE                     'opcode flag set to normal
WRD_COM1INQ   byte  8,"COM1-IN?"              
COD_COM1INQ   word  NFW_LIT, cmd_COM_CAN_IN, @COD_SYSTEM    
              word  NFW_EXIT
'
'Empty the com port 1 device input buffer
              word  @WRD_COM1INQ
              word  of_NONE                     'opcode flag set to normal
WRD_COM1CLEAR byte  8,"COM1-CLR"              
COD_COM1CLEAR word  NFW_LIT, cmd_COM_CLEAR, @COD_SYSTEM    
              word  NFW_EXIT
'              
'Set com port 1 device baud rate
              word  @WRD_COM1CLEAR
              word  of_NONE                     'opcode flag set to normal
WRD_COM1BAUD  byte  9,"COM1-BAUD"              
COD_COM1BAUD  word  NFW_LIT, cmd_COM_BAUD, @COD_SYSTEM    
              word  NFW_EXIT
'              
'P COM1-OUT ( c -- ) 
'Write the character value (low 8 bits) of the top data stack cell to the first serial port.
              word  @WRD_COM1BAUD
              word  of_NONE                     'opcode flag set to normal
WRD_COM1OUT   byte  8,"COM1-OUT"              
COD_COM1OUT   word  NFW_LIT, cmd_COM_OUT, @COD_SYSTEM    
              word  NFW_EXIT
'
'P COM1-OUT? (  -- f )
'Return TRUE if a character can be sent to the first serial output device or FALSE if no data can be written.
'This is a defered word so it may be reassigned at runtime. See IS.
                word  @WRD_COM1OUT
                word  of_NONE                     'opcode flag set to normal
WRD_COM1OUTQ  byte  9,"COM1-OUT?"              
COD_COM1OUTQ  word  NFW_LIT, cmd_COM_CAN_OUT, @COD_SYSTEM    
                word  NFW_EXIT
'
'Param_SystemFreq                
                word  @WRD_COM1OUTQ
                word  of_NONE                     'opcode flag set to normal
WRD_CLKFREQ     byte  7,"CLKFREQ"              
COD_CLKFREQ     word  @COD_ZERO, @COD_QFETCH    
                word  NFW_EXIT
'
'U AT-XY ( X Y -- ) set the terminal cursor position                
              word  @WRD_CLKFREQ
              word  of_DEFER                   'opcode flag set to defered
WRD_ATXY      byte  5,"AT-XY"              
              word  @COD_2DROP    
              word  NFW_EXIT
'
'A XY ( -- X Y ) read the terminal cursor position                
              word  @WRD_ATXY
              word  of_DEFER                   'opcode flag set to defered
WRD_XY        byte  2,"XY"              
              word  @COD_NOP    
              word  NFW_EXIT
'
'P MAX-XY ( -- X Y ) read the number of terminal rows and columns                
              word  @WRD_XY
              word  of_DEFER                   'opcode flag set to defered
WRD_MAXXY     byte  6,"MAX-XY"              
              word  @COD_NOP    
              word  NFW_EXIT
'
'U PAGE ( -- ) clear the terminal screen                
              word  @WRD_MAXXY
              word  of_DEFER                   'opcode flag set to defered
WRD_PAGE      byte  4,"PAGE"              
              word  @COD_NOP    
              word  NFW_EXIT
'
'P CLR-EOL ( -- ) clear the terminal's current screen line
'                 from the X position to the end of the line.                
              word  @WRD_PAGE
              word  of_DEFER                   'opcode flag set to defered
WRD_CLREOL    byte  7,"CLR-EOL"              
              word  @COD_NOP    
              word  NFW_EXIT
'
'P CURSOR ( f -- ) if f is TRUE terminal cursor is displayed, 
'                  if f is FALSE termianl cursor is hidden.                
              word  @WRD_CLREOL
              word  of_DEFER                   'opcode flag set to defered
WRD_CURSOR    byte  6,"CURSOR"              
              word  @COD_DROP    
              word  NFW_EXIT
'
'C EMIT ( c -- )
'Write the character value (low 8 bits) to the current console.
'This is a defered word so it may be reassigned at runtime. See IS.
'   DEFER "name"  (  --  ) 
              word  @WRD_CURSOR
              word  of_DEFER                   'opcode flag set to defered
WRD_EMIT      byte  4,"EMIT"              
COD_EMIT      word  @COD_com1out    
              word  NFW_EXIT
'
'C EMIT? ( -- f )
'Return TRUE if a character can be sent to the console output device or FALSE if no data can be written.
'This is a defered word so it may be reassigned at runtime.
              word  @WRD_EMIT
              word  of_DEFER                   'opcode flag set to defered
WRD_CANEMIT   byte  5,"EMIT?"              
COD_CANEMIT   word  @COD_com1outq    
              word  NFW_EXIT
'
'C KEY ( -- c )
'If a character is available, return true. Otherwise, return false.
              word  @WRD_CANEMIT
              word  of_DEFER                     'opcode flag set to defered
WRD_KEY       byte  3,"KEY"              
COD_KEY       word  @COD_COM1IN    
              word  NFW_EXIT
'
'C KEY? ( -- f )
'Return TRUE if a character is available from the console input device or FALSE if no data ready
              word  @WRD_KEY
              word  of_DEFER                     'opcode flag set to defered
WRD_ISKEY     byte  4,"KEY?"              
COD_ISKEY     word  @COD_COM1INQ    
              word  NFW_EXIT
'
'C EMPTY-KEYS ( --  )
'Empty the console device input buffer
              word  @WRD_ISKEY
              word  of_DEFER                     'opcode flag set to defered
WRD_CLRKEYS   byte  10,"EMPTY-KEYS"              
COD_CLRKEYS   word  @COD_COM1CLEAR    
              word  NFW_EXIT
'              
'P (cr) ( -- )
'Write the character values for a CR and LF to the console.
              word  @WRD_CLRKEYS
              word  of_NONE                     'opcode flag set to normal
WRD_CRLF      byte  4,"(cr)"              
COD_CRLF      word  NFW_LIT, 13, @COD_EMIT
              word  NFW_LIT, 10, @COD_EMIT
              word  NFW_EXIT
'
'P COM1>CON ( -- )
'Set all console defered words to there COM1 equivelents
              word  @WRD_CRLF
              word  of_NONE                     'opcode flag set to normal
WRD_COM1CON   byte  8,"COM1>CON"              
COD_COM1CON   word  NFW_ALIT, @COD_COM1OUT, NFW_ALIT, @COD_EMIT, @COD_STORE  
              word  NFW_ALIT, @COD_COM1OUTQ, NFW_ALIT, @COD_CANEMIT, @COD_STORE
              word  NFW_ALIT, @COD_COM1IN, NFW_ALIT, @COD_KEY, @COD_STORE  
              word  NFW_ALIT, @COD_COM1INQ, NFW_ALIT, @COD_ISKEY, @COD_STORE
              word  NFW_ALIT, @COD_COM1CLEAR, NFW_ALIT, @COD_CLRKEYS, @COD_STORE
              word  NFW_EXIT

'Write the character values for a CR and LF to the console.
              word  @WRD_COM1CON
              word  of_DEFER                    'opcode flag set to defered
WRD_CR        byte  2,"CR"              
COD_CR        word  @COD_CRLF
              word  NFW_EXIT
'
'Write the space character to the console.
              word  @WRD_CR
              word  of_NONE                     'opcode flag set to normal
WRD_SPACE     byte  5,"SPACE"              
COD_SPACE     word  @COD_BL, @COD_EMIT
              word  NFW_EXIT
'
'P ( n -- )
'Write N backspace characters to the console.
              word  @WRD_SPACE
              word  of_NONE                     'opcode flag set to normal
WRD_BSPACE    byte  9,"BACKSPACE"              
COD_BSPACE    word  @COD_BS, @COD_EMIT 
              word  NFW_EXIT

'P (chars) ( n c -- )  Write n characters c to the console.
              word  @WRD_BSPACE
              word  of_NONE                     'opcode flag set to normal
WRD_REPTCHAR  byte  7,"(chars)"              
COD_REPTCHAR  word  @COD_SWAP
COD_REPTCHAR1 word  @COD_DUP, @COD_ZGT, NFW_QBRANCH, @COD_REPTCHAR2
              word  @COD_SUB1, @COD_OVER, @COD_EMIT
              word  NFW_BRANCH, @COD_REPTCHAR1
COD_REPTCHAR2 word  @COD_2DROP
              word  NFW_EXIT
'
'P BACKSPACES ( n -- )
'Write N backspace characters to the console.
              word  @WRD_REPTCHAR
              word  of_NONE                     'opcode flag set to normal
WRD_BSPACES   byte  10,"BACKSPACES"              
COD_BSPACES   word  @COD_BS, @COD_REPTCHAR 
              word  NFW_EXIT

'C SPACES ( n -- )
'Write N backspace characters to the console.
              word  @WRD_BSPACES
              word  of_NONE                     'opcode flag set to normal
WRD_SPACES    byte  6,"SPACES"              
COD_SPACES    word  @COD_BL, @COD_REPTCHAR 
              word  NFW_EXIT

'P UMIN     u1 u2 -- u      unsigned minimum
'   2DUP U> IF SWAP THEN DROP ;
              word  @WRD_SPACES
              word  of_NONE                     'opcode flag set to normal
WRD_UMIN      byte  4,"UMIN"              
COD_UMIN      word  @COD_2DUP, @COD_UGT, NFW_QBRANCH, @COD_UMIN1, @COD_SWAP 
COD_UMIN1     word  @COD_DROP
              word  NFW_EXIT
'
'P UMAX    u1 u2 -- u       unsigned maximum
'   2DUP U< IF SWAP THEN DROP ;
              word  @WRD_UMIN
              word  of_NONE                     'opcode flag set to normal
WRD_UMAX      byte  4,"UMAX"              
COD_UMAX      word  @COD_2DUP, @COD_ULT, NFW_QBRANCH, @COD_UMAX1, @COD_SWAP 
COD_UMAX1     word  @COD_DROP
              word  NFW_EXIT
'
'C S"       --      compile in-line counted string
'At run time the address and character count are placed on
'the top of the data stack     ( -- c-adr u)
'   COMPILE (S")  [ HEX ]
'   22 WORD C@ 1+ ALIGNED ALLOT ; IMMEDIATE
              word  @WRD_UMAX
              word  of_IMMED                     'opcode flag set to IMMEDIATE
WRD_SQUOTE    byte  2,"S",$22              
COD_SQUOTE    word NFW_LIT, NFW_XSQUOTE, @COD_XTCOMMA
              word NFW_LIT, $22, @COD_WORD, @COD_CFETCH, @COD_ADD1
              word @COD_ALLOT, @COD_ALIGN
              word  NFW_EXIT
'
'X C"       --      compile in-line counted string
'At run time the address of the character count is placed on
'the top of the data stack     ( -- c-adr )
'   COMPILE (S")  [ HEX ]
'   22 WORD C@ 1+ ALIGNED ALLOT ; IMMEDIATE
              word  @WRD_SQUOTE
              word  of_IMMED                     'opcode flag set to IMMEDIATE
WRD_CQUOTE    byte  2,"C",$22              
COD_CQUOTE    word  @COD_SQUOTE
              'NFW_XSQUOTE leaves the string's first char address and character count
              'Character count in memory is the byte before first character.
              'First drop the character count
              word  NFW_ALIT, @COD_DROP, @COD_XTCOMMA
              'Point the address to the character count byte
              word  NFW_ALIT, @COD_SUB1, @COD_XTCOMMA
              word  NFW_EXIT
'
'C ."       --         compile string to print
'   POSTPONE S"  POSTPONE TYPE ; IMMEDIATE
              word  @WRD_CQUOTE
              word  of_IMMED                     'opcode flag set to normal
WRD_DOTQUOTE  byte  2,".",$22
              'see if we are in compile mode
COD_DOTQUOTE  word  @COD_STATE, @COD_FETCH, @COD_ZEQ, NFW_QBRANCH, @COD_DOTQUOTE1
              word  NFW_LIT, $22, @COD_WORD, @COD_COUNT, @COD_TYPE
              word NFW_BRANCH, @COD_DOTQUOTE2
              'we are in compie mode
COD_DOTQUOTE1 word  @COD_SQUOTE
              word  NFW_ALIT, @COD_TYPE, @COD_XTCOMMA 
COD_DOTQUOTE2 word  NFW_EXIT
'
' read a string of characters into buffer
' return length of string in buffer.
              word  @WRD_DOTQUOTE
              word  of_NONE                     'opcode flag set to normal
WRD_ACCEPT    byte  6,"ACCEPT"            
COD_ACCEPT    word  @COD_TO_R, @COD_ZERO
              'Read char from input device.
COD_ACCEPT3   word  @COD_KEY
              'Test if char is a CR character, if it is we are done
              word  @COD_DUP, NFW_LIT, $0D, @COD_NEQ, NFW_QBRANCH, @COD_ACCEPT1
              'Test if we have a control character
              word  @COD_DUP, NFW_LIT, 32, @COD_GTE, NFW_QBRANCH, @COD_ACCEPT2
              'Just a regular character see if we have room for it
              word  @COD_OVER, @COD_R_FETCH, @COD_LT, NFW_QBRANCH, @COD_ACCEPT4
              'Char fits in buffer so emit char and save to buffer
              word  @COD_DUP, @COD_EMIT, @COD_TO_R, @COD_2DUP, @COD_ADD, @COD_R_TO
              word  @COD_SWAP, @COD_CSTORE, @COD_ADD1, NFW_BRANCH, @COD_ACCEPT3
              'No room in buffer drop character and read next char
COD_ACCEPT4   word  @COD_DROP, NFW_BRANCH, @COD_ACCEPT3 
              'We have a control character see if it is a backspace
COD_ACCEPT2   word  @COD_BS, @COD_EQ, NFW_QBRANCH, @COD_ACCEPT3              
              'Backspace character see if we any chars to delete
              word  @COD_DUP, @COD_ZGT, NFW_QBRANCH, @COD_ACCEPT3
              'Emit a backspace and delete last char in buffer by reducing count
              word  @COD_BSPACE, @COD_SUB1, NFW_BRANCH, @COD_ACCEPT3
              'clean stacks and leaving char coount
COD_ACCEPT1   word  @COD_DROP, @COD_NIP, @COD_R_DROP
              word  NFW_EXIT
'              
'Copy a string of characters to the console device
'The TOS contains the number of bytes in the string and
'the second value on the data stack is the address of the first character to copy.
'C TYPE  ( c-addr u -- )
              word  @WRD_ACCEPT
              word  of_NONE                     'opcode flag set to normal
WRD_TYPE      byte  4,"TYPE"              
COD_TYPE      word  @COD_DUP, NFW_QBRANCH, @COD_TYPE3
              word  @COD_ZERO, @COD_XDO
COD_TYPE1     word  @COD_COUNT                  'COUNT same as; DUP C! SWAP 1+ SWAP
              'see if we have a control character
              word  NFW_LIT, 32, @COD_MAX, @COD_EMIT
              word  NFW_XLOOP, @COD_TYPE1, @COD_DUP
              'All done so just exit
COD_TYPE3     word  @COD_2DROP
              word  NFW_EXIT
'
'
'FORTH DICTIONARY MANAGEMENT 

'C HERE    -- addr      returns dictionary ptr
              word  @WRD_TYPE
              word  of_NONE                     'opcode flag set to normal
WRD_HERE      byte  4,"HERE"                 
COD_HERE      word  @COD_DP, @COD_FETCH  
              word  NFW_EXIT
'
'C ALLOT   n --         allocate n bytes in dict
              word  @WRD_HERE
              word  of_NONE                     'opcode flag set to normal
WRD_ALLOT     byte  5,"ALLOT"                
COD_ALLOT     word  @COD_DP, @COD_PLUSSTORE
              'make sure we are not out of memory
              word  @COD_SPFETCH, @COD_HERE, NFW_LIT, 128, @COD_ADD, @COD_ULT
'              word  NFW_QBRANCH, @COD_ALLOT1
              'we ran out of memory, report error and abort 
              word  NFW_XSQUOTE
              byte  13, "out of memory"
              word  @COD_QABORT
COD_ALLOT1    word  NFW_EXIT
'
'C ,     x --           append cell to dict
              word  @WRD_ALLOT
              word  of_NONE                     'opcode flag set to normal
WRD_COMMA     byte  1,","                    
COD_COMMA     word  @COD_HERE, @COD_STORE, @COD_ONE, @COD_CELLS, @COD_ALLOT
              word  NFW_EXIT
'
'C C,    char --        append char to dict
              word  @WRD_COMMA
              word  of_NONE                     'opcode flag set to normal
WRD_CCOMMA    byte  2,"C,"                   
COD_CCOMMA    word  @COD_HERE, @COD_CSTORE, @COD_ONE, @COD_CHARS, @COD_ALLOT
              word  NFW_EXIT
'
'C >BODY     xt -- a-addr      adrs of param field
'In our case xt is also the address of the first data area of the word.           
              word  @WRD_CCOMMA
              word  of_NONE                     'opcode flag set to normal
WRD_TO_BODY   byte  5,">BODY"                
COD_TO_BODY   word  NFW_EXIT
'
'X COMPILE,   xt --         append execution token optimaize if posible
              word  @WRD_TO_BODY
              word  of_NONE                     'opcode flag set to normal
WRD_XTCOMMA   byte  8,"COMPILE,"             
COD_XTCOMMA   word  @COD_XTTOCODE, @COD_COMMA
              word  NFW_EXIT
'
'P EXIT,     --      append hi-level EXIT action
              word  @WRD_XTCOMMA
              word  of_NONE                     'opcode flag set to normal
WRD_EXITCOMMA byte  5,"EXIT,"                 
COD_EXITCOMMA word  NFW_LIT, NFW_EXIT, @COD_COMMA
              word  NFW_EXIT
'
'P BACK,     --      set HERE back one cell by adjusting the DP
              word  @WRD_EXITCOMMA
              word  of_NONE                     'opcode flag set to normal
WRD_BACKCOMMA byte  5,"BACK,"                 
COD_BACKCOMMA word  @COD_HERE, @COD_CELL, @COD_SUB, @COD_DP, @COD_STORE 
              word  NFW_EXIT
'

'FORTH CONTROL STRUCTURES
'
'P ,BRANCH   xt --    append a branch instruction
' xt is the branch operator to use, e.g. qbranch
' or (loop).  It does NOT append the destination
' address.  This is equivalent to ,XT in.
                        word @WRD_BACKCOMMA
                        word of_NONE            'opcode flag set to normal
WRD_COMMABRANCH         byte  7,",BRANCH"
COD_COMMABRANCH         word  @COD_COMMA
                        word NFW_EXIT
'                       
'P ,DEST    dest --        append a branch address
' This appends the given destination address to
' the branch instruction.  On the propeller this is ','
' ...other CPUs may use relative addressing.
              word @WRD_COMMABRANCH
              word of_NONE                      'opcode flag set to normal
WRD_COMMADEST byte 5,",DEST"
COD_COMMADEST word @COD_COMMA
              word NFW_EXIT
'                       
'P !DEST   dest adrs --    change a branch dest'n
' Changes the destination address found at 'adrs'
' to the given 'dest'.  This is '!'
' ...other CPUs may need relative addressing.
              word @WRD_COMMADEST
              word of_NONE                      'opcode flag set to normal
WRD_STOREDEST byte 5,"!DEST"
COD_STOREDEST word @COD_STORE
              word NFW_EXIT
'
'                       
'FORTH INTERPRETER
' 
' Note that NFA>LFA, NFA>CFA, IMMED?, NFA and
' FIND are dependent on the structure of the Forth
' header.  This may be common across many CPUs,
' or it may be different.
'
'C SOURCE (  -- c-adr u )   current input buffer
'c-addr is the address of, and u is the number of characters in, the input buffer.
              word  @WRD_STOREDEST
              word  of_NONE                     'opcode flag set to normal
WRD_SOURCE    byte  6,"SOURCE"             
COD_SOURCE    word  @COD_TSOURCE, @COD_2FETCH
              word  NFW_EXIT
'
'S /STRING  ( c-addr1 u1 n -- c-addr2 u2 )  trim string
'Adjust the character string at c-addr1 by n characters. The resulting character string, specified
'by c-addr2 u2, begins at c-addr1 plus n characters and is u1 minus n characters long.
                word  @WRD_SOURCE
                word  of_NONE                     'opcode flag set to normal
WRD_SLASHSTRING byte  7,"/STRING"             
COD_SLASHSTRING word  @COD_ROT, @COD_OVER, @COD_ADD, @COD_ROT, @COD_ROT, @COD_SUB
                word  NFW_EXIT
'
'P >counted  src n dst --     copy to counted str
'   2DUP C! CHAR+ SWAP CMOVE ;
              word  @WRD_SLASHSTRING
              word  of_NONE                     'opcode flag set to normal
WRD_TOCOUNTED byte  8,">COUNTED"             
COD_TOCOUNTED word  @COD_2DUP, @COD_CSTORE, @COD_CHARPLUS, @COD_SWAP, @COD_CMOVE
              word  NFW_EXIT
'
'P SCAN    c-addr u c -- c-addr' u'                         
'Scan the character string at address "c-addr" with a length of "u" for the
'character "c".
'On Exit" "c-addr'" contains the address of first matching character and "u'"
'contains the number of character remaining in the string 
              word  @WRD_TOCOUNTED
              word  of_NONE                     'opcode flag set to normal
WRD_SCAN      byte  4,"SCAN"             
              'Make sure count <> 0
COD_SCAN      word  @COD_OVER, NFW_QBRANCH, @COD_SCAN_EXIT
              word  @COD_SWAP, @COD_ZERO, @COD_XDO, @COD_SWAP
              'loop and test each character in string ( c adr -- )
COD_SCAN1     word  @COD_2DUP, @COD_CFETCH, @COD_NEQ, NFW_QBRANCH, @COD_SCAN_EQ
              'Character does not match, increment address and try again
              word  @COD_ADD1, NFW_XLOOP, @COD_SCAN1
              'Made to end of string, indicate no character matched
              word  @COD_ZERO, @COD_ROT, NFW_BRANCH, @COD_SCAN_EXIT
              'Found character equal to c so we are done.
COD_SCAN_EQ   word  @COD_R_TO, @COD_R_TO        'Same as "UNLOOP", but Limit and Index are saved on stack
              word  @COD_SWAP, @COD_SUB, @COD_ROT                 
COD_SCAN_EXIT word  @COD_DROP
              word  NFW_EXIT
'
'P SKIP    c-addr u c -- c-addr' u'
'Scan the character string at address "c-addr" with a length of "u" for the
'first character not equal to "c".
'On Exit" "c-addr'" contains the address of first character that does not match "c"
'and "u'"'contains the number of character remaining in the string
              word  @WRD_SCAN
              word  of_NONE                     'opcode flag set to normal
WRD_SKIP      byte  4,"SKIP"             
              'Make sure count <> 0
COD_SKIP      word  @COD_OVER, NFW_QBRANCH, @COD_SKIP_EXIT
              word  @COD_SWAP, @COD_ZERO, @COD_XDO, @COD_SWAP
              'loop and test each character in string ( c adr -- )
COD_SKIP1     word  @COD_2DUP, @COD_CFETCH, @COD_EQ, NFW_QBRANCH, @COD_SKIP_NEQ
              'Character matches, increment address and try again
              word  @COD_ADD1, NFW_XLOOP, @COD_SKIP1
              'Made to end of string, indicate all characters matched
              word  @COD_ZERO, @COD_ROT, NFW_BRANCH, @COD_SKIP_EXIT
              'Found character <> c so we are done.
COD_SKIP_NEQ  word  @COD_R_TO, @COD_R_TO        'Same as "UNLOOP", but Limit and Index are saved on stack
              word  @COD_SWAP, @COD_SUB, @COD_ROT                 
COD_SKIP_EXIT word  @COD_DROP
              word  NFW_EXIT
'
'P S= c-addr1 c-addr2 u -- n string compare
'  0 ?DO 2DUP C@ SWAP C@ - ?DUP 
'  IF NIP NIP NEGATE UNLOOP EXIT 
'  ELSE $00010001. D+ 
'  THEN LOOP 2DROP 0 ; 
'             n<0: s1<s2, n=0: s1=s2, n>0: s1>s2
              word  @WRD_SKIP
              word  of_NONE                     'opcode flag set to normal
              'Make sure count <> 0
WRD_STEQ      byte  2,"S="
              ' string length is zero report a match and exit
COD_STEQ      word  @COD_DUP, NFW_QBRANCH, @COD_STEQ3
              'Ready do loop index and limit values
              word  @COD_ZERO, @COD_XDO
              'Loop through string comparing each character
COD_STEQ1     word  @COD_OVER, @COD_CFETCH, @COD_OVER, @COD_CFETCH, @COD_SUB 
              word  @COD_IFDUP, @COD_ZEQ, NFW_QBRANCH, @COD_STEQ2
              'We have a match try next character if any
              word  NFW_DLIT, 1, 1, @COD_DADD, NFW_XLOOP, @COD_STEQ1
              'Indicate we have a match and exit
              word  @COD_ZERO, NFW_BRANCH, @COD_STEQ3 
              'We failed to match report which is larger and exit
COD_STEQ2     word  @COD_UNLOOP         
COD_STEQ3     word  @COD_NIP, @COD_NIP
              word  NFW_EXIT
'
'
'C WORD   char -- c-addr   word delim'd by char
'Copy next word delimited by char from input buffer
'to dictionary memory pointed to by HERE.
'The word is stored as a counted string. The
'address of the counted string is returned.
'The string is padded by a space that is not counted in the length byte.
'   DUP  SOURCE >IN @ /STRING   -- c c adr n
'   DUP >R   ROT SKIP           -- c adr' n'
'   OVER >R  ROT SCAN           -- adr" n"
'   DUP IF CHAR- THEN        skip trailing delim.
'   R> R> ROT -   >IN +!        update >IN offset
'   TUCK -                      -- adr' N
'   HERE >counted               --
'   HERE                        -- a
'   BL OVER COUNT + C! ;    append trailing blank
              word  @WRD_STEQ
              word  of_NONE                     'opcode flag set to normal
WRD_WORD      byte  4,"WORD"              
COD_WORD      word  @COD_DUP, @COD_SOURCE, @COD_TOIN, @COD_FETCH, @COD_SLASHSTRING
              word  @COD_DUP, @COD_TO_R, @COD_ROT, @COD_SKIP
              word  @COD_OVER, @COD_TO_R, @COD_ROT, @COD_SCAN
              word  @COD_DUP, NFW_QBRANCH, @COD_WORD1, @COD_SUB1
COD_WORD1     word  @COD_R_TO, @COD_R_TO, @COD_ROT, @COD_SUB, @COD_TOIN, @COD_PLUSSTORE
              word  @COD_TUCK, @COD_SUB
              word  @COD_HERE, @COD_TOCOUNTED, @COD_HERE
              word  @COD_BL, @COD_OVER, @COD_COUNT, @COD_ADD, @COD_CSTORE
              word  NFW_EXIT
'

'X PARSE   char -- c-addr u   word delim'd by char
'Find the end of the remaining input buffer delimited by char.
'The word returns the start address of the string and the character count.
              word  @WRD_WORD
              word  of_NONE                     'opcode flag set to normal
WRD_PARSE     byte  5,"PARSE"
              'Get the address of the unused portion of the input buffer
COD_PARSE     word  @COD_SOURCE, @COD_TOIN, @COD_FETCH, @COD_SLASHSTRING
              'We are pointing to the address of the string start.
              'If the count is 0 we are done
              word  @COD_DUP, NFW_QBRANCH, @COD_PARSE1
              ' we have; char c-adr u on the top of the stack
              'save copy of string start address and total chars in string to return stack
              word  @COD_2DUP, @COD_2TO_R
              'Move character to top of stack. See if delimiter is in string
              word  @COD_ROT, @COD_SCAN
              'Scan left c-adr u on top of stack, we only need u
              word  @COD_NIP
              'Get string start address and length from return stack
              word  @COD_2R_TO, @COD_ROT
              'At this point we have; c-adr u u' on the stack
              word  @COD_SUB
              'Adjust >IN to reflect characters used
              word  @COD_DUP, @COD_TOIN, @COD_PLUSSTORE
              word  NFW_BRANCH, @COD_PARSE2
              'input buffer empty so dump char and return
COD_PARSE1    word  @COD_ROT, @COD_DROP
COD_PARSE2    word  NFW_EXIT


'P NFA>CFA   nfa -- cfa    name adr -> code field
'Convert the name field address to the code field address.
'The TOS contains the start address of a forth word which points
'to the length byte of the name string. The TOS address value is adjusted
'to point to the link field address. 
              word  @WRD_PARSE
              word  of_NONE                     'opcode flag set to normal
WRD_NTOC      byte  7,"NFA>CFA"              
COD_NTOC      word  @COD_DUP, @COD_CFETCH, NFW_LIT, $7F, @COD_AND, @COD_ADD, @COD_ADD1
              'Align the address on a word (16bit) boundry
              word  @COD_DUP, @COD_ONE, @COD_AND, @COD_ADD    
              word  NFW_EXIT
'              
'P NFA>LFA   nfa -- lfa    name adr -> link field
'Convert the name field address to the link field address.
'The TOS contains the start address of a forth word which points
'to the length byte of the name string. The TOS address value is adjusted
'to point to the link field address. 
              word  @WRD_NTOC
              word  of_NONE                     'opcode flag set to normal
WRD_NTOL      byte  7,"NFA>LFA"              
COD_NTOL      word  NFW_LIT, 4, @COD_SUB
              word  NFW_EXIT
'
'P NFA>FFA   nfa -- ffa    name adr -> flag field
'Convert the name field address to the flag field address.
'The TOS contains the start address of a forth word which points
'to the length byte of the name string. The TOS address value is adjusted
'to point to the link field address. 
              word  @WRD_NTOL
              word  of_NONE                     'opcode flag set to normal
WRD_NTOF      byte  7,"NFA>FFA"              
COD_NTOF      word  @COD_SUB2
              word  NFW_EXIT
'
'P OPFLAG?   (nfa FLAG -- f)   f = true if flag is set or false if flag is clear
              word  @WRD_NTOF
              word  of_NONE                     'opcode flag set to normal
WRD_OPFLAGQ   byte  7,"OPFLAG?"
COD_OPFLAGQ   word  @COD_SWAP, @COD_NTOF, @COD_FETCH, @COD_AND
              word  @COD_ZNE
              word  NFW_EXIT

'P IMMED?    nfa -- f      fetch immediate flag
'   1- C@ ;                     nonzero if immed
              word  @WRD_OPFLAGQ
              word  of_NONE                     'opcode flag set to normal
WRD_IMMEDQ    byte  6,"IMMED?"
COD_IMMEDQ    word  NFW_LIT, of_IMMED, @COD_OPFLAGQ
              word  NFW_EXIT
'
'P VALUE?    nfa -- f      fetch VALUE flag
'   1- C@ ;                     nonzero if name is a reference to value
              word  @WRD_IMMEDQ
              word  of_NONE                     'opcode flag set to normal
WRD_VALUEQ    byte  6,"VALUE?"
COD_VALUEQ    word  NFW_LIT, of_VALUE, @COD_OPFLAGQ
              word  NFW_EXIT
'              
'P DEFER?    nfa -- f      fetch DEFER flag
'   1- C@ ;                     nonzero if name is a reference to a defered word
              word  @WRD_VALUEQ
              word  of_NONE                     'opcode flag set to normal
WRD_DEFERQ    byte  6,"DEFER?"
COD_DEFERQ    word  NFW_LIT, of_DEFER, @COD_OPFLAGQ
              word  NFW_EXIT
'              
'P NFA   c-addr -- nfa | 0    0 = if not found
'   LATEST @ BEGIN             -- a nfa
'       2DUP OVER C@ CHAR+     -- a nfa a nfa n+1
'       S=                     -- a nfa f
'       DUP IF
'           DROP
'           NFA>LFA @ DUP      -- a link link
'       THEN
'   0= UNTIL                   -- a nfa  OR  a 0
'   DUP IF
'       NIP DUP NFA>CFA        -- nfa xt
'       SWAP IMMED?            -- xt iflag
'       0= 1 OR                -- xt 1/-1
'   THEN ;
              word  @WRD_DEFERQ
              word  of_NONE                     'opcode flag set to normal
WRD_NFA       byte  3,"NFA"
COD_NFA       word  @COD_LATEST, @COD_FETCH
COD_NFA1      word  @COD_2DUP, @COD_OVER, @COD_CFETCH, @COD_CHARPLUS
              'Comapre the forth word name with the string at c-adr starting with length byte
              'Since we only care about exact matches (S= returns zero) it is ok to include
              'the string length in the match.
              word  @COD_STEQ, @COD_DUP, NFW_QBRANCH, @COD_NFA2
              'No match read prior word address and if <> 0 check again
              word  @COD_DROP, @COD_NTOL, @COD_FETCH, @COD_DUP
              'See if the link address = 0, if not try again
              word  @COD_ZEQ, NFW_QBRANCH, @COD_NFA1
              'Prior word link is = 0 so all done, no match
              word  @COD_NIP, NFW_BRANCH, @COD_NFA3
              'String matched code thread name, Return name address
COD_NFA2      word  @COD_DROP, @COD_NIP
COD_NFA3      word  NFW_EXIT
'
'              
'C FIND   c-addr -- c-addr 0   if not found
'C                  xt  1      if immediate
'C                  xt -1      if "normal"
'
              word  @WRD_NFA
              word  of_NONE                     'opcode flag set to normal
WRD_FIND      byte  4,"FIND"
COD_FIND      word  @COD_DUP, @COD_NFA, @COD_DUP, NFW_QBRANCH, @COD_FIND1
              'Word name at c-addr was found. Drop name c-addr and
              'check immed flag.
              word  @COD_NIP, @COD_DUP, @COD_NTOC, @COD_SWAP
              word  @COD_IMMEDQ, @COD_ZEQ, @COD_ONE, @COD_OR 
COD_FIND1     word  NFW_EXIT
'
'
'P LIT, ( u -- )
'Compile a literal whose value is on the data stack
              word  @WRD_FIND
              word  of_NONE                     'opcode flag set to normal
WRD_LITCOMMA  byte  4, "LIT,"
COD_LITCOMMA  word  NFW_LIT, NFW_LIT, @COD_XTCOMMA, @COD_COMMA
              word  NFW_EXIT
'              
'C LITERAL  x --        append numeric literal
'   STATE @ IF ['] LIT ,XT , THEN ; IMMEDIATE
' This tests STATE so that it can also be used
' interpretively.  (ANSI doesn't require this.)
              word  @WRD_LITCOMMA
              word  of_IMMED                     'opcode flag set to immediate
WRD_LITERAL   byte  7, "LITERAL"
COD_LITERAL   word  @COD_STATE, @COD_FETCH, NFW_QBRANCH, @COD_LITERAL1
              word  @COD_LITCOMMA
COD_LITERAL1  word  NFW_EXIT
'
'D LITERAL  x --        append double precision numeric literal
'   STATE @ IF ['] SWAP, LIT ,XT , LIT ,XT , THEN ; IMMEDIATE
' This tests STATE so that it can also be used
' interpretively.  (ANSI doesn't require this.)
              word  @WRD_LITERAL
              word  of_IMMED                     'opcode flag set to immediate
WRD_DLITERAL  byte  8, "2LITERAL"
COD_DLITERAL  word  @COD_STATE, @COD_FETCH, NFW_QBRANCH, @COD_DLITERAL1
              word  @COD_SWAP, NFW_LIT, NFW_DLIT, @COD_XTCOMMA, @COD_COMMA, @COD_COMMA
COD_DLITERAL1 word  NFW_EXIT
'
'P DIGIT?   c -- n -1   if c is a valid digit
'P            -- x  0   otherwise
'   [ HEX ] DUP 39 > 100 AND +     silly looking
'   DUP 140 > 107 AND -   30 -     but it works!
'   DUP BASE @ U< ;
              word  @WRD_DLITERAL
              word  of_NONE                     'opcode flag set to normal
WRD_DIGITQ    byte  6, "DIGIT?"
COD_DIGITQ    word  @COD_DUP, NFW_LIT, $39, @COD_GT, NFW_LIT, $100, @COD_AND, @COD_ADD
              word  @COD_DUP, NFW_LIT, $140, @COD_GT, NFW_LIT, $107, @COD_AND
              word  @COD_SUB, NFW_LIT, $30, @COD_SUB
              word  @COD_DUP, @COD_BASE, @COD_FETCH, @COD_ULT
              word  NFW_EXIT
'
'P ?CHAR adr n c -- adr' n' f
'Returns true in "f" if "c" is first character in string,
'Remove first string character if a match is found.
' 
              word  @WRD_DIGITQ
              word  of_NONE                     'opcode flag set to normal
WRD_QCHAR     byte  5, "?CHAR"
COD_QCHAR     word  @COD_TO_R, @COD_OVER, @COD_CFETCH, @COD_R_TO, @COD_EQ
              word  @COD_DUP, NFW_QBRANCH, @COD_QCHAR1
              'string starts with char c
              word  @COD_TO_R, @COD_ONE, @COD_SLASHSTRING, @COD_R_TO
COD_QCHAR1    word  NFW_EXIT
'


'P ?SIGN   adr n -- adr' n' f  get optional sign
'P  advance adr/n if sign; return NZ if negative
'   OVER C@                 -- adr n c
'   2C - DUP ABS 1 = AND    -- +=-1, -=+1, else 0
'   DUP IF 1+               -- +=0, -=+2
'       >R 1 /STRING R>     -- adr' n' f
'   THEN ;
              word  @WRD_QCHAR
              word  of_NONE                     'opcode flag set to normal
WRD_QSIGN     byte  5, "?SIGN"
COD_QSIGN     word  NFW_LIT, $2D, @COD_QCHAR
              word  NFW_EXIT
'
'P ?BASE   adr n -- adr' n'  see if optional base indicator character '$ # %' is at start of number string. 
'If the string starts with a valid base representation character advance adr/n
'to consume base character then set BASE variable to match $ = 4, # = 10, and % = 2.
              word  @WRD_QSIGN
              word  of_NONE                     'opcode flag set to normal
WRD_QBASE     byte  5, "?BASE"
COD_QBASE     word  NFW_LIT, $24, @COD_QCHAR, NFW_QBRANCH, @COD_QBASE2
              word  @COD_HEX, NFW_BRANCH, @COD_QBASE1
COD_QBASE2    word  NFW_LIT, $25, @COD_QCHAR, NFW_QBRANCH, @COD_QBASE3
              word  @COD_BINARY
COD_QBASE3    word  NFW_LIT, $23, @COD_QCHAR, NFW_QBRANCH, @COD_QBASE1
              word  @COD_DECIMAL
COD_QBASE1    word  NFW_EXIT
'
'P ?DOUBLE   c-adr n -- c-adr n' f  get optional double indicator "."
'P  See if the last character of the number string is a period indicating a double
'P if double return f = -1 and redunce n by 1. If single cell value return f = 0, n stay the same
              word  @WRD_QBASE
              word  of_NONE                     'opcode flag set to normal
WRD_QDBL      byte  7, "?DOUBLE"
COD_QDBL      word  @COD_2DUP, @COD_ADD, @COD_SUB1
              word  @COD_CFETCH, NFW_LIT, $2E   
              word  @COD_EQ, @COD_DUP, @COD_ROT
              word  @COD_ADD, @COD_SWAP
              word  NFW_EXIT
'
'C >NUMBER  ud adr u -- ud' adr' u'
'C                      convert string to number
'   BEGIN
'   DUP WHILE
'       OVER C@ DIGIT?
'       0= IF DROP EXIT THEN
'       >R 2SWAP BASE @ UD*
'       R> M+ 2SWAP
'       1 /STRING
'   REPEAT ;
              word  @WRD_QDBL
              word  of_NONE                     'opcode flag set to normal
WRD_TONUMBER  byte  7, ">NUMBER"
COD_TONUMBER  word  @COD_DUP, NFW_QBRANCH, @COD_TONUM3
              word  @COD_OVER, @COD_CFETCH, @COD_DIGITQ
              word  @COD_ZEQ, NFW_QBRANCH, @COD_TONUM2, @COD_DROP, NFW_BRANCH, @COD_TONUM3
COD_TONUM2    word  @COD_TO_R, @COD_2SWAP, @COD_BASE, @COD_FETCH, @COD_UDMUL
              word  @COD_R_TO,  @COD_MADD, @COD_2SWAP
              word  @COD_ONE, @COD_SLASHSTRING, NFW_BRANCH, @COD_TONUMBER
COD_TONUM3    word  NFW_EXIT
'
'P ?DNUMBER c-addr -- d -1      valid single number
'                     d  1      valid double number
'P                 -- c-addr 0  if convert error
'Convert a string to a a double cell value and report if string was
'a double value (1) or a single value (-1) or an invalid string (0).
'
'
              word  @WRD_TONUMBER
              word  of_NONE                     'opcode flag set to normal
WRD_QDNUMBER  byte  8, "?DNUMBER"
              'Save the current base number value incase we change it
COD_QDNUMBER  word  @COD_BASE, @COD_FETCH, @COD_TO_R
              word  @COD_DUP, @COD_ZERO, @COD_DUP, @COD_ROT, @COD_COUNT
              word  @COD_QSIGN, @COD_TO_R, @COD_QDBL, @COD_TO_R
              'See if string starts with a BASE indicator character '$ % #' if so change base
              word  @COD_QBASE
              'Returned length of token string length = 0, number done
              word  @COD_TONUMBER, NFW_QBRANCH, @COD_QDNUM1
              'Not a vaid number, cleanup stack and leave c-ard, return 0 at top of stack
              'Remove sign and double flags from return stack
              word  @COD_R_TO, @COD_R_TO, @COD_2DROP
              'Remove double from stack and report error
              word  @COD_2DROP, @COD_DROP, @COD_ZERO
              word  NFW_BRANCH, @COD_QDNUM3
              'We have a valid number see if its a double
COD_QDNUM1    word  @COD_R_TO, NFW_QBRANCH, @COD_QDNUM4
              'we have a double cell value, remove n and c-adr
              word  @COD_DROP, @COD_ROT, @COD_DROP
              'Indicate we have a double value
COD_QDNUM5    word  @COD_R_TO, NFW_QBRANCH, @COD_QDNUM6, @COD_DNEGATE
              'Indicate we have a double value
COD_QDNUM6    word  @COD_ONE, NFW_BRANCH, @COD_QDNUM3
              'We have a single cell value    
COD_QDNUM4    word  @COD_2DROP, @COD_NIP, @COD_R_TO, NFW_QBRANCH, @COD_QDNUM2, @COD_NEGATE
COD_QDNUM2    word  @COD_STOD, NFW_LIT, -1
COD_QDNUM3    word  @COD_R_TO, @COD_BASE, @COD_STORE
              word  NFW_EXIT
'
'P XT>NATIVE  xt -- xt
'             xt -- op
'Convert an execution address to a native opcode.
'If the execution token address points to a native
'code execution thread the address is converted to
'a virtual CPU opcode. If the execution token address
'is not pointing to a native code thread the xt on the
'stack is not changed. This is a factor for INTERPRET.
              word  @WRD_QDNUMBER
              word  of_NONE                     'opcode flag set to normal
WRD_XTTOCODE  byte  9, "XT>NATIVE"
COD_XTTOCODE  word  @COD_DUP, NFW_ALIT, @WordTableStart, @COD_GTE, NFW_QBRANCH, @COD_XTTOCODE1
              word  @COD_DUP, NFW_ALIT, @COD_NOP, @COD_LTE, NFW_QBRANCH, @COD_XTTOCODE1
                    'Code address is in range of native codes so read opcode
              word  @COD_FETCH
COD_XTTOCODE1 word  NFW_EXIT
'    
'P INTERPRET    i*x c-addr u -- j*x
'P                      interpret given buffer
' This is a common factor of EVALUATE and QUIT.
' ref. dpANS-6, 3.4 The Forth Text Interpreter
'   'SOURCE 2!  0 >IN !
'   BEGIN
'   BL WORD DUP C@ WHILE        -- textadr
'       FIND                    -- a 0/1/-1
'       ?DUP IF                 -- xt 1/-1
'           1+ STATE @ 0= OR    immed or interp?
'           IF EXECUTE ELSE ,XT THEN
'       ELSE                    -- textadr
'           ?DNUMBER
'           IF POSTPONE LITERAL     converted ok
'           ELSE COUNT TYPE 3F EMIT CR ABORT  err
'           THEN
'       THEN
'   REPEAT DROP ;
              word  @WRD_XTTOCODE
              word  of_NONE                     'opcode flag set to normal
WRD_INTER     byte  9, "INTERPRET"
COD_INTER     word  @COD_TSOURCE, @COD_2STORE, @COD_ZERO, @COD_TOIN, @COD_STORE
COD_INTER1    word  @COD_BL, @COD_WORD, @COD_DUP, @COD_CFETCH, NFW_QBRANCH, @COD_INTER6
              word  @COD_FIND, @COD_IFDUP, NFW_QBRANCH,  @COD_INTER3
              word  @COD_ADD1, @COD_STATE, @COD_FETCH, @COD_ZEQ, @COD_OR
              word  NFW_QBRANCH, @COD_INTER2
              word  @COD_EXECUTE, NFW_BRANCH, @COD_INTER1
COD_INTER2    word  @COD_XTCOMMA
              word  NFW_BRANCH, @COD_INTER1
COD_INTER3    word  @COD_QDNUMBER, @COD_DUP, NFW_QBRANCH, @COD_INTER5
              'Have a valid number see if it is a double cell value
              word  @COD_ONE, @COD_EQ, NFW_QBRANCH, @COD_INTER4
              'We have a double cell value
              word  @COD_DLITERAL, NFW_BRANCH, @COD_INTER1
              'We have a single value number, drop the high cell
COD_INTER4    word  @COD_DROP, @COD_LITERAL, NFW_BRANCH, @COD_INTER1
              'Error parsing current token, show bad token then "?" and abort
COD_INTER5    word  @COD_DROP, @COD_COUNT, @COD_TRUE, @COD_ROT, @COD_ROT, @COD_QABORT
COD_INTER6    word  @COD_DROP
              word  NFW_EXIT
'
'C EVALUATE   i*x c-addr u -- j*x  interprt string
'   'SOURCE 2@ >R >R  >IN @ >R
'   INTERPRET
'   R> >IN !  R> R> 'SOURCE 2! ;
              word  @WRD_INTER
              word  of_NONE                     'opcode flag set to normal
WRD_EVALUATE  byte  8,"EVALUATE"
COD_EVALUATE  word  @COD_TRUE, @COD_TO_SOURCE_ID
              word  @COD_SOURCE, @COD_2TO_R
              word  @COD_TOIN, @COD_FETCH, @COD_TO_R, @COD_INTER
              word  @COD_R_TO, @COD_TOIN, @COD_STORE, @COD_2R_TO
              word  @COD_TSOURCE, @COD_2STORE
              'Block extensions require BLK set to 0
              word  @COD_ZERO, @COD_BLK, @COD_STORE
              'indicate we are no longer evaluating a string
              word  @COD_FALSE, @COD_TO_SOURCE_ID
              word  NFW_EXIT
'
'C QUIT      --    R: i*x --    interpret from kbd
'   L0 LP !  R0 RP!   0 STATE !
'   BEGIN
'       TIB DUP TIBSIZE ACCEPT  SPACE
'       INTERPRET
'       STATE @ 0= IF CR ." OK" THEN
'   AGAIN ;
              word  @WRD_EVALUATE
              word  of_NONE                     'opcode flag set to normal
WRD_QUIT      byte  4,"QUIT"
COD_QUIT      word  @COD_L0, @COD_LP, @COD_STORE
              word  @COD_LEFTBRAC
              word  @COD_ZERO, @COD_BLK, @COD_STORE
              word  @COD_S0, @COD_SPSTORE
             'indicate we are no longer evaluating a string
              word  @COD_FALSE, @COD_TO_SOURCE_ID
COD_QUIT1     word  @COD_TIB, @COD_DUP, @COD_TIBSIZE, @COD_ACCEPT, @COD_SPACE
              word  @COD_INTER
              word  @COD_STATE, @COD_FETCH, @COD_ZEQ, NFW_QBRANCH, @COD_QUIT1
              word  NFW_XSQUOTE
              byte  3," OK"
              word  @COD_TYPE
              word  @COD_CR
              word  NFW_BRANCH, @COD_QUIT1       'QUIT never exits....
              word  NFW_EXIT
'
'X SAVE-INPUT  ( -- u1 u2 .. n ) Save the input state
              word  @WRD_QUIT
              word  of_NONE                     'opcode flag set to normal
WRD_SAVEINP   byte  10,"SAVE-INPUT"
              'Save the current input source state LENGTH, SOURCE, IN, SOURCE-ID
COD_SAVEINP   word  @COD_SOURCE                 'Pushes two values to the stack
              word  @COD_TOIN, @COD_FETCH
              word  @COD_SOURCE_ID
              word  @COD_BLK, @COD_FETCH
              word  NFW_LIT, 5
              word  NFW_EXIT
'
'X RESTORE-INPUT  ( u1 u2 .. n -- ) Restore the input state
              word  @WRD_SAVEINP
              word  of_NONE                     'opcode flag set to normal
WRD_RESTINP    byte  13,"RESTORE-INPUT"
              'Restore the current input source state LENGTH, SOURCE, IN, SOURCE-ID
COD_RESTINP   word  NFW_LIT, 5, @COD_NEQ
              word  NFW_XSQUOTE
              byte  19, "Invalid input state"
              word  @COD_QABORT
              'same number of params so try to restore
COD_RESTINP1  word  @COD_TO_R                   'Save source block number to return stack
              word  @COD_TO_SOURCE_ID           'Restore the input type (is EVALUATE)
              word  @COD_TOIN, @COD_STORE       'Restore the source input character index
              word  @COD_TSOURCE, @COD_2STORE   'Restore the source input buffer address and length
              'If the input source was a device block, make sure
              'we reload the block and reset source to the buffer address
              word  @COD_R_TO, @COD_DUP, @COD_BLK, @COD_STORE
              word  @COD_ZNE, NFW_QBRANCH, @COD_RESTINP2
              'Need to load the prior block buffer
              word  @COD_BLK, @COD_FETCH, @COD_BLOCK
              word  @COD_DROP
COD_RESTINP2  word  NFW_EXIT
'


'P (load)  ( u -- )
              word  @WRD_RESTINP
              word  of_NONE                     'opcode flag set to normal
WRD_DOLOAD    byte  6,"(load)"
              'Make sure the block number > 0
COD_DOLOAD    word  @COD_IFDUP, @COD_ZEQ
              word  NFW_XSQUOTE
              byte  20, "Can not load block 0"
              word  @COD_QABORT
              'Load the BLOCK and save the block number to BLK
              word  @COD_DUP, @COD_BLK, @COD_STORE, @COD_BLOCK
              'Interpert the block
              word  NFW_LIT, 1024, @COD_INTER
              'Restore the prior input source state from return stack
              'All done just exit
              word  NFW_EXIT
'
'B LOAD  ( u -- )
              word  @WRD_DOLOAD
              word  of_NONE                     'opcode flag set to normal
WRD_LOAD      byte  4,"LOAD"
             'Save the current input source state LENGTH, SOURCE, IN, SOURCE-ID, BLK to return stack'
COD_LOAD      word  @COD_TO_R, @COD_SAVEINP, @COD_R_TO
              word  @COD_DOLOAD
              'Restore the prior input source state from return stack
              word  @COD_RESTINP
              'All done just exit
              word  NFW_EXIT

'B THRU ( u1 u2 -- ) load blocks u1 through u2
              word  @WRD_LOAD
              word  of_NONE                             'opcode flag set to normal
WRD_THRU      byte  4,"THRU"
              'Save the current input source state LENGTH, SOURCE, IN, SOURCE-ID, BLK to return stack'
              word  @COD_CR, @COD_2TO_R, @COD_SAVEINP, @COD_2R_TO
              word  @COD_ADD1, @COD_SWAP, @COD_XDO
              'try to load the block
COD_THRU1     word  @COD_I, @COD_DOLOAD
              word  NFW_XLOOP, @COD_THRU1
              'Restore the prior input source state from return stack
              word  @COD_RESTINP
              word  NFW_EXIT
'
'P  :ABORT        --      remove the last definition if it is hidden
              word  @WRD_THRU
              word  of_NONE                              'opcode flag set to normal
WRD_COMPABORT byte  6,":ABORT"
              'Turn off compile mode
COD_COMPABORT word   @COD_LEFTBRAC
              'See if the last word finished compiling
              word  @COD_HIDDENQ, NFW_QBRANCH, @COD_COMPABOR1
              'word is hidden, remove it
              word  @COD_LATEST, @COD_FETCH, @COD_NTOL, @COD_DUP
              word  @COD_DP, @COD_STORE, @COD_FETCH, @COD_LATEST, @COD_STORE
COD_COMPABOR1 word  NFW_EXIT
'
'C ABORT    i*x --   R: j*x --   clear stk & QUIT
'   S0 SP!  QUIT ;
              word  @WRD_COMPABORT
              word  of_NONE                             'opcode flag set to normal
WRD_ABORT     byte  5,"ABORT"
COD_ABORT     word  @COD_S0, @COD_SPSTORE               'Clear the data stack
              word  @COD_R0, @COD_RPSTORE               'Clear the return stack
              word  @COD_L0, @COD_LP, @COD_STORE        'Clear the leave (control) stack
              word  @COD_COMPABORT, @COD_CLRKEYS        'If we are compiling undo CREATE, Empty input buffer
              word  @COD_DECIMAL                        'Restore BASE to base 10
              word  NFW_BRANCH, @COD_QUIT               'QUIT never returns
              word  NFW_EXIT
'
'P ?ABORT   f c-addr u --      abort & print msg
'   ROT IF TYPE ABORT THEN 2DROP ;
              word  @WRD_ABORT
              word  of_NONE                             'opcode flag set to normal
WRD_QABORT    byte  6,"?ABORT"
COD_QABORT    word  @COD_ROT, NFW_QBRANCH, @COD_QABORT1
              word  @COD_SPACE, @COD_TYPE, NFW_LIT, $3F, @COD_EMIT, @COD_CR
              word  @COD_ABORT
COD_QABORT1   word  @COD_2DROP
              word  NFW_EXIT
'
'C ABORT"  i*x 0  -- i*x   R: j*x -- j*x  x1=0
'C         i*x x1 --       R: j*x --      x1<>0
              word  @WRD_QABORT
              word  of_IMMED                             'opcode flag set to immediate
WRD_ABORTQ    byte  6,"ABORT", $22
COD_ABORTQ    word  @COD_SQUOTE
              word  NFW_ALIT, @COD_QABORT, @COD_XTCOMMA
              word  NFW_EXIT
'
'C '    -- xt           find word in dictionary
              word  @WRD_ABORTQ
              word  of_NONE                             'opcode flag set to none
WRD_TICK      byte  1, $27
COD_TICK      word  @COD_BL, @COD_WORD, @COD_DUP
              'see if we are at end of line, word length = 0
              word  @COD_CFETCH, @COD_ZEQ
              'end of input buffer and no valid name, report error and fail
              word  NFW_XSQUOTE
              byte  13, "name expected"
              word  @COD_QABORT
              'we have a name see if it exists in the dictionary              
COD_TICK2     word  @COD_FIND, @COD_ZEQ
              'If we failed to find the display failed word then abort
              word  NFW_QBRANCH, @COD_TICK1
              'Display error then abort
COD_TICK3     word  @COD_COUNT, @COD_TYPE, NFW_LIT, $3F, @COD_EMIT, @COD_CR
              word  @COD_ABORT
COD_TICK1     word  NFW_EXIT
'
'C CHAR   -- char           parse ASCII character
'   BL WORD 1+ C@ ;
              word  @WRD_TICK
              word  of_NONE                             'opcode flag set to normal
WRD_CHAR      byte  4, "CHAR"
COD_CHAR      word  @COD_BL, @COD_WORD, @COD_ADD1, @COD_CFETCH
              word  NFW_EXIT


'C [CHAR]   --          compile character literal
'   CHAR  ['] LIT ,XT  , ; IMMEDIATE
              word  @WRD_CHAR
              word  of_IMMED                             'opcode flag set to immediate
WRD_BRACCHAR  byte  6,"[CHAR]"
COD_BRACCHAR  word  @COD_CHAR
              word  @COD_LITCOMMA
              word  NFW_EXIT
'
'C (    --                     skip input until )
'   [ HEX ] 29 WORD DROP ; IMMEDIATE
              word  @WRD_BRACCHAR
              word  of_IMMED                             'opcode flag set to immediate
WRD_PAREN     byte  1,"("
COD_PAREN     word  NFW_LIT, $29, @COD_WORD, @COD_DROP
              word  NFW_EXIT
'
'X .(    --                     skip input until )
              word  @WRD_PAREN
              word  of_IMMED                             'opcode flag set to immediate
WRD_DOTPAREN  byte  2,".("
COD_DOTPAREN  word  NFW_LIT, $29, @COD_WORD, @COD_COUNT, @COD_TYPE, @COD_CR
              word  NFW_EXIT
'
'X \    --                     skip remaining characters in the input buffer
'B      if BLK <> 0 then skip to the next CR (eol) char
              word  @WRD_DOTPAREN
              word  of_IMMED                             'opcode flag set to immediate
WRD_BACKSLSH  byte  1,"\"
COD_BACKSLSH  word  @COD_BLK, @COD_FETCH, NFW_QBRANCH, @COD_BACKSLSH1
              'BLK <> 0, skip to end of block line
              word  @COD_TOIN, @COD_FETCH, NFW_LIT, 63, @COD_ADD
              word  NFW_LIT, $FFC0, @COD_AND, NFW_LIT, 1024, @COD_MIN
              word  @COD_TOIN, @COD_STORE
              word  NFW_BRANCH, @COD_BACKSLSH2
              'BLK = 0, skip to end of the line
COD_BACKSLSH1 word  NFW_LIT, $0D, @COD_WORD, @COD_DROP
COD_BACKSLSH2 word  NFW_EXIT
'
'FORTH COMPILER

'C CREATE   --      create an empty definition
'   LATEST @ , 0 C,         link & immed field
'   HERE LATEST !           new "latest" link
'   BL WORD C@ 1+ ALLOT         name field
'   docreate ,CF                code field
              word  @WRD_BACKSLSH
              word  of_NONE                             'opcode flag set to normal
WRD_CREATE    byte  6,"CREATE"
              'First write the link field and set the flag field
COD_CREATE    word  @COD_ALIGN, @COD_LATEST, @COD_FETCH, @COD_COMMA, NFW_LIT, of_NONE, @COD_COMMA
              ''Update the forth variable LATEST to the name field address of the new word
              word  @COD_HERE, @COD_LATEST, @COD_STORE
              'Now get the name of the new forth word, Calculate space to store name string.
              word  @COD_BL, @COD_WORD, @COD_CFETCH, @COD_ADD1
              'Word Align the allocation count for string.
              word  @COD_ALLOT, @COD_ALIGN
              {
              'Word Align the allocation count for string.
              word  @COD_DUP, @COD_ONE, @COD_AND, @COD_ADD 
              word  @COD_ALLOT
              }
              'By default words defuned by create must return there first data address
              'We need a minumum of 2 cells to rewrite a word's execution code for DOES>.
              'We add an NOP as the first execution token.
              word  NFW_LIT, NFW_NOP, @COD_XTCOMMA
              word  NFW_LIT, NFW_EXIT_ADR, @COD_XTCOMMA
              word  NFW_EXIT
'
'P (does>)  ( xt -- )     change action of latest def'n
              word  @WRD_CREATE
              word  of_IMMED                             'opcode flag set to immediate
WRD_XDOES     byte  7,"(does>)"              
              'change the last created word to a does/branch
COD_XDOES     word  NFW_LIT, NFW_DOES_BRANCH, @COD_LATEST, @COD_FETCH, @COD_NTOC, @COD_STORE
              'now set the jump address for the headless code thread
              word  @COD_LATEST, @COD_FETCH, @COD_NTOC, @COD_CELLPLUS, @COD_STORE
              'compile in an exit so we will not execute headless code thread that follows
              word  NFW_EXIT
'              
'C DOES>    --      change action of latest def'n
              word  @WRD_XDOES
              word  of_IMMED                             'opcode flag set to immediate
WRD_DOES      byte  5,"DOES>"              
COD_DOES      word  @COD_HERE, NFW_LIT, 4, @COD_CELLS, @COD_ADD, @COD_LITCOMMA
              word  NFW_ALIT, @COD_XDOES, @COD_XTCOMMA
              word  @COD_EXITCOMMA
              word  NFW_EXIT
'              
'P DEFER "name"  (  --  )   Create the word name whose behavior will be determined later and may be varied. See IS.
              word  @WRD_DOES
              word  of_NONE                             'opcode flag set to normal
WRD_DEFER     byte  5,"DEFER"
COD_DEFER     word  @COD_CREATE
              'Replace the first opcode with NOP EXIT
              word  @COD_BACKCOMMA, @COD_BACKCOMMA
              word  NFW_ALIT, NFW_NOP, @COD_XTCOMMA
              word  NFW_LIT, NFW_EXIT, @COD_XTCOMMA
              'Set the opcode flag to indicate the constant is a VALUE
              word  @COD_LATEST, @COD_FETCH, @COD_NTOF, @COD_DUP, @COD_FETCH
              word  NFW_LIT, of_DEFER, @COD_OR, @COD_SWAP, @COD_STORE
              word  NFW_EXIT
'
'C VARIABLE  "name" ( -- )     define a Forth variable to hold single cell value
'   CREATE 1 CELLS ALLOT ;
' Action of RAM variable is identical to CREATE,
' so we don't need a DOES> clause to change it.
              word  @WRD_DEFER
              word  of_NONE                             'opcode flag set to normal
WRD_VAR       byte  8,"VARIABLE"
COD_VAR       word  @COD_CREATE
              'Replace the first opcode with EXIT_ADR
              word  @COD_BACKCOMMA, @COD_BACKCOMMA, NFW_LIT, NFW_EXIT_ADR, @COD_XTCOMMA
              'Allocate one cell of storage for variable data
              word  @COD_ONE, @COD_CELLS, @COD_ALLOT
              word  NFW_EXIT
'
'C 2VARIABLE  "name" ( -- )     define a Forth variable to hold double cell value
'   CREATE 2 CELLS ALLOT ;
              word  @WRD_VAR
              word  of_NONE                             'opcode flag set to normal
WRD_2VAR      byte  9,"2VARIABLE"
COD_2VAR      word  @COD_VAR, @COD_ONE, @COD_CELLS, @COD_ALLOT
              word  NFW_EXIT
'
'P QVARIABLE  "name" ( -- )
'Define a Forth variable to hold a quad cell value that of quad byte aligned (0 4 8 ...)
'   CREATE 2 CELLS ALLOT ;
              word  @WRD_2VAR
              word  of_NONE                             'opcode flag set to normal
WRD_QVAR      byte  9,"QVARIABLE"
COD_QVAR      word  @COD_CREATE
              'Replace the first opcodes with address of quad aligned data
              'address and EXIT opcode
              word  @COD_BACKCOMMA, @COD_BACKCOMMA
              'Return quad aligned address after copiled EXIT
              word  @COD_HERE, NFW_LIT, 6, @COD_ADD, @COD_DUP, @COD_QALIGNED, @COD_DUP
              'Compile NWF_LIT 
              word  @COD_LITCOMMA, @COD_EXITCOMMA
              word  @COD_SWAP, @COD_SUB, NFW_LIT, 4, @COD_ADD, @COD_ALIGNED, @COD_ALLOT
              word  NFW_EXIT
'
'P QBYTES  "name" ( n -- )
'Define a Forth variable to hold a n bytes that of quad byte aligned (0 4 8 ...)
'   QVARIABLE 4 - 0 MAX ALLOT ;         'QVARIABLE allocates 4 bytes
              word  @WRD_QVAR
              word  of_NONE                             'opcode flag set to normal
WRD_QBYTES    byte  6,"QBYTES"
              word  @COD_QVAR, NFW_LIT, 4, @COD_SUB
              word  @COD_ZERO, @COD_MAX, @COD_ALLOT
              word  NFW_EXIT
'
'C CONSTANT  "name" ( n -- )     define a Forth constant
              word  @WRD_QBYTES
              word  of_NONE                             'opcode flag set to normal
WRD_CONST     byte  8,"CONSTANT"
COD_CONST     word  @COD_CREATE
              'Replace the first opcode with LIT
              word  @COD_BACKCOMMA, @COD_BACKCOMMA, @COD_LITCOMMA  
              'Allocate space for the constant and store the value
              word  @COD_EXITCOMMA      
              word  NFW_EXIT
'
'X VALUE  "name" ( n -- )
'In MV4th16 a value is the same as a constant.
              word  @WRD_CONST
              word  of_NONE                             'opcode flag set to normal
WRD_VALUE     byte  5,"VALUE"
COD_VALUE     word  @COD_CONST
              'Set the opcode flag to indicate the constant is a VALUE
              word  @COD_LATEST, @COD_FETCH, @COD_NTOF, @COD_DUP, @COD_FETCH
              word  NFW_LIT, of_VALUE, @COD_OR, @COD_SWAP, @COD_STORE
              word  NFW_EXIT
'
'
'P (assign) "name" ( n code-bias op-flag -- ) 
'Store a number/address into a VALUE or DEFFERed data word.
              word  @WRD_VALUE
              word  of_IMMED                    'opcode flag set to immediate
WRD_ASSIGN    byte  8,"(assign)"
COD_ASSIGN    word  @COD_BL, @COD_WORD, @COD_DUP
              'see if we are at end of line, word length = 0
              word  @COD_CFETCH, @COD_ZEQ
              'Display name not a valid VALUE token error then abort
              word  NFW_XSQUOTE
              byte  13, "Name expected"
              word  @COD_QABORT
              'we have a word see if we can find in dictionary ( n code-bias op-flag c-adr -- )
              word  @COD_DUP, @COD_NFA, @COD_DUP, @COD_ZEQ
              'Display undefined name token error then abort  ( n code-bias op-flag c-adr nfa f -- )
              word  @COD_ROT, @COD_COUNT
              'string address and char count on stack ( n code-bias op-flag nfa f c-adr cnt -- )
              word  @COD_QABORT
              'We found the search name see if it is a VALUE token ( n code-bias op-flag nfa -- )
              word  @COD_DUP, @COD_ROT, @COD_OPFLAGQ, @COD_ZEQ
              word  NFW_XSQUOTE
              byte  14, "Incorrect type"
              word  @COD_QABORT
              'Convert name field address to code field address ( n code-bias nfa -- )
              word  @COD_NTOC
              'First xt in a value is always LITERAL skip over it, for deferred word first value is xt to call  ( n code-bias cfa -- )
              word  @COD_SWAP, @COD_CELLS, @COD_ADD               
              'We have the address of value data cell folowed by the new value's value at top of stack ( n cfa -- )
              'See if we are in compile mode
              word  @COD_STATE, @COD_FETCH, NFW_QBRANCH, @COD_ASSIGN1
              'We are in compile mode compile simple phrase; LITERAL address !
              word  @COD_LITCOMMA
              word  NFW_ALIT, @COD_STORE, @COD_XTCOMMA
              word  NFW_BRANCH, @COD_ASSIGN2    
              'We are interactive mode 
COD_ASSIGN1   word  @COD_STORE          
COD_ASSIGN2   word  NFW_EXIT
'
'
'X TO  "name"  ( n -- )
'Store a number into a VALUE data word. In MV4th16 a value is the same as a constant.
              word  @WRD_ASSIGN
              word  of_IMMED                    'opcode flag set to immediate
WRD_TO        byte  2,"TO"
              word  @COD_ONE, NFW_LIT, of_VALUE, @COD_ASSIGN
              word  NFW_EXIT
'
'P IS  "name"  ( xt -- )
'Give the behavior identified by xt to the DEFER word name. Word "name" must be a defered word
              word  @WRD_TO
              word  of_IMMED                    'opcode flag set to immediate
WRD_IS        byte  2,"IS"
              word  @COD_ZERO, NFW_LIT, of_DEFER, @COD_ASSIGN
              word  NFW_EXIT
'
'
'C 2CONSTANT  "name" ( n -- )     define a Forth double constant
'   CREATE , DOES> (machine code fragment)
              word  @WRD_IS
              word  of_NONE                             'opcode flag set to normal
WRD_2CONST    byte  9,"2CONSTANT"
COD_2CONST    word  @COD_CREATE
              word  @COD_SWAP
              'Create Leaves a NOP and EXIT_ADR opcode, remove from current definition    
              word  @COD_BACKCOMMA, @COD_BACKCOMMA 
              'Place the low, then high word from the double on the stack
              word  NFW_LIT, NFW_DLIT, @COD_XTCOMMA, @COD_COMMA, @COD_COMMA  
              'Place an EXIT opcode at the end of the definition.
              word  @COD_EXITCOMMA      
              word  NFW_EXIT
'
'X MARKER  "name"  ( -- )      create code thread to restore state of dictionary
'                     before this marker was created (save DP @, LATEST @)..
              word  @WRD_2CONST
              word  of_NONE                             'opcode flag set to normal
WRD_MARKER    byte  6,"MARKER"
              'save current dictionary state to data stack
COD_MARKER    word  @COD_HERE, @COD_LATEST, @COD_FETCH
              'Create a dictionary entry for the new marker
              word  @COD_CREATE
              'Remove first opcode after create (NFW_EXIT_ADR)
              word  @COD_BACKCOMMA, @COD_BACKCOMMA 
              'compile the value of LATEST as a literal value 
              word  @COD_LITCOMMA 
              'compile the value of HERE as a literal value 
              word  @COD_LITCOMMA
              'compile code to restore the value of the DP (HERE)
              word  NFW_ALIT, @COD_DP, @COD_XTCOMMA, NFW_ALIT, @COD_STORE, @COD_XTCOMMA
              'compile code to restore the value of the LATEST
              word  NFW_ALIT, @COD_LATEST, @COD_XTCOMMA, NFW_ALIT, @COD_STORE, @COD_XTCOMMA
              'complete code thread with an EXIT opcode
              word  @COD_EXITCOMMA      
              word  NFW_EXIT
'
'P USER     n --        define user variable 'n'
              word  @WRD_MARKER
              word  of_NONE                             'opcode flag set to normal
WRD_USER      byte  4,"USER"
COD_USER      word  @COD_CREATE
              'Create Leaves a EXIT_ADR opcode, remove from current definition    
              word  @COD_BACKCOMMA, @COD_BACKCOMMA 
              'Replace the first opcode with LIT
              word  @COD_XUSER, @COD_LITCOMMA  
              'Allocate space for the constant address and store the address value
              word  @COD_EXITCOMMA      
              word  NFW_EXIT
'
'C RECURSE  --      recurse current definition
'   LATEST @ NFA>CFA ,XT ; IMMEDIATE
              word  @WRD_USER
              word  of_IMMED                             'opcode flag set to immediate
WRD_RECURSE   byte  7,"RECURSE"
COD_RECURSE   word  @COD_LATEST, @COD_FETCH, @COD_NTOC, @COD_XTCOMMA
              word  NFW_EXIT
'
'C [        --      enter interpretive state
'   0 STATE ! ; IMMEDIATE
              word  @WRD_RECURSE
              word  of_IMMED                             'opcode flag set to immediate
WRD_LEFTBRAC  byte  1,"["
COD_LEFTBRAC  word  @COD_FALSE, @COD_STATE, @COD_STORE
              word  NFW_EXIT
'
'C ]        --      enter compiling state
'   -1 STATE ! ;
              word  @WRD_LEFTBRAC
              word  of_NONE                              'opcode flag set to normal
WRD_RIGHTBRAC byte  1,"]"
COD_RIGHTBRAC word  @COD_TRUE, @COD_STATE, @COD_STORE
              word  NFW_EXIT
'
'P HIDE    ( -- )    "hide" latest definition
              word  @WRD_RIGHTBRAC
              word  of_NONE                              'opcode flag set to normal
WRD_HIDE      byte  4,"HIDE"
COD_HIDE      word  @COD_LATEST, @COD_FETCH, @COD_DUP, @COD_CFETCH, NFW_LIT, $80, @COD_OR
              word  @COD_SWAP, @COD_CSTORE
              word  NFW_EXIT
'
'P REVEAL  ( -- )     "reveal" latest definition
              word  @WRD_HIDE
              word  of_NONE                              'opcode flag set to normal
WRD_REVEAL    byte  6,"REVEAL"
COD_REVEAL    word  @COD_LATEST, @COD_FETCH, @COD_DUP, @COD_CFETCH, NFW_LIT, $7F, @COD_AND
              word  @COD_SWAP, @COD_CSTORE
              word  NFW_EXIT
'
'P HIDDEN?  ( -- f )     see if the latest definition is hiddden
              word  @WRD_REVEAL
              word  of_NONE                              'opcode flag set to normal
WRD_HIDDENQ   byte  6,"HIDDEN?"
COD_HIDDENQ   word  @COD_LATEST, @COD_FETCH, @COD_CFETCH, NFW_LIT, $80, @COD_AND
              word  @COD_ZNE
              word  NFW_EXIT
'
'C IMMEDIATE  ( -- )  make last def'n immediate
              word  @WRD_HIDDENQ
              word  of_IMMED                              'opcode flag set to normal
WRD_IMMED     byte  9,"IMMEDIATE"
COD_IMMED     word  @COD_LATEST, @COD_FETCH, @COD_NTOF, @COD_DUP
              word  @COD_FETCH, NFW_LIT, of_IMMED, @COD_OR
              word  @COD_SWAP, @COD_STORE
              word  NFW_EXIT
'
'C :        --      begin a colon definition
'   CREATE HIDE ] !COLON ;
              word  @WRD_IMMED
              word  of_NONE                              'opcode flag set to normal
WRD_COLON     byte  1,":"
COD_COLON     word  @COD_CREATE, @COD_BACKCOMMA, @COD_BACKCOMMA, @COD_HIDE, @COD_RIGHTBRAC
              word  NFW_EXIT
'
'X :NONAME     --      begin a headless definition
'   ALIGN HERE [ ;
              word  @WRD_COLON
              word  of_NONE                              'opcode flag set to normal
WRD_COLNONAME byte  7,":NONAME"
COD_COLNONAME word  @COD_ALIGN, @COD_HERE, @COD_RIGHTBRAC
              word  NFW_EXIT
'
'C ;
'   REVEAL  EXIT,
'   POSTPONE [  ; IMMEDIATE
              word  @WRD_COLNONAME
              word  of_IMMED                           'opcode flag set to immediate
WRD_SEMICOLON byte  1,";"
COD_SEMICOLON word  @COD_REVEAL, @COD_EXITCOMMA, @COD_LEFTBRAC
              word  NFW_EXIT
'
'C [']  --         find word & compile as literal
'   '  ['] LIT ,XT  , ; IMMEDIATE
' When encountered in a colon definition, the
' phrase  ['] xxx  will cause   LIT,xxt  to be
' compiled into the colon definition (where
' (where xxt is the execution token of word xxx).
' When the colon definition executes, xxt will
' be put on the stack.  (All xt's are one cell.)
'C ;
              word  @WRD_SEMICOLON
              word  of_IMMED                    'opcode flag set to immediate
WRD_BRACTICK  byte  3,"[']"
COD_BRACTICK  word  @COD_TICK                   'get xt of 'xxx'
              word  @COD_LITCOMMA
              word  NFW_EXIT
'
'C POSTPONE  --   postpone compile action of word
'   BL WORD DUP FIND ROT OVER
'   0= SWAP COUNT ?ABORT
'   0< IF   -- xt  non immed: add code to current
'                  def'n to compile xt later.
'       ['] LIT ,XT  ,      add "LIT,xt,COMMAXT"
'       ['] ,XT ,XT         to current definition
'   ELSE  ,XT      immed: compile into cur. def'n
'   THEN ; IMMEDIATE
              word  @WRD_BRACTICK
              word  of_IMMED                    'opcode flag set to immediate
WRD_POSTPONE  byte  8,"POSTPONE"
COD_POSTPONE  word  @COD_BL, @COD_WORD, @COD_DUP, @COD_FIND, @COD_ROT, @COD_OVER, @COD_ZEQ
              word  @COD_SWAP, @COD_COUNT
              'if FIND failed to find word abort
              word  @COD_QABORT
              'If the word is immediate just compile it now, if not compile code
              'to compile word when code thread is run
              word  @COD_ZLT, NFW_QBRANCH, @COD_POST1
              'Compile instruction to compile the word later
              word  @COD_LITCOMMA
              word  NFW_ALIT, @COD_XTCOMMA, @COD_XTCOMMA, NFW_BRANCH, @COD_POST2
              'We have an immediate execution token so compile the token now
COD_POST1     word  @COD_XTCOMMA
COD_POST2     word  NFW_EXIT
'
'X [COMPILE]  Compilation: (  <spaces>name  -- )
'If name has other than default compilation semantics, append them
'to the current definition; otherwise append the execution semantics of name.
              word  @WRD_POSTPONE
              word  of_IMMED                    'opcode flag set to immediate
WRD_CMP_BRAC  byte  9,"[COMPILE]"
COD_CMP_BRAC  word  @COD_TICK
              word  @COD_XTCOMMA 
              word  NFW_EXIT
'
'
'FORTH CONTROL STRUCTURES
'
'C IF       -- adrs    conditional forward branch
'   ['] qbranch ,BRANCH  HERE DUP ,DEST ;
'   IMMEDIATE
              word  @WRD_CMP_BRAC
              word  of_IMMED                    'opcode flag set to immediate
WRD_IF        byte  2,"IF"
COD_IF        word  NFW_LIT, NFW_QBRANCH, @COD_COMMABRANCH
              word  @COD_HERE, @COD_DUP, @COD_COMMADEST 
              word  NFW_EXIT
'
'C THEN     adrs --        resolve forward branch
'   HERE SWAP !DEST ; IMMEDIATE
              word  @WRD_IF
              word  of_IMMED                    'opcode flag set to immediate
WRD_THEN      byte  4,"THEN"
COD_THEN      word  @COD_HERE, @COD_SWAP, @COD_STOREDEST
              word  NFW_EXIT
'
'C ELSE     adrs1 -- adrs2    branch for IF..ELSE
'   ['] branch ,BRANCH  HERE DUP ,DEST
'   SWAP  POSTPONE THEN ; IMMEDIATE
              word  @WRD_THEN
              word  of_IMMED                    'opcode flag set to immediate
WRD_ELSE      byte  4,"ELSE"
COD_ELSE      word  NFW_LIT, NFW_BRANCH, @COD_COMMABRANCH
              word  @COD_HERE, @COD_DUP, @COD_COMMADEST
              word  @COD_SWAP, @COD_THEN
              word  NFW_EXIT
'
'C BEGIN    -- adrs        target for bwd. branch
'   HERE ; IMMEDIATE
              word  @WRD_ELSE
              word  of_IMMED                    'opcode flag set to immediate
WRD_BEGIN     byte  5,"BEGIN"
COD_BEGIN     word  NFW_BRANCH, @COD_HERE
              word  NFW_EXIT
'
'C UNTIL    adrs --   conditional backward branch
'   ['] qbranch ,BRANCH  ,DEST ; IMMEDIATE
'   conditional backward branch
              word  @WRD_BEGIN
              word  of_IMMED                    'opcode flag set to immediate
WRD_UNTIL     byte  5,"UNTIL"
COD_UNTIL     word  NFW_LIT, NFW_QBRANCH, @COD_COMMABRANCH 
              word  @COD_COMMADEST
              word  NFW_EXIT
'
'X AGAIN    adrs --      uncond'l backward branch
'   ['] branch ,BRANCH  ,DEST ; IMMEDIATE
'   unconditional backward branch
              word  @WRD_UNTIL
              word  of_IMMED                    'opcode flag set to immediate
WRD_AGAIN     byte  5,"AGAIN"
COD_AGAIN     word  NFW_LIT, NFW_BRANCH, @COD_COMMABRANCH 
              word  @COD_COMMADEST
              word  NFW_EXIT
'
'C WHILE    -- adrs         branch for WHILE loop
'   POSTPONE IF ; IMMEDIATE
              word  @WRD_AGAIN
              word  of_IMMED                    'opcode flag set to immediate
WRD_WHILE     byte  5,"WHILE"
COD_WHILE     word  NFW_BRANCH, @COD_IF 
              word  NFW_EXIT
'
'C REPEAT   adrs1 adrs2 --     resolve WHILE loop
'   SWAP POSTPONE AGAIN POSTPONE THEN ; IMMEDIATE
              word  @WRD_WHILE
              word  of_IMMED                    'opcode flag set to immediate
WRD_REPEAT    byte  6,"REPEAT"
COD_REPEAT    word  @COD_SWAP, @COD_AGAIN, @COD_THEN 
              word  NFW_EXIT
'
'X CASE   ( -- 0 ) Start of case statement
              word  @WRD_REPEAT
              word  of_IMMED                    'opcode flag set to immediate
WRD_CASE      byte  4,"CASE"
              'Leave the CASE OF count on the stack
              word  @COD_ZERO 
              word  NFW_EXIT
'
'X OF     ( #of -- orig #of+1 / x -- )  Start of case match clause
              word  @WRD_CASE
              word  of_IMMED                    'opcode flag set to immediate
WRD_OF        byte  2,"OF"
              'Leave the CASE OF count on the stack
              word  @COD_ADD1, @COD_TO_R
              word  NFW_ALIT, @COD_OVER, @COD_XTCOMMA 
              word  NFW_ALIT, @COD_EQ, @COD_XTCOMMA 
              word  @COD_IF
              word  NFW_ALIT, @COD_DROP, @COD_XTCOMMA
              word  @COD_R_TO 
              word  NFW_EXIT
'              
'P OF-RANGE     ( #of -- orig #of+1 / l h -- )  Start of case match clause when true
              word  @WRD_OF
              word  of_IMMED                    'opcode flag set to immediate
WRD_OFRANGE   byte  8,"OF-RANGE"
              'Save the CASE OF count on the return stack after adding one
              word  @COD_ADD1, @COD_TO_R
              word  NFW_ALIT, @COD_2TO_R, @COD_XTCOMMA 
              word  NFW_ALIT, @COD_DUP, @COD_XTCOMMA 
              word  NFW_ALIT, @COD_2R_TO, @COD_XTCOMMA 
              word  NFW_ALIT, @COD_WITHIN, @COD_XTCOMMA 
              word  @COD_IF
              word  NFW_ALIT, @COD_DROP, @COD_XTCOMMA
              word  @COD_R_TO 
              word  NFW_EXIT
'              
'X ENDOF  ( orig1 #of -- orig2 #of ) marks end of CASE OF clause
              word  @WRD_OFRANGE
              word  of_IMMED                    'opcode flag set to immediate
WRD_ENDOF     byte  5,"ENDOF"
              word  @COD_TO_R
              word  @COD_ELSE 
              word  @COD_R_TO 
              word  NFW_EXIT
'
'X ENDCASE ( orig 1..orign #of -- ) marks end of CASE clause
              word  @WRD_ENDOF
              word  of_IMMED                    'opcode flag set to immediate
WRD_ENDCASE   byte  7,"ENDCASE"
              'Close all IF ELSE statments with THEN
              'First drop the CASE test value from the stack
              word  NFW_ALIT, @COD_DROP, @COD_XTCOMMA
              'Number of OF statments generated on top of stack
              word  @COD_IFDUP, NFW_QBRANCH, @COD_ENDCASE2
              word  @COD_ZERO, @COD_XDO
COD_ENDCASE1  word  @COD_THEN
              word  NFW_XLOOP, @COD_ENDCASE1
COD_ENDCASE2  word  NFW_EXIT
'              
'P OTHERWISE ( -- ) Indicates start of no match statement
              word  @WRD_ENDCASE
              word  of_IMMED                    'opcode flag set to immediate
WRD_CASEOTHER byte  9,"OTHERWISE"
              'Compatibilty word does nothing in MV4th
              word  NFW_EXIT
'
'P >L   x --   L: -- x        move to leave stack
'   CELL LP +!  LP @ ! ;      (L stack grows up)
              word  @WRD_CASEOTHER
              word  of_NONE                    'opcode flag set to normal
WRD_TO_L      byte  2,">L"
COD_TO_L      word  @COD_CELL, @COD_LP, @COD_PLUSSTORE, @COD_LP, @COD_FETCH, @COD_STORE 
              word  NFW_EXIT
'
'P L>   -- x   L: x --      move from leave stack
'   LP @ @  CELL NEGATE LP +! ;
              word  @WRD_TO_L
              word  of_NONE                    'opcode flag set to normal
WRD_L_TO      byte  2,"L>"
COD_L_TO      word  @COD_LP, @COD_FETCH, @COD_FETCH
              word  @COD_CELL, @COD_NEGATE, @COD_LP, @COD_PLUSSTORE
              word  NFW_EXIT
'
'X ?DO       -- adrs   L: -- 0
'   2DUP ,XT = ,XT
'   IF 2DROP ,XT ELSE
'   ['] xdo ,XT   HERE     target for bwd branch
'   0 >L ; IMMEDIATE           marker for LEAVEs
              word  @WRD_L_TO
              word  of_IMMED                    'opcode flag set to immediate
WRD_QDO       byte  3,"?DO"
              word  NFW_ALIT, @COD_2DUP, @COD_XTCOMMA
              word  NFW_ALIT, @COD_EQ, @COD_XTCOMMA, @COD_IF
              'Loop values are equal do not loop
              word  NFW_ALIT, @COD_2DROP, @COD_XTCOMMA
              word  @COD_ELSE
              word  @COD_ZERO, @COD_TO_L
              word  @COD_TO_L
              word  NFW_ALIT, @COD_XDO, @COD_XTCOMMA, @COD_HERE
              word  NFW_EXIT
'
'C DO       -- adrs   L: -- 0
'   ['] xdo ,XT   HERE     target for bwd branch
'   0 >L ; IMMEDIATE           marker for LEAVEs
              word  @WRD_QDO
              word  of_IMMED                    'opcode flag set to immediate
WRD_DO        byte  2,"DO"
              word  NFW_ALIT, @COD_XDO, @COD_XTCOMMA, @COD_HERE
              word  @COD_ZERO, @COD_TO_L
              word  NFW_EXIT
'
'P ENDLOOP   adrs xt --   L: 0 a1 a2 .. aN --
'   ,BRANCH  ,DEST                backward loop
'   BEGIN L> ?DUP WHILE POSTPONE THEN REPEAT ;
'                                 resolve LEAVEs
' This is a common factor of LOOP and +LOOP.
              word  @WRD_DO
              word  of_NONE                    'opcode flag set to normal
WRD_ENDLOOP   byte  7,"ENDLOOP"
COD_ENDLOOP   word  @COD_COMMABRANCH, @COD_COMMADEST
COD_ENDLOOP1  word  @COD_L_TO, @COD_IFDUP, NFW_QBRANCH, @COD_ENDLOOP2
              word  @COD_HERE, @COD_SWAP, @COD_STOREDEST, NFW_BRANCH, @COD_ENDLOOP1
COD_ENDLOOP2  word  NFW_EXIT
'
'C LOOP    adrs --   L: 0 a1 a2 .. aN --
'   ['] xloop ENDLOOP ;  IMMEDIATE
              word  @WRD_ENDLOOP
              word  of_IMMED                    'opcode flag set to immediate
WRD_LOOP      byte  4,"LOOP"
              word  NFW_LIT, NFW_XLOOP, @COD_ENDLOOP
              word  NFW_EXIT
'
'C +LOOP   adrs --   L: 0 a1 a2 .. aN --
'   ['] xplusloop ENDLOOP ;  IMMEDIATE
              word  @WRD_LOOP
              word  of_IMMED                    'opcode flag set to immediate
WRD_PLUSLOOP  byte  5,"+LOOP"
COD_PLUSLOOP  word  NFW_LIT, NFW_XPLUSLOOP, @COD_ENDLOOP
              word  NFW_EXIT
'
'C LEAVE ( -- ) ( L: -- adr )
'Compile words to discard the current loop control parameters.
'Continue execution immediately'following the innermost syntactically
'enclosing DO ... LOOP or DO ... +LOOP.
              word  @WRD_PLUSLOOP
              word  of_IMMED                    'opcode flag set to immediate
WRD_LEAVE     byte  5,"LEAVE"
              word  NFW_ALIT, @COD_UNLOOP, @COD_XTCOMMA
              word  NFW_LIT, NFW_BRANCH, @COD_COMMABRANCH
              word  @COD_HERE, @COD_DUP, @COD_COMMADEST
              word  @COD_TO_L
              word  NFW_EXIT
'
'
'OTHER FORTH OPERATORS
'
'X WITHIN   n1|u1 n2|u2 n3|u3 -- f   n2<=n1<n3?
'  OVER - >R - R> U< ;          per ANS document
              word  @WRD_LEAVE
              word  of_NONE                    'opcode flag set to normal
WRD_WITHIN    byte  6,"WITHIN"
COD_WITHIN    word  @COD_OVER, @COD_SUB, @COD_TO_R, @COD_SUB, @COD_R_TO, @COD_ULT
              word  NFW_EXIT
'
';C MOVE    addr1 addr2 u --     smart move
'             VERSION FOR 1 ADDRESS UNIT = 1 CHAR
'  >R 2DUP SWAP DUP R@ +     -- ... dst src src+n
'  WITHIN IF  R> CMOVE>        src <= dst < src+n
'       ELSE  R> CMOVE  THEN ;          otherwise
              word  @WRD_WITHIN
              word  of_NONE                    'opcode flag set to normal
WRD_MOVE      byte  4,"MOVE"
COD_MOVE      word  @COD_TO_R, @COD_2DUP, @COD_SWAP, @COD_DUP, @COD_R_FETCH, @COD_ADD
              word  @COD_WITHIN, NFW_QBRANCH, @COD_MOVE1
              word  @COD_R_TO, @COD_CMOVEREV, NFW_BRANCH, @COD_MOVE2
COD_MOVE1     word  @COD_R_TO, @COD_CMOVE
COD_MOVE2     word  NFW_EXIT
'
'C ENVIRONMENT?  c-addr u -- false   system query
'                         -- i*x true
'   2DROP 0 ;       the minimal definition!
              word  @WRD_MOVE
              word  of_NONE                    'opcode flag set to normal
WRD_ENVIRONQ  byte  12,"ENVIRONMENT?"
              word  @COD_2DROP, @COD_ZERO
              word  NFW_EXIT
'
'
'X MS  ( u -- )   delay for u milli-seconds
              word  @WRD_ENVIRONQ
              word  of_NONE                    'opcode flag set to normal
WRD_MS        byte  2,"MS"
COD_MS        word  @COD_IFDUP, @COD_ZNE, NFW_QBRANCH, @COD_MS1 
'                   ( uMS )
              word  @COD_CLKFREQ, NFW_DLIT, 1000, 0, @COD_DDIV
'                   ( uMS dOneMs )
              word  @COD_ROT, @COD_ZERO, @COD_XDO
'                   ( wait for one MS )
'                   ( dOneMS )
COD_MS3       word  @COD_COG_CNT
'                   ( wait loop )
'                   ( dOneMs dCntStart )
COD_MS2       word  @COD_COG_CNT, @COD_2OVER, @COD_DSUB
'                   ( dOneMs dCntStart dCntDelta )
              word  @COD_2TO_R, @COD_2OVER, @COD_2R_TO
'                   ( dOneMs dCntStart dOneMs dCntDelta )
              word  @COD_UDGT, @COD_NOT, NFW_QBRANCH, @COD_MS2       
'                   ( dOneMs dCntStart  )
              word  @COD_2DROP, NFW_XLOOP, @COD_MS3
'                   ( dOneMs  )
              word  @COD_2DROP
COD_MS1       word  NFW_EXIT
'
'
'FORTH DEBUGING WORDS AND STARTUP CODE
'
'T WORDS   ( -- )       list all words in dict.
'Display a list of all forth word names to the console
'              word  @WRD_DUMP
              word  @WRD_MS
              word  of_NONE                     'opcode flag set to normal
WRD_WORDS     byte  5,"WORDS"              
              word  @COD_CR, @COD_LATEST, @COD_FETCH
              'Display the word, save a copy of the forth word address
COD_WORDS1    word  @COD_DUP, @COD_COUNT, NFW_LIT, $7F, @COD_AND, @COD_TYPE, @COD_SPACE
              'Get next word address from current word link field
              'If the link field is zero we reached the last forth word.
              word  @COD_NTOL, @COD_FETCH, @COD_DUP, NFW_QBRANCH, @COD_WORDS2
              word  NFW_BRANCH, @COD_WORDS1    
COD_WORDS2    word  @COD_DROP
              word  NFW_EXIT
'             
'T .S     ( -- )        print stack contents
              word  @WRD_WORDS
              word  of_NONE                             'opcode flag set to normal
WRD_DOTS      byte  2,".S"
              'First make sure stack has not underflowed
              word  @COD_SPFETCH, @COD_S0, @COD_GT 
              'Stack under flow..
              word  NFW_XSQUOTE
              byte  9, "underflow"
              word  @COD_QABORT 
              'Stack is ok show values  
COD_DOTS3     word  @COD_CR, @COD_SPFETCH, @COD_S0, @COD_SUB, NFW_QBRANCH, @COD_DOTS2
              word  @COD_SPFETCH, NFW_LIT, 2, @COD_SUB, @COD_S0, NFW_LIT, 2, @COD_SUB, @COD_XDO
COD_DOTS1     word  @COD_I, @COD_FETCH, @COD_UDOT, NFW_LIT, -2, NFW_XPLUSLOOP, @COD_DOTS1
COD_DOTS2     word  NFW_EXIT
'
'
'? .     ( a-adr --  )    Display the value stored at the aligned address a-addr.
              word  @WRD_DOTS
              word  of_NONE                     'opcode flag set to normal
WRD_QUESTION  byte  1,"?"              
              word  @COD_FETCH, @COD_DOT
              word  NFW_EXIT
'
'Forth BLOCK extensions
'
'Initialize block extension variables, 
              word  @WRD_QUESTION
              word  of_NONE                             'opcode flag set to normal
WRD_BLKINIT   byte  13,"EMPTY-BUFFERS"
COD_BLKINIT   word  NFW_LIT, -1, NFW_ALIT, @var_BufferBlock, @COD_STORE          '-1 = no block in buffer
              word  @COD_ZERO, NFW_ALIT, @var_BufferDirty, @COD_STORE          '0 = buffer is not modified
              word  NFW_EXIT
'
'
'P blk-cnt  ( -- u ) number of device blocks on storage device
              word  @WRD_BLKINIT
              word  of_NONE                     'opcode flag set to normal
WRD_BLK_CNT   byte  9,"blk-count"              
COD_BLK_CNT   word  NFW_LIT, cmd_BLK_COUNT, @COD_SYSTEM    
              word  NFW_EXIT
'
'P block-read  (u -- )  Load block u into the block buffer
'ABORT if read failed.
'              word  @WRD_BLKWCNT
              word  @WRD_BLK_CNT
              word  of_NONE                             'opcode flag set to normal
WRD_BLKLOAD   byte  7,"readblk"
COD_BLKLOAD   word  @COD_DUP, NFW_ALIT, @Param_BlockBufferBase, @COD_FETCH, NFW_LIT
              'If we could not read block, read returned 0, report error
              word  cmd_READBLK, @COD_SYSTEM, @COD_ZEQ
              word  NFW_XSQUOTE
              byte  20, "Unable to read block"
              word  @COD_QABORT
              'We loaded the block set variables to indicate
              word  NFW_ALIT, @var_BufferBlock, @COD_STORE                     'Save the last read block number
              word  @COD_ZERO, NFW_ALIT, @var_BufferDirty, @COD_STORE          '0 = buffer is not modified
COD_EXITLOAD  word  NFW_EXIT
'
'P block-write  (u -- ) save block buffer to block u on the block device 
'ABORT if read failed.
              word  @WRD_BLKLOAD
              word  of_NONE                             'opcode flag set to normal
WRD_BLKSAVE   byte  8,"writeblk"
COD_BLKSAVE   word  @COD_DUP, NFW_ALIT, @Param_BlockBufferBase, @COD_FETCH, NFW_LIT
              word  cmd_WRITEBLK, @COD_SYSTEM, @COD_ZEQ
              word  NFW_XSQUOTE
              byte  20, "Unable to write block"
              'if we could not write block, write returned 0, report error
              word  @COD_QABORT
              'We SAVEed the block set variables to indicate
              word  NFW_ALIT, @var_BufferBlock, @COD_STORE                     'Save the block number for the buffer
              word  @COD_ZERO, NFW_ALIT, @var_BufferDirty, @COD_STORE          '0 = buffer is not modified
              'All done exit and return block buffer address
COD_BLKSAVE2  word  NFW_EXIT
'
'If the buffer number u is assigned to a different block that is marked as dirty save the block.
'Function returns the address of the block buffer at the top of the data stack.
'B BUFFER     (u -- adr)
              word  @WRD_BLKSAVE
              word  of_NONE                             'opcode flag set to normal
WRD_BUFFER    byte  6,"BUFFER"
              'If the block number is the same as the one in the block buffer just return address
COD_BUFFER    word  @COD_DUP, NFW_ALIT, @var_BufferBlock, @COD_FETCH, @COD_NEQ, NFW_QBRANCH, @COD_BUFFER1
              'Block numbers are different, If the current block buffer is dirty save it,
              'See if the block number < 0 if it is assign buffer to block and return buffer address
COD_BUFFER2   word  NFW_ALIT, @var_BufferBlock, @COD_FETCH, @COD_ZLT, @COD_NOT, NFW_QBRANCH, @COD_BUFFER1
              'If the block is not marked as dirty if not assign buffer to block and return buffer address
              word  NFW_ALIT, @var_BufferDirty, @COD_FETCH, NFW_QBRANCH, @COD_BUFFER1
              'The buffer is dirty we must save it first
              word  NFW_ALIT, @var_BufferBlock, @COD_FETCH, @COD_BLKSAVE
              'Set the block buffer number to the requested block (1..XX) 
COD_BUFFER1   word  NFW_ALIT, @var_BufferBlock, @COD_STORE, NFW_ALIT, @Param_BlockBufferBase, @COD_FETCH
              word  NFW_EXIT
'
'Read a block from the mass store device if block is not already in buffer
'If the buffer is assigned to a different block and is dirty save the block.
'Function returns the address of the block buffer at the top of the data stack.
'B BLOCK      (u -- adr)
              word  @WRD_BUFFER
              word  of_NONE                             'opcode flag set to normal
WRD_BLOCK     byte  5,"BLOCK"
              'If the block number is the same as the one in the block buffer just return address
COD_BLOCK     word  @COD_DUP, NFW_ALIT, @var_BufferBlock, @COD_FETCH, @COD_EQ, NFW_QBRANCH, @COD_BLOCK2
              'block number is one in buffer so noting more to do, drop block number from stack, then exit with buffer address
              word  @COD_DROP, NFW_BRANCH, @COD_BLOCK1
              'Free the buffer updating the block device if reuired. We do not need buffer address so drop it
COD_BLOCK2    word  @COD_DUP, @COD_BUFFER, @COD_DROP    
              word  @COD_BLKLOAD
              'Return the block buffer address                 
COD_BLOCK1    word  NFW_ALIT, @Param_BlockBufferBase, @COD_FETCH
              word  NFW_EXIT
'
'Mark the current block buffer as dirty (modified)
'B UPDATE      ( -- )
              word  @WRD_BLOCK
              word  of_NONE                             'opcode flag set to normal
WRD_UPDATE    byte  6,"UPDATE"
              word  NFW_LIT, -1, NFW_ALIT, @var_BufferDirty, @COD_STORE
              word  NFW_EXIT
'
'Save the any dirty block buffers to the block device
'B SAVE-BUFFERS  ( -- )
              word  @WRD_UPDATE
              word  of_NONE                             'opcode flag set to normal
WRD_SAVBUFS   byte  12,"SAVE-BUFFERS"
              'See if the block buffer is assigned to a device block, if not we are done
COD_SAVBUFS   word  NFW_ALIT, @var_BufferBlock, @COD_FETCH,  @COD_ZLT, @COD_NOT, NFW_QBRANCH, @COD_SAVBUFS1
              'If the block is not marked as dirty we are done
              word  NFW_ALIT, @var_BufferDirty, @COD_FETCH, NFW_QBRANCH, @COD_SAVBUFS1
              'The buffer is dirty we must save it. BLKSAVE will clear the dirty flag
              word  NFW_ALIT, @var_BufferBlock, @COD_FETCH, @COD_BLKSAVE 
COD_SAVBUFS1  word  NFW_EXIT
'
'Save the any dirty block buffers to the block device, then clear the buffer block assignments
'B FLUSH  ( -- )
              word  @WRD_SAVBUFS
              word  of_NONE                             'opcode flag set to normal
WRD_FLUSH     byte  5,"FLUSH"
              'no COD_XXX lable to save IDE symbol table space              
              'See if the block buffer is assigned to a device block, if not we are done
              word  @COD_SAVBUFS, @COD_BLKINIT
              word  NFW_EXIT
'
'Attempt to fill the input buffer from the input source, returning a true flag if successful.
'B REFILL  ( -- flag )
              word  @WRD_FLUSH
              word  of_NONE                             'opcode flag set to normal
WRD_REFILL    byte  6,"REFILL"
              'no COD_XXX lable to save IDE symbol table space              
              'if BLK = 0 then SOURCE is not a mass storage block
COD_REFILL    word  @COD_BLK, @COD_FETCH, NFW_QBRANCH, @COD_REFILL3
              'Input source is a block, see if we can select next block
              word  @COD_BLK, @COD_FETCH, @COD_ADD1
              'See if next block is valid block number
              word  @COD_BLK_CNT, @COD_LT, NFW_QBRANCH, @COD_REFILL4
              'We have a next block, assign it as the current input
              word  @COD_ONE, @COD_BLK, @COD_PLUSSTORE
              word  @COD_BLK, @COD_FETCH, @COD_BLOCK, NFW_LIT, 1024    
COD_REFILL5   word  @COD_TSOURCE, @COD_2STORE, @COD_ZERO, @COD_TOIN, @COD_STORE
              word  @COD_TRUE, NFW_BRANCH, @COD_REFILL1
              'Input is not assigned to a block buffer..
              'If the SOURCE-ID <> 0 we are evaluating a string return false
COD_REFILL3   word  @COD_SOURCE_ID,  NFW_QBRANCH, @COD_REFILL2
              'SOURCE-ID <> 0 return false
COD_REFILL4   word  @COD_FALSE, NFW_BRANCH, @COD_REFILL1
             'normal input source, read next line into TIB 
COD_REFILL2   word  @COD_TIB, @COD_DUP, @COD_TIBSIZE, @COD_ACCEPT, NFW_BRANCH, @COD_REFILL5
COD_REFILL1   word  NFW_EXIT
'
'B LIST   ( u -- ) List the text contens of the BLOCK number u in 64x16 line mode
              word  @WRD_REFILL
              word  of_NONE                             'opcode flag set to normal
WRD_LIST      byte  4,"LIST"
              'See if the block buffer is assigned to a device block, if not we are done
              'no COD_XXX lable to save IDE symbol table space              
              word  @COD_BASE                  'Put BASE addr on stack 
              word  @COD_FETCH                 'Put current base on stack 
              word  @COD_SWAP                  'Get number of list screen to top 
              word  @COD_DECIMAL               'Sets decimal mode 
              word  @COD_CR                    'Output [CR][LF]
              word  @COD_DUP                   'Duplicate top value on stack (block number) 
              word  @COD_SCR                   'Set most recently listed
              word  @COD_STORE                 'Store word at addr 
              word  NFW_XSQUOTE                'Output following string
              byte  6,"SCR # "
              word  @COD_TYPE
              word  @COD_DUP                   'Duplicate top value on stack (block number) 
              word  @COD_DOT                   'Output the screen number
              word  @COD_BLOCK                 'Place first line address on data stack
              word  NFW_DLIT                   'Puts next 2 bytes on the stack
              word  $0010                      'From 0 to 15
              word  $0000                      '16 lines to do
              word  @COD_XDO                    'Put start & end loop values on RPP
              'repeat for each line to display
COD_LIST1     word  @COD_CR                    'Output [CR][LF] 
              word  @COD_I                     'Line number onto data stack
              word  NFW_LIT                    'Puts next 2 bytes on the stack 
              word  0003                       'Fromat right justified 3 characters
              word  @COD_DOTR                  'Output formatted
              word  @COD_SPACE                 'Output space
              word  @COD_DUP 
              word  NFW_LIT, 64, @COD_TYPE     'Display the line
              word  NFW_LIT, 64, @COD_ADD 
              word  NFW_XLOOP                  'Increment loop & branch if not done
              word  @COD_LIST1
              'Done lopping for all lines
              word  @COD_DROP               
              word  @COD_CR                    'Output [CR][LF]
              word  @COD_BASE                  'Put BASE addr on stack 
              word  @COD_STORE                 'Restore original base
              word  NFW_EXIT
'
'P  COG-NEW ( q-adr u -- n )
'Run the code at address q-adr (must be a quad byte aligned address)
'u (upper 14 bits) becomes the value of the COG's PAR register.
'n = the ID of the COG started or -1 (FFFF) if no cog available.
              word  @WRD_LIST
              word  of_NONE                     'opcode flag set to normal
WRD_COGNEW    byte  7,"COG-NEW"              
              'no COD_XXX lable to save IDE symbol table space              
COD_COGNEW    word  NFW_LIT, cmd_COG_NEW, @COD_SYSTEM    
              word  NFW_EXIT
'
'P  COG-INIT ( q-adr u1 u2 -- )
'Run the code at address q-adr (must be a quad byte aligned address),
'u1 (upper 14 bits) becomes the value of the COG's PAR register.
'u2 is the index of the cog to start or restart.
              word  @WRD_COGNEW
              word  of_NONE                     'opcode flag set to normal
WRD_COGINIT   byte  8,"COG-INIT"              
              'no COD_XXX lable to save IDE symbol table space              
              word  NFW_LIT, cmd_COG_INIT, @COD_SYSTEM    
              word  NFW_EXIT
'
'P  COG-STOP ( u --  )
'Stop the COG with ID u (u = 0..7)
              word  @WRD_COGINIT
              word  of_NONE                     'opcode flag set to normal
WRD_COGSTOP   byte  8,"COG-STOP"              
COD_COGSTOP   word  NFW_LIT, cmd_COG_STOP, @COD_SYSTEM    
              word  NFW_EXIT
'
'P  COG_LOAD ( uParAdr uBlkNum -- n2 )
'Start the first available COG with block uBlkNum (uBlkNum = 0..blk-count -1)
'If the size of the COG image binary > 1024 the block n+1 will also be loaded
'It is assumed that the binary image was created with the Parallax spin IDE.
'This means the size is the 6th word in the image and the load address for the
'COG image is at offset 24 in the first block (16 for object init and 8 for
'dummy spin start method. uParAdr contains the value pasted to the COG
'PAR register, only the upper 14 bits are pasted (quad byte aligned).
'On exit the COG index is returned (1..7) or 0 if no COG is available.
              word  @WRD_COGSTOP
              word  of_NONE                     'opcode flag set to normal
WRD_COGLOAD   byte  8,"COG-LOAD"
              'save the input state incase we are LOADing from a block
              word  @COD_2TO_R, @COD_SAVEINP, @COD_2R_TO, @COD_BLK, @COD_STORE
              'Read first block into dictionary memory
              word  @COD_HERE, @COD_QALIGNED, @COD_DUP
              word  @COD_BLK, @COD_FETCH, @COD_BLOCK, @COD_ROT
              word  NFW_LIT, 1024, @COD_CMOVE
              'get the size of the COG object we are loading (6th word from start)
              word  @COD_DUP, NFW_LIT, 12, @COD_ADD, @COD_FETCH
              word  NFW_LIT, 1024, @COD_GT, NFW_QBRANCH, @COD_COGLOAD1
              'we need to load the a second block
              word  @COD_DUP, @COD_BLK, @COD_FETCH, @COD_ADD1, @COD_BLOCK
              word  @COD_SWAP, NFW_LIT, 1024, @COD_ADD
              word  NFW_LIT, 1024, @COD_CMOVE
              'we loaded what we needed to start the COG, save COG index to return stack
COD_COGLOAD1  word  NFW_LIT, 24, @COD_ADD, @COD_SWAP, @COD_COGNEW
              word  NFW_LIT, 2, @COD_MS
              word  @COD_TO_R, @COD_RESTINP, @COD_R_TO
              word  NFW_EXIT
'
'P  LOCK-NEW (  -- u )
'Allocate a lock, u is the lock ID (0..7) or -1 if no locks available 
              word  @WRD_COGLOAD
              word  of_NONE                     'opcode flag set to normal
WRD_LOCKNEW   byte  8,"LOCK-NEW"              
              'no COD_XXX lable to save IDE symbol table space              
              word  NFW_LIT, cmd_LOCK_NEW, @COD_SYSTEM    
              word  NFW_EXIT

'P  LOCK-RET ( u -- )
'Return lock u (0..7) making lock available 
              word  @WRD_LOCKNEW
              word  of_NONE                     'opcode flag set to normal
WRD_LOCKRET   byte  8,"LOCK-RET"
              'no COD_XXX lable to save IDE symbol table space              
              word  NFW_LIT, cmd_LOCK_RET, @COD_SYSTEM    
              word  NFW_EXIT
'
'P AUTORUN ( -- )
'Load block # 1 and execute. Called on a cold start.
'Prompts user for 5 seconds to cancel autorun on default console device (COM1).
              word  @WRD_LOCKRET
              word  of_NONE                     'opcode flag set to normal
WRD_AUTORUN   byte  7,"AUTORUN"              
COD_AUTORUN   word  NFW_XSQUOTE
              byte  21, "ESC to abort auto run"
              word  @COD_TYPE
              'wait up to 3 seconds for a key press
              word  NFW_LIT, 30, @COD_ZERO, @COD_XDO
COD_AUTORUN1  word  @COD_ISKEY, NFW_QBRANCH, @COD_AUTORUN2
              'we have a key see if it is an ESC key
              word  @COD_KEY, NFW_LIT, 27, @COD_EQ
              word  NFW_QBRANCH, @COD_AUTORUN2
              ' we have an ESC key abort auto load. 
              ' indicate we want to abort
              word  @COD_UNLOOP, NFW_BRANCH, @COD_AUTORUN3
              { wait 0.1 seconds }
COD_AUTORUN2  word  NFW_LIT, 100, @COD_MS
              word  NFW_XLOOP, @COD_AUTORUN1
              'no ESC key was presssed so do load
              word  NFW_LIT, 21, @COD_BSPACES
              word  @COD_ONE, @COD_LOAD
COD_AUTORUN3  word  @COD_CR
              word  NFW_EXIT
'

'Startup and intialize forth engine.
'This must be the last Forth word defined.
              word  @WRD_AUTORUN
              word  of_NONE                     'opcode flag set to normal
WRD_COLD      byte  4,"COLD"
              ' stop any cogs started by non core forth words              
COD_COLD      word  NFW_LIT, 8, NFW_LIT, 3, @COD_XDO
COD_COLD1     word  @COD_I, @COD_COGSTOP
              word  NFW_XLOOP, @COD_COLD1
              ' initialize all system variable
              word  @COD_INIT                   'initialize user area global variables
              word  @COD_BLKINIT                'Initialize block extension variables
             'indicate we are no longer evaluating a string
              word  @COD_FALSE, @COD_TO_SOURCE_ID
              ' Set COM1 as console device
              word  @COD_COM1CON
              ' sign on
              word  NFW_XSQUOTE
              byte  12,"MV4th16 0.92"
              word  @COD_TYPE, @COD_CR
              ' Perform auto run
              word  @COD_AUTORUN
              ' sign on
              word  NFW_XSQUOTE
              byte  3," OK"
              word  @COD_TYPE, @COD_CR
              word  NFW_BRANCH, @COD_QUIT       'QUIT never returns
              word  NFW_EXIT
'
'Must be last address of last declared forth word, after COLD.
'Need enough space to build native word list at startup. Space
'is reclaimed when MV4th16 starts.
UserCodeEnd    word  0 [512] '[c_ReservedWords]
'